/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "shared/core/RCEngine.h"
#include "shared/core/RCTableImpl.h"
#include "core/Transaction.h"

#include <string>

#include <boost/shared_ptr.hpp>

using namespace std;

void RCEngine::ApplyPendingChanges(THD* thd, const string& table_path, struct st_table_share* table_share) {}

void RCEngine::DropPendingChanges(THD* thd, const string& table_path, struct st_table_share* table_share) {
	RCTable* tab = GetTable(table_path, table_share);
	if(tab != NULL) {
		tab->Rollback(GetThreadParams(*thd)->GetTransactionID());
	}
}

bool RCEngine::RefreshTableApplyPendingChanges(RCTable* tab)
{
	return false;
}

void RCEngine::Commit(THD* thd, bool all)
{
	if(all || !(thd->options & OPTION_NOT_AUTOCOMMIT)) {
		unsigned transaction_id = 0;
		{
			IBGuard guard(threadparams_mutex);
			transaction_id = rceng->GetThreadParams(*thd)->GetTransactionID();
		}

		IBGuard guard(tables_mutex);
		TablesListAutoPtr tables_list = rceng->GetListOfModifiedTables(*thd);
		BHASSERT(tables_list.get(), "Unable to obtain a list of tables to commit.");

		if(tables_list->size()) {
			TablesList::iterator iter = tables_list->begin();
			TablesList::iterator end = tables_list->end();

			for(; iter != end; ++iter)
				ApplyPendingChanges(thd, (*iter), NULL);

			VerifyTablesIntegrity(*tables_list);

			for(iter = tables_list->begin(); iter != end; ++iter) {
				RCTable* tab = GetTable(*iter, NULL);
				if(tab != NULL)
					tab->CommitSaveSession(transaction_id, 0);
				table_lock_manager.ReleaseLocks(*thd, *iter);
			}

			if(rsi_manager)
				rsi_manager->UpdateDef();
		}
		if(all && thd->lock)
			thd->lock = 0;
		thd->server_status &= ~SERVER_STATUS_IN_TRANS;

		rceng->GetThreadParams(*thd)->GetTransaction()->Commit();
		rceng->GetThreadParams(*thd)->EndTransaction();
	}
}

RCTablePtr RCEngine::RefreshTable(const char* table_path, uint session_id, bool force) throw(DatabaseRCException)
{
	char db_name[256] = { 0 };
	char tab_name[256] = { 0 };
	GetNames(table_path, db_name, tab_name, 256);
	//pthread_mutex_lock(&m_sync_tables);
    tables_mutex.Lock();
	RCTablePtr tab = m_tables[db_name][tab_name];

	if(tab && !force) {
        tables_mutex.Unlock();
		return tab;
	}

	string tab_path = table_path;
	tab_path += m_tab_ext;
	vector<DTCollation> charsets;
	if(tab)
		charsets = tab->GetCharsets();
	try {
		tab = (RCTablePtr)(new RCTable(tab_path.c_str(), charsets, -1, 0, true));
	} catch (DatabaseRCException&) {
        tables_mutex.Unlock();
		//pthread_mutex_unlock(&m_sync_tables);
		throw;
	}

	m_tables[db_name][tab_name] = tab;
	// load dirty data if session_id <> 0
	if(session_id != 0) {
		tab->LoadDirtyData(session_id);
	}
    tables_mutex.Unlock();
	//pthread_mutex_unlock(&m_sync_tables);
}
