/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "common/CommonDefinitions.h"
#include "core/RCAttrPack.h"
#include "core/tools.h"
#include "core/WinTools.h"
#include "RCAttr_load.h"
#include "shared/loader/NewValueSet.h"

RCAttrLoad::RCAttrLoad(int a_num,int t_num,char *a_path,int conn_mode,unsigned int s_id, DTCollation collation) throw(DatabaseRCException)
	: RCAttrLoadBase(a_num, t_num, a_path, conn_mode, s_id, collation )
{

}

RCAttrLoad::~RCAttrLoad()
{
	LogWarnigs();
}

int RCAttrLoad::LoadData(NewValuesSet* nvs)
{
	if(PackType() == PackN) {
		if(rsi_manager && rsi_hist_update == NULL)
			rsi_hist_update = (RSIndex_Hist*)rsi_manager->GetIndexForUpdate(
							RSIndexID(RSI_HIST, table_number, attr_number), GetCurReadLocation());
	} else {
		if(rsi_manager && rsi_cmap_update == NULL)
			rsi_cmap_update = (RSIndex_CMap*)rsi_manager->GetIndexForUpdate(
							RSIndexID(RSI_CMAP, table_number, attr_number), GetCurReadLocation());
	}
#ifdef FUNCTIONS_EXECUTION_TIMES
	double start = clock();
#endif
	if(nvs->NoValues() == 0)
		return 0;
	LoadPackInfoForLoader();
	bool new_pack_created = false;
	if(NoPack() == 0 || loc_no_obj[NoPack()-packs_omitted-1] == 0xFFFF) {
		CreateNewPackage();
		new_pack_created = true;
	} else {
		LockPackForUse(NoPack()-packs_omitted-1);
	}
	int load_nulls = 0;
	int load_obj = nvs->NoValues();
	int current_pack = NoPack()-packs_omitted-1;
	uint cur_obj = (uint)loc_no_obj[current_pack] + 1;
	if(pack_file[current_pack] == PF_NO_OBJ)
		cur_obj = 0;
	uint cur_nulls = loc_no_nulls[current_pack];

	_int64 load_min, load_max;
	if(!ATI::IsRealType(TypeName())) {
		load_min=PLUS_INF_64;
		load_max=MINUS_INF_64;
	} else {
		load_min = *(_int64*)&PLUS_INF_DBL;
		load_max = *(_int64*)&MINUS_INF_DBL;
	}

	bool is_real_type = ATI::IsRealType(TypeName());
	double& real_sum = *(double*)&local_sum[current_pack];
	bool nulls_conferted = false;
	if(this->PackType() == PackN) {
		_int64 null_value = -1;
		bool null = false;
		int nonv = nvs->NoValues();
		for(int i = 0; i < nonv; i++) {
			bool insert = false;
			_int64 v = 0;
			null = false;
			if(nvs->IsNull(i)) {
				null = true;
				if(Type().GetNullsMode() == NO_NULLS) {
					nulls_conferted = true;
					illegal_nulls = true;
					if(ATI::IsStringType(TypeName())) {
						if(null_value == -1)
							null_value = v = EncodeValue_T(ZERO_LENGTH_STRING, 1);
						else
							v = null_value;
					} else {
						rclog << DecodeValue_S(0).val << ".";
						null_value = v = 0;
					}
					null = false;
				} else
					load_nulls++;
			}
		}
		if(!is_real_type) {
			_int64 tmp_sum;
			nvs->GetIntStats(load_min, load_max, tmp_sum);
			local_sum[current_pack] += tmp_sum;
			if(nulls_conferted) {
				if(load_min > null_value)
					load_min = null_value;
				if(load_max < null_value)
					load_max = null_value;
			}
		} else {
			double tmp_sum;
			nvs->GetRealStats(*(double*)&load_min, *(double*)&load_max, tmp_sum);
			real_sum += tmp_sum;
			if(nulls_conferted) {
				if(*(double*)&load_min > *(double*)&null_value)
					*(double*)&load_min = *(double*)&null_value;
				if(*(double*)&load_max < *(double*)&null_value)
					*(double*)&load_max = *(double*)&null_value;
			}
		}

		bool is_exclusive = false; // true if the new data lie outside of the current i_min-i_max
		SetNaturalSizeSaved(GetNaturalSizeSaved() + nvs->SumarizedSize()); //this is needed for LOOKUP column
		if(load_nulls != nvs->NoValues()) {
			if(NoObj() == 0) {
				SetMinInt64(load_min);
				SetMaxInt64(load_max);
				is_exclusive = true;
			} else {
				if(!ATI::IsRealType(TypeName())) {
					is_exclusive = (GetMinInt64()>load_max || GetMaxInt64()<load_min);
					if(GetMinInt64()>load_min)
						SetMinInt64(load_min);
					if(GetMaxInt64()<load_max)
						SetMaxInt64(load_max);
				} else {
					_int64 a_min = GetMinInt64();
					_int64 a_max = GetMaxInt64();
					is_exclusive = (*(double*)(&a_min)>*(double*)(&load_max) || *(double*)(&a_max)
							<*(double*)(&load_min));
					if(*(double*)(&a_min)>*(double*)(&load_min))
						SetMinInt64(load_min);
					if(*(double*)(&a_max)<*(double*)(&load_max))
						SetMaxInt64(load_max); // 1-level statistics
				}
			}
		}
		if((cur_nulls + load_nulls) == 0 && load_min == load_max && (cur_obj == 0 || (local_min[current_pack]
				== load_min && local_max[current_pack] == load_max))) {// no need to store any values - uniform package
			pack_mode[current_pack] = 0;
			is_stored[current_pack] = false;
			pack_file[current_pack] = PF_NOT_KNOWN; // will not be used
			local_min[current_pack] = load_min;
			local_max[current_pack] = load_max;
			// sum is already calculated
			if(IsUniqueUpdated() && IsUnique() && is_exclusive) {
				if(load_obj-load_nulls==1)
					SetUnique(true); // only one new value, different than the previous
				else
					SetUnique(false); // at least two identical values found
			} else
				SetUniqueUpdated(false); // may be unique or not
		} else if(load_nulls == load_obj && (pack_file[NoPack()-packs_omitted-1] == PF_NO_OBJ ||
				pack_file[NoPack()-packs_omitted-1] == PF_NULLS_ONLY)) {
			// no need to store any values - uniform package (nulls only)
			pack_mode[current_pack]=0;
			is_stored[current_pack] = false;
			pack_file[current_pack]=PF_NULLS_ONLY;
			local_min[current_pack]=PLUS_INF_64;
			local_max[current_pack]=MINUS_INF_64;
			local_sum[current_pack]=0;
			// uniqueness status not changed
		} else {
			_uint64 new_max_val = 0;

			if(cur_obj==0 || pack_mode[current_pack] == 0 || pack_mode[current_pack] == 3) {
				// new package (also in case of expanding so-far-uniform package)
				_uint64 uniform_so_far = 0;
				if(pack_file[current_pack]==PF_NULLS_ONLY) {
					uniform_so_far=(_uint64)NULL_VALUE_64;
					local_min[current_pack]=load_min;
					local_max[current_pack]=load_max;
				} else {
					if(ATI::IsRealType(TypeName())) {
						uniform_so_far=local_min[current_pack]; // fill with uniform-so-far
					} else if(local_min[current_pack]>load_min)
						uniform_so_far=_uint64(local_min[current_pack]-load_min);

					if(!ATI::IsRealType(TypeName())) {
						if(cur_obj==0 || local_min[current_pack]>load_min)
							local_min[current_pack]=load_min;
						if(cur_obj==0 || local_max[current_pack]<load_max)
							local_max[current_pack]=load_max;
					} else {
						if(cur_obj==0 || *(double*)&(local_min[current_pack])>*(double*)&(load_min))
							local_min[current_pack]=load_min;
						if(cur_obj==0 || *(double*)&(local_max[current_pack])<*(double*)&(load_max))
							local_max[current_pack]=load_max;
					}
				}

				BHASSERT(packs[current_pack], "'packs[current_pack]' should not be null");
				if(ATI::IsRealType(TypeName())) {
					// full 64-bit scope
					((AttrPackN*)packs[current_pack])->Prepare(cur_obj+load_obj,
							_uint64(0xFFFFFFFFFFFFFFFFLL));
				} else {
					new_max_val=_uint64(local_max[current_pack]-local_min[current_pack]);
					((AttrPackN*)packs[current_pack])->Prepare(cur_obj + load_obj, new_max_val);
				}
				pack_file[current_pack] = PF_NOT_KNOWN;
				pack_mode[current_pack] = 2;
				is_stored[current_pack] = true;
				// Now fill the beginning of the table by so far uniform values (if there is any beginning)
				if(cur_obj > 0) {
					if(uniform_so_far != NULL_VALUE_64) {
						if((local_min[current_pack] != local_max[current_pack]) || (load_nulls > 0))
							for(int i = 0; i < (int)cur_obj; i++)
								((AttrPackN*)packs[current_pack])->SetVal64(i, uniform_so_far);
					} else {
						for(int i = 0; i < (int)cur_obj; i++) {
							packs[current_pack]->SetNull(i);
						}
					}
				}
			} else {
				// expand existing package
				if(ATI::IsRealType(TypeName())) {
					if(*(double*)&local_min[current_pack]>*(double*)&load_min)
						*(double*)&local_min[current_pack]=*(double*)&load_min;
					if(*(double*)&local_max[current_pack]<*(double*)&load_max)
						*(double*)&local_max[current_pack]=*(double*)&load_max;
					((AttrPackN*)packs[current_pack])->Expand(cur_obj+load_obj,
							_uint64(0xFFFFFFFFFFFFFFFFLL), 0);
				} else {
					_int64 offset=0;
					if(local_min[current_pack]>load_min) {
						offset=local_min[current_pack];
						offset-=load_min;
						local_min[current_pack]=load_min;
					}
					if(local_max[current_pack]<load_max)
						local_max[current_pack]=load_max;
					BHASSERT(packs[current_pack], "'packs[current_pack]' should not be null");
					new_max_val=_uint64(local_max[current_pack]-local_min[current_pack]);
					((AttrPackN*)packs[current_pack])->Expand(cur_obj+load_obj, new_max_val, offset);
				}
			}

			_uint64 v = 0;
			int obj = 0;
			bool isnull;
			for(int i = 0; i < load_obj; i++) {
				isnull = nvs->IsNull(i);
				if(isnull && Type().GetNullsMode() == NO_NULLS) {
					isnull = false;
					v = null_value;
				} else
					v = *(_uint64*)nvs->GetDataBytesPointer(i);
				obj = i + cur_obj;
				if(local_min[current_pack] == local_max[current_pack] && !ATI::IsRealType(TypeName())) {// special case: no data stored except nulls
					if(isnull)
						packs[current_pack]->SetNull(obj);
				} else {
					if(!ATI::IsRealType(TypeName()))
						v -= local_min[current_pack];
					if(isnull) {
						//((AttrPackN*)packs[NoPack()-packs_omitted-1])->SetVal64(obj, 0);
						((AttrPackN*)packs[current_pack])->SetNull(obj);
					} else
						((AttrPackN*)packs[current_pack])->SetVal64(obj, v);
				}
			}
			if(IsUniqueUpdated() && IsUnique() && is_exclusive // there is a chance for uniqueness
					&& !ATI::IsRealType(TypeName()) // temporary limitations
					&& _int64(local_max[current_pack] - local_min[current_pack]) > 0
					&& _int64(local_max[current_pack] - local_min[current_pack]) < 8000000) {
				Filter f_val(local_max[current_pack] - local_min[current_pack] + 1);
				f_val.Reset();
				bool repetition_found = false;
				if(local_min[current_pack] == local_max[current_pack]) {
					if(cur_obj+load_obj-loc_no_nulls[current_pack]-load_nulls>1) // more than one non-null uniform object?
						repetition_found = true;
				} else
					for(int i = 0; i < load_obj; i++)
						if(nvs->IsNull(i)==false) {
							BHASSERT_WITH_NO_PERFORMANCE_IMPACT(*(_int64*)nvs->GetDataBytesPointer(i) <= local_max[current_pack]);
							BHASSERT_WITH_NO_PERFORMANCE_IMPACT(*(_int64*)nvs->GetDataBytesPointer(i) >= local_min[current_pack]);
							v = *(_int64*)nvs->GetDataBytesPointer(i) - local_min[current_pack];
							if(f_val.Get(v)) {
								repetition_found = true;
								break;
							}
							f_val.Set(v);
						}
				if(repetition_found)
					SetUnique(false); // else remains true
			} else
				SetUniqueUpdated(false);
		}
	} else if(PackType() == PackS) {
		for(int i = 0; i < nvs->NoValues(); i++) {
			if(nvs->IsNull(i)) {
				if(Type().GetNullsMode() == NO_NULLS) {
					illegal_nulls = true;
				} else
					load_nulls++;
			}
		}
		SetNaturalSizeSaved(GetNaturalSizeSaved() + nvs->SumarizedSize());
		AttrPackS*& packS = ((AttrPackS*&)packs[current_pack]);
		// no need to store any values - uniform package
		if(load_nulls == load_obj && (pack_file[current_pack]==PF_NO_OBJ || pack_file[current_pack]==PF_NULLS_ONLY)) {
			pack_mode[current_pack] = 0;
			is_stored[current_pack] = false;
			pack_file[current_pack] = PF_NULLS_ONLY;
			local_min[current_pack] = 0;
			local_max[current_pack] = -1;
			local_sum[current_pack] = 0;
			local_size[current_pack] = 0;
		} else {
			BHASSERT(packS, "'packS' should not be null"); //pack object must exist
			SetUnique(false); // we will not check uniqueness for strings now
			SetUniqueUpdated(false);
			if(cur_obj == 0 || pack_mode[current_pack] == 0 || pack_mode[current_pack] == 3)// new package (also in case of expanding so-far-null package)
			{
				packS->Prepare(loc_no_nulls[current_pack]); // data size: one byte for each null, plus more...
				pack_mode[current_pack] = 2; // status: "in memory"
				is_stored[current_pack] = true;
				pack_file[current_pack] = PF_NOT_KNOWN;
			}
			if(!new_pack_created && current_pack == 0 &&
					local_min[current_pack] == 0 && local_max[current_pack] == -1 &&
					GetPackOntologicalStatus(current_pack) != NULLS_ONLY) {
				RCDataTypePtr min_ptr;
				RCDataTypePtr max_ptr;
				GetMinMaxValuesPtrs(current_pack, min_ptr, max_ptr);
				if(!min_ptr->IsNull())
					strncpy((uchar*)(&local_min[current_pack]), *(RCBString*)(min_ptr.get()), sizeof(_int64));
				if(!max_ptr->IsNull())
					strncpy((uchar*)(&local_max[current_pack]), *(RCBString*)(min_ptr.get()),
							min((*(RCBString*)(min_ptr.get())).size(), sizeof(_int64)));
			}

			RCBString min_s;
			RCBString max_s;
			ushort max_size = 0;
			nvs->GetStrStats(min_s, max_s, max_size);

			packS->Expand(nvs->NoValues());
			bool isnull = false;
			char const* v = 0;
			uint size = 0;
			RCBString zls(ZERO_LENGTH_STRING, 0);
			for(int i = 0; i < nvs->NoValues(); i++) {
				isnull = nvs->IsNull(i);
				if(isnull && Type().GetNullsMode() == NO_NULLS) {
					isnull = false;
					v = ZERO_LENGTH_STRING;
					size = 0;
					if(min_s.IsNull() || min_s > zls)
						min_s = zls;
					if(max_s.IsNull() || max_s < zls)
						max_s = zls;
				} else {
					v = nvs->GetDataBytesPointer(i);
					size = nvs->Size(i);
				}
				packS->BindValue(isnull, (uchar*)v, size);
			}

			if(new_pack_created || GetPackOntologicalStatus(current_pack) == NULLS_ONLY) {
				if(!min_s.IsNull())
					SetPackMin(current_pack, min_s);
					//strncpy((uchar*)(&local_min[current_pack]), min_s, sizeof(_int64));
				if(!max_s.IsNull())
					SetPackMax(current_pack, max_s);
					//strncpy((uchar*)(&local_max[current_pack]), max_s, min(max_s.size(), sizeof(_int64)));
				local_size[current_pack] = max_size;
			} else {
				if(!min_s.IsNull() && !min_s.GreaterEqThanMinUTF(local_min[current_pack], Type().GetCollation(), true))
					SetPackMin(current_pack, min_s);
					//strncpy((uchar*)(&local_min[current_pack]), min_s, sizeof(_int64));

				if(!max_s.IsNull() && !max_s.LessEqThanMaxUTF(local_max[current_pack], Type().GetCollation(), true))
					SetPackMax(current_pack, max_s);
					//strncpy((uchar*)(&local_max[current_pack]), max_s, sizeof(_int64));

				if(local_size[current_pack] < max_size)
					local_size[current_pack] = max_size;
			}
			pack_mode[current_pack] = 2;
			is_stored[current_pack] = true;
		}
	}

	loc_no_nulls[current_pack] += load_nulls;
	loc_no_obj[current_pack] = ushort(cur_obj + load_obj - 1);
	SetNoObj(NoObj() + load_obj);
	//no_nulls += load_nulls;
	_int64 nnulls = NoNulls() + load_nulls;
	SetNoNulls(nnulls);
#ifdef FUNCTIONS_EXECUTION_TIMES
	double step3 = clock();
#endif
	if(loc_no_obj[current_pack] == 0xFFFF) {


		SavePack(current_pack);
	} else if(PackType() == PackS && packs[current_pack] && !packs[current_pack]->IsEmpty())
		// pack not full, because no more data in the current buf
		// copy data from buffer to free the buffer
		// alternative - mark use of the current buffer,
		//   release it when data pack will get full with data from the next buffer
		//   - difficult to design and implement
		((AttrPackS*&)packs[current_pack])->CopyBinded();

#ifdef FUNCTIONS_EXECUTION_TIMES
	double end = clock();
	stringstream s;
	s << "a[" << attr_number << "].LoadData(...)";
	fet->Add(s.str(), end - start);
#endif

	return 0;
}


void RCAttrLoad::LoadPack(uint n)
{
	BHASSERT(n < NoPack()-packs_omitted, "should be 'n < no_pack-packs_omitted'");
	if(pack_mode[n]!=1 || !ShouldExist(n))
		return;

	LoadPackInherited(n);
}


int RCAttrLoad::Save()
{
	if(current_state!=1)
	{
		rclog << lock << "Error: cannot save. It is read only session." << unlock;
		throw;
	}
	LoadPackInfo();

	int npack=NoPack();
	for(uint i=0;i<npack-packs_omitted;i++)
		SavePack(i);

	if(rsi_hist_update || rsi_cmap_update)
		SaveRSI();
	BHASSERT(FileFormat()==10, "should be 'file_format==10'");
	SaveDPN();
	SaveHeader();
	return 0;
}

int RCAttrLoad::DoSavePack(int n)		// WARNING: assuming that all packs with lower numbers are already saved!
{
	if(current_state != 1)
	{
		rclog << lock << "Error: cannot save. Read only session." << unlock;
		throw;
	}
	LoadPackInfo();

	int p_type=-1;

	if(!packs[n] || packs[n]->IsEmpty() || pack_mode[n]==0 || pack_mode[n]==3)
	{
		return 0;
	}

	if(packs[n]->UpToDate())
	{

		return 2;
	}

	double start_spt = clock();
	int fout = -1;
	RCAttrLoad* rcattr = this;
	//SetEvent(params->spevent);

	rcattr->packs[n]->Compress();

	//////////////// Update RSI ////////////////

	rcattr->UpdateRSI_Hist(n);		// updates pack n in RSIndices (NOTE: it may be executed in a random order of packs),
	rcattr->UpdateRSI_CMap(n);		// but does not save it to disk. Use SaveRSI to save changes (after the whole load)


	uint fsize = 0;
	IBFile fattr;
	if(rcattr->pack_file[n] == PF_NOT_KNOWN)
		fsize = 0;
	else {
		char file_name[1024];

			try {
				fattr.OpenCreate(rcattr->AttrPackFileName(file_name,1024,n));
				fattr.Seek(0,SEEK_END);
				fsize = fattr.Tell();
				fattr.Close();
			} catch(DatabaseRCException& e)	{
				rclog << lock << "Error: cannot get filesize" << file_name << ". " << GetErrorMessage(errno) << unlock;
				throw e;
			}
	}

	if(rcattr->pack_file[n]==PF_NOT_KNOWN || fsize > rcattr->pack_addrs[n]+ rcattr->packs[n]->PreviousSaveSize())					// new package
	{
		rcattr->pack_file[n]=rcattr->GetSaveFileLoc(rcattr->GetCurSaveLocation());
		rcattr->pack_addrs[n]=rcattr->GetSavePosLoc(rcattr->GetCurSaveLocation());
	}
	else if(rcattr->pack_file[n]!=PF_NULLS_ONLY && rcattr->pack_file[n]!=PF_NO_OBJ &&
		rcattr->pack_file[n]%2!=rcattr->GetCurSaveLocation())  // or an existing one, but from the wrong file (switching session)
	{
		rcattr->SetSaveFileLoc(1-rcattr->GetCurSaveLocation(), rcattr->pack_file[n]);
		rcattr->SetSavePosLoc(1-rcattr->GetCurSaveLocation(),rcattr->pack_addrs[n]);
		rcattr->pack_file[n]=rcattr->GetSaveFileLoc(rcattr->GetCurSaveLocation());
		rcattr->pack_addrs[n]=rcattr->GetSavePosLoc(rcattr->GetCurSaveLocation());
	}
	////////////////////////////////////////////////////////
	// Now the package has its final location.
	if ((rcattr->pack_mode[n]==1 || rcattr->pack_mode[n]==2) && rcattr->pack_file[n]>=rcattr->GetTotalPackFile())	// find the largest used file number
		rcattr->SetTotalPackFile(rcattr->pack_file[n]+1);
	BHASSERT(rcattr->pack_file[n] >= 0, "should be 'rcattr->pack_file[n] >= 0'");
	uint pack_des = rcattr->pack_addrs[n];
	if(rcattr->packs[n] && !rcattr->packs[n]->IsEmpty())
		rcattr->SetSavePosLoc(rcattr->GetCurSaveLocation(), rcattr->pack_addrs[n] + rcattr->packs[n]->TotalSaveSize());
	if( rcattr->GetSavePosLoc(rcattr->GetCurSaveLocation()) > rcattr->file_size_limit)	// more than 1.5 GB - start the next package from the new file
	{
		int cur_save_loc = rcattr->GetCurSaveLocation();
		rcattr->SetSavePosLoc(cur_save_loc, 0);
		int save_location = rcattr->GetSaveFileLoc(cur_save_loc);
		rcattr->SetSaveFileLoc(cur_save_loc, save_location + 2);
	}

	////////////////////////////////////////////////////////
	char file_name[1024];

	try
	{
		fattr.OpenCreate(rcattr->AttrPackFileName(file_name,1024,n));
		fattr.Seek(pack_des, SEEK_SET);
		rcattr->packs[n]->Save(&fattr);
		fattr.Close();
	}
	catch(DatabaseRCException& e)
	{
		rclog << lock << "Error: can not save data pack to " << file_name << ". " << GetErrorMessage(errno) << unlock;
		throw e;
	}

	RCAttrLoad::count_spt--;
	rcattr->UnlockPackFromUse(n);
	return 0;
}

