/* Copyright (C)  2005-2009 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "AggregatorAdvanced.h"
#include <math.h>

void AggregatorVarPop64::PutAggregatedValue(unsigned char *buf, _int64 v, _int64 factor)
{
	if(v != NULL_VALUE_64) {
		(*((_int64*)buf)) += factor;
		*((double*)(buf + 8)) += factor * double(v) / prec_factor;
		*((double*)(buf + 16)) += factor * (double(v) / prec_factor) * (double(v) / prec_factor);
	}
}

_int64 AggregatorVarPop64::GetValue64(unsigned char *buf)
{
	if(NoObj(buf) < 1)
		return NULL_VALUE_64;
	double vd = VarPop(buf);
	return *(_int64*)(&vd);
}

///////////////////////////////////////////////////////////////////////////////////////////////

void AggregatorVarSamp64::PutAggregatedValue(unsigned char *buf, _int64 v, _int64 factor)
{
	if(v != NULL_VALUE_64) {
		(*((_int64*)buf)) += factor;
		*((double*)(buf + 8)) += factor * double(v) / prec_factor;
		*((double*)(buf + 16)) += factor * (double(v) / prec_factor) * (double(v) / prec_factor);
	}
}

_int64 AggregatorVarSamp64::GetValue64(unsigned char *buf)
{
	if(NoObj(buf) < 2)
		return NULL_VALUE_64;
	double vd = VarSamp(buf);
	return *(_int64*)(&vd);
}

///////////////////////////////////////////////////////////////////////////////////////////////

void AggregatorStdPop64::PutAggregatedValue(unsigned char *buf, _int64 v, _int64 factor)
{
	if(v != NULL_VALUE_64) {
		(*((_int64*)buf)) += factor;
		*((double*)(buf + 8)) += factor * double(v) / prec_factor;
		*((double*)(buf + 16)) += factor * (double(v) / prec_factor) * (double(v) / prec_factor);
	}
}

_int64 AggregatorStdPop64::GetValue64(unsigned char *buf)
{
	if(NoObj(buf) < 1)
		return NULL_VALUE_64;
	double vd = sqrt(VarPop(buf));
	return *(_int64*)(&vd);
}

///////////////////////////////////////////////////////////////////////////////////////////////

void AggregatorStdSamp64::PutAggregatedValue(unsigned char *buf, _int64 v, _int64 factor)
{
	if(v != NULL_VALUE_64) {
		(*((_int64*)buf)) += factor;
		*((double*)(buf + 8)) += factor * double(v) / prec_factor;
		*((double*)(buf + 16)) += factor * (double(v) / prec_factor) * (double(v) / prec_factor);
	}
}

_int64 AggregatorStdSamp64::GetValue64(unsigned char *buf)
{
	if(NoObj(buf) < 2)
		return NULL_VALUE_64;
	double vd = sqrt(VarSamp(buf));
	return *(_int64*)(&vd);
}

///////////////////////////////////////////////////////////////////////////////////////////////

void AggregatorVarPopD::PutAggregatedValue(unsigned char *buf, _int64 v, _int64 factor)
{
	if(v != NULL_VALUE_64) {
		(*((_int64*)buf)) += factor;
		*((double*)(buf + 8)) += *((double*)(&v)) * factor;
		*((double*)(buf + 16)) += *((double*)(&v)) * *((double*)(&v)) * factor;
	}
}

_int64 AggregatorVarPopD::GetValue64(unsigned char *buf)
{
	if(NoObj(buf) < 1)
		return NULL_VALUE_64;
	double vd = VarPop(buf);
	return *(_int64*)(&vd);
}

void AggregatorVarPopD::PutAggregatedValue(unsigned char *buf, const RCBString& v, _int64 factor)
{
	stats_updated = false;
	RCNum val(RC_REAL);
	if(!v.IsEmpty() && RCNum::ParseReal(v, val, RC_REAL) == BHRC_SUCCESS && !val.IsNull()) {
		double d_val = double(val);
		PutAggregatedValue(buf, *((_int64*)(&d_val)), factor);
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////


void AggregatorVarSampD::PutAggregatedValue(unsigned char *buf, _int64 v, _int64 factor)
{
	if(v != NULL_VALUE_64) {
		(*((_int64*)buf)) += factor;
		*((double*)(buf + 8)) += *((double*)(&v)) * factor;
		*((double*)(buf + 16)) += *((double*)(&v)) * *((double*)(&v)) * factor;
	}
}

_int64 AggregatorVarSampD::GetValue64(unsigned char *buf)
{
	if(NoObj(buf) < 2)
		return NULL_VALUE_64;
	double vd = VarSamp(buf);
	return *(_int64*)(&vd);
}

void AggregatorVarSampD::PutAggregatedValue(unsigned char *buf, const RCBString& v, _int64 factor)
{
	stats_updated = false;
	RCNum val(RC_REAL);
	if(!v.IsEmpty() && RCNum::ParseReal(v, val, RC_REAL) == BHRC_SUCCESS && !val.IsNull()) {
		double d_val = double(val);
		PutAggregatedValue(buf, *((_int64*)(&d_val)), factor);
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////

void AggregatorStdPopD::PutAggregatedValue(unsigned char *buf, _int64 v, _int64 factor)
{
	if(v != NULL_VALUE_64) {
		(*((_int64*)buf)) += factor;
		*((double*)(buf + 8)) += *((double*)(&v)) * factor;
		*((double*)(buf + 16)) += *((double*)(&v)) * *((double*)(&v)) * factor;
	}
}

_int64 AggregatorStdPopD::GetValue64(unsigned char *buf)
{
	if(NoObj(buf) < 1)
		return NULL_VALUE_64;
	double vd = sqrt(VarPop(buf));
	return *(_int64*)(&vd);
}

void AggregatorStdPopD::PutAggregatedValue(unsigned char *buf, const RCBString& v, _int64 factor)
{
	stats_updated = false;
	RCNum val(RC_REAL);
	if(!v.IsEmpty() && RCNum::ParseReal(v, val, RC_REAL) == BHRC_SUCCESS && !val.IsNull()) {
		double d_val = double(val);
		PutAggregatedValue(buf, *((_int64*)(&d_val)), factor);
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////

void AggregatorStdSampD::PutAggregatedValue(unsigned char *buf, _int64 v, _int64 factor)
{
	if(v != NULL_VALUE_64) {
		(*((_int64*)buf)) += factor;
		*((double*)(buf + 8)) += *((double*)(&v)) * factor;
		*((double*)(buf + 16)) += *((double*)(&v)) * *((double*)(&v)) * factor;
	}
}

_int64 AggregatorStdSampD::GetValue64(unsigned char *buf)
{
	if(NoObj(buf) < 2)
		return NULL_VALUE_64;
	double vd = sqrt(VarSamp(buf));
	return *(_int64*)(&vd);
}

void AggregatorStdSampD::PutAggregatedValue(unsigned char *buf, const RCBString& v, _int64 factor)
{
	stats_updated = false;
	RCNum val(RC_REAL);
	if(!v.IsEmpty() && RCNum::ParseReal(v, val, RC_REAL) == BHRC_SUCCESS && !val.IsNull()) {
		double d_val = double(val);
		PutAggregatedValue(buf, *((_int64*)(&d_val)), factor);
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////

void AggregatorBitAnd::PutAggregatedValue(unsigned char *buf, const RCBString& v, _int64 factor)
{
	stats_updated = false;
	RCNum val(RC_BIGINT);
	if(!v.IsEmpty() && RCNum::Parse(v, val, RC_BIGINT) == BHRC_SUCCESS && !val.IsNull()) {
		PutAggregatedValue(buf, _int64(val), factor);
	}
}

void AggregatorBitOr::PutAggregatedValue(unsigned char *buf, const RCBString& v, _int64 factor)
{
	stats_updated = false;
	RCNum val(RC_BIGINT);
	if(!v.IsEmpty() && RCNum::Parse(v, val, RC_BIGINT) == BHRC_SUCCESS && !val.IsNull()) {
		PutAggregatedValue(buf, _int64(val), factor);
	}
}

void AggregatorBitXor::PutAggregatedValue(unsigned char *buf, const RCBString& v, _int64 factor)
{
	stats_updated = false;
	RCNum val(RC_BIGINT);
	if(!v.IsEmpty() && RCNum::Parse(v, val, RC_BIGINT) == BHRC_SUCCESS && !val.IsNull()) {
		PutAggregatedValue(buf, _int64(val), factor);
	}
}
