/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include <boost/bind.hpp>

#include "common/defs.h"
#include "CompiledQuery.h"

#include "Query.h"
#include "MysqlExpression.h"
#include "vc/MultiValColumn.h"

using namespace std;

CompiledQuery::CompiledQuery()
{
	no_tabs = 0;
	no_params = 0;
	no_filters = 0;
}

CompiledQuery::CompiledQuery(CompiledQuery const & cq)
{
	this->steps = cq.steps;

	no_tabs = cq.no_tabs;
	no_attrs = cq.no_attrs;
	no_params = cq.no_params;
	no_filters = cq.no_filters;
	no_virt_cols = cq.no_virt_cols;
}

CompiledQuery::~CompiledQuery()
{
	steps.clear();
}

CompiledQuery& CompiledQuery::operator=(CompiledQuery const &cq)
{
	this->steps = cq.steps;

	no_tabs=cq.no_tabs;
	no_attrs=cq.no_attrs;
	no_params=cq.no_params;
	no_filters=cq.no_filters;
	no_virt_cols = cq.no_virt_cols;

	return *this;
}

CompiledQuery::CQStep::CQStep(const CompiledQuery::CQStep & s)
	: type( s.type ),
	t1( s.t1 ),
	t2( s.t2 ),
	t3( s.t3 ),
	a1( s.a1 ),
	a2( s.a2 ),
	f1( s.f1 ),
	f2( s.f2 ),
	f3( s.f3 ),
	p1( s.p1 ),
	e1( s.e1 ),
	e2( s.e2 ),
	e3( s.e3 ),
	op( s.op ),
	tmpar( s.tmpar ),
	jt( s.jt ),
	cop( s.cop ),
	alias( NULL ),
	mysql_expr( s.mysql_expr ),
	virt_cols(s.virt_cols),
	n1( s.n1 ),
	n2( s.n2 )
{

	if(s.alias) {
		size_t alias_ct(strlen(s.alias)+1);
		alias = new char[alias_ct];
		strcpy(alias, s.alias);
	}
	else
		alias = NULL;
}

CompiledQuery::CQStep & CompiledQuery::CQStep::operator=(const CompiledQuery::CQStep & s)
{
	if(&s != this) {
		CQStep tmp_step(s);
		swap(tmp_step);
	}
	return *this;
}
void CompiledQuery::CQStep::swap(CQStep& s) {
	if(&s != this) {
		using std::swap;
		swap(type, s.type);
		swap(t1, s.t1);
		swap(t2, s.t2);
		swap(t3, s.t3);
		swap(a1, s.a1);
		swap(a2, s.a2);
		swap(f1, s.f1);
		swap(f2, s.f2);
		swap(f3, s.f3);
		swap(p1, s.p1);
		swap(e1, s.e1);
		swap(e2, s.e2);
		swap(e3, s.e3);
		swap(op, s.op);
		swap(tmpar, s.tmpar);
		swap(jt, s.jt);
		swap(cop, s.cop);
		swap(n1, s.n1);
		swap(n2, s.n2);
		swap(mysql_expr, s.mysql_expr);
		swap(virt_cols, s.virt_cols);
		swap(alias, s.alias);
	}
}

void string_add(char ** str, int* size_left, const char * buf)
{
	if (*size_left > 0) {
		//memcpy_s(*str, *size_left, buf, std::min(*size_left, (int) strlen(buf)));
		memcpy(*str, buf, std::min(*size_left, (int) strlen(buf)) );
		*size_left -= (int)strlen(buf);
		*str += strlen(buf);  //may be out of range, but at the same time *size_left will be <0
	};
};

void CompiledQuery::CQStep::Print(char** str, int* size_left)
{
	char buf[512] = "";
	char b1[100], b2[100], b3[100];
	b1[0] = b2[0] = b3[0] = '\0';

	char b_op[20];
	switch(op) {
		case O_EQ:
			strcpy(b_op,"=");
			break;
		case O_EQ_ALL:
			strcpy(b_op,"=ALL");
			break;
		case O_EQ_ANY:
			strcpy(b_op,"=ANY");
			break;
		case O_NOT_EQ:
			strcpy(b_op,"<>");
			break;
		case O_NOT_EQ_ALL:
			strcpy(b_op,"<>ALL");
			break;
		case O_NOT_EQ_ANY:
			strcpy(b_op,"<>ANY");
			break;
		case O_LESS:
			strcpy(b_op,"<");
			break;
		case O_LESS_ALL:
			strcpy(b_op,"<ALL");
			break;
		case O_LESS_ANY:
			strcpy(b_op,"<ANY");
			break;
		case O_MORE:
			strcpy(b_op,">");
			break;
		case O_MORE_ALL:
			strcpy(b_op,">ALL");
			break;
		case O_MORE_ANY:
			strcpy(b_op,">ANY");
			break;
		case O_LESS_EQ:
			strcpy(b_op,"<=");
			break;
		case O_LESS_EQ_ALL:
			strcpy(b_op,"<=ALL");
			break;
		case O_LESS_EQ_ANY:
			strcpy(b_op,"<=ANY");
			break;
		case O_MORE_EQ:
			strcpy(b_op,">=");
			break;
		case O_MORE_EQ_ALL:
			strcpy(b_op,">=ALL");
			break;
		case O_MORE_EQ_ANY:
			strcpy(b_op,">=ANY");
			break;
		case O_IS_NULL:
			strcpy(b_op,"IS NULL");
			break;
		case O_NOT_NULL:
			strcpy(b_op,"IS NOT NULL");
			break;
		case O_BETWEEN:
			strcpy(b_op,"BETWEEN");
			break;
		case O_IN:
			strcpy(b_op,"IN");
			break;
		case O_LIKE:
			strcpy(b_op,"LIKE");
			break;
		case O_ESCAPE:
			strcpy(b_op,"ESCAPE");
			break;
		case O_EXISTS:
			strcpy(b_op,"EXISTS");
			break;
		case O_NOT_LIKE:
			strcpy(b_op,"NOT LIKE");
			break;
		case O_NOT_BETWEEN:
			strcpy(b_op,"NOT BETWEEN");
			break;
		case O_NOT_IN:
			strcpy(b_op,"NOT IN");
			break;
		case O_NOT_EXISTS:
			strcpy(b_op,"NOT EXISTS");
			break;
		case O_FALSE:
			strcpy(b_op, "FALSE");
			break;
		case O_TRUE:
			strcpy(b_op,"TRUE");
			break;
		default:
			strcpy(b_op,"?");
	}

	char b_cop[20]; // enum ColOperation {LIST, COUNT, COUNT_DISTINCT, SUM, MIN, MAX, AVG};

	switch(cop) {
		case LISTING:
			strcpy(b_cop, "LIST");
			break;
		case COUNT:
			strcpy(b_cop, "COUNT");
			break;
		case SUM:
			strcpy(b_cop, "SUM");
			break;
		case MIN:
			strcpy(b_cop, "MIN");
			break;
		case MAX:
			strcpy(b_cop, "MAX");
			break;
		case AVG:
			strcpy(b_cop, "AVG");
			break;
		case STD_POP:
			strcpy(b_cop, "STD_POP");
			break;
		case STD_SAMP:
			strcpy(b_cop, "STD_SAMP");
			break;
		case VAR_POP:
			strcpy(b_cop, "VAR_POP");
			break;
		case VAR_SAMP:
			strcpy(b_cop, "VAR_SAMP");
			break;
		case BIT_AND:
			strcpy(b_cop, "BIT_AND");
			break;
		case BIT_OR:
			strcpy(b_cop, "BIT_OR");
			break;
		case BIT_XOR:
			strcpy(b_cop, "BIT_XOR");
			break;
		case GROUP_CONCAT:
			strcpy(b_cop, "GROUP_CONCAT");
			break;
		case GROUP_BY:
			strcpy(b_cop,"GROUP_BY");
			break;
		case DELAYED:
			strcpy(b_cop,"DELAYED");
			break;
		default:
			strcpy(b_cop,"[no name yet]");
	}

	char b_tmpar[20]; // enum TMParameter	{	TM_DISTINCT, TM_TOP, TM_EXISTS, TM_COUNT }; // Table Mode

	switch(tmpar) {
		case TM_DISTINCT:
			strcpy(b_tmpar,"DISTINCT");
			break;
		case TM_TOP:
			strcpy(b_tmpar,"LIMIT");
			break;
		case TM_EXISTS:
			strcpy(b_tmpar,"EXISTS");
			break;
		default:
			strcpy(b_tmpar,"???");
	}

	char b_jt[20]; // enum JoinType		{	JO_INNER, JO_LEFT, JO_RIGHT, JO_FULL };

	switch(jt) {
		case JO_INNER:
			strcpy(b_jt,"INNER");
			break;
		case JO_LEFT:
			strcpy(b_jt,"LEFT");
			break;
		case JO_RIGHT:
			strcpy(b_jt,"RIGHT");
			break;
		case JO_FULL:
			strcpy(b_jt,"FULL");
			break;
		default:
			strcpy(b_jt,"????");
	}

	switch(type) {
		case TABLE_ALIAS:
			if(alias)
				sprintf(buf,"T:%d = TABLE_ALIAS(T:%d,\"%s\")",N(t1.n),N(t2.n),alias);
			else
				sprintf(buf,"T:%d = TABLE_ALIAS(T:%d)",N(t1.n),N(t2.n));
			break;
		case CREATE_FILTER:
			sprintf(buf,"F:%d = CREATE_FILTER(T:%d,%s,%s,%s,%s,%s)",
				N(f1.n),
				N(t1.n),
				e1.ToString(b1, _countof(b1), t1.n),
				b_op,
				e2.ToString(b2, _countof(b2), t1.n),
				e3.ToString(b3, _countof(b3), t1.n),
				n1 == 0 ? "WHERE" : (n1 == 1 ? "HAVING" : (n1 == 2 ? "ON INNER" : (n1==3 ? "ON LEFT" : "ON RIGHT")))
				);
			break;
		case AND_F:
			sprintf(buf,"F:%d = AND(F:%d,F:%d)",N(f3.n),N(f1.n),N(f2.n));
			break;
		case OR_F:
			sprintf(buf,"F:%d = OR(F:%d,F:%d)",N(f3.n),N(f1.n),N(f2.n));
			break;
		case NOT_F:
			sprintf(buf,"F:%d = NOT(F:%d)",N(f2.n),N(f1.n));
			break;
		case COPY_F:
			sprintf(buf,"F:%d = COPY(F:%d)",N(f2.n),N(f1.n));
			break;
		case AND_E:
			sprintf(buf,"F:%d = AND(F:%d,%s,%s,%s,%s)",
				N(f2.n),
				N(f1.n),
				e1.ToString(b1, _countof(b1), t1.n),
				b_op,
				e2.ToString(b2, _countof(b2), t1.n),
				e3.ToString(b3, _countof(b3), t1.n));
			break;
		case TMP_TABLE:
			sprintf(buf,"T:%d = TMP_TABLE(T:%d)",N(t2.n),N(t1.n));
			break;
		case CREATE_VC:
			if(mysql_expr.size() == 1) {
				char s1[200];
				strncpy(s1, Query::GetItemName(mysql_expr[0]->GetItem()).c_str(), 199);
				s1[199] = '\0';
				sprintf(buf, "VC:%d.%d = CREATE_VC(T:%d,EXPR(\"%s\"))", N(t1.n), N(a1.n), N(t1.n), s1);
			} else if(mysql_expr.size() > 1) {
				char s1[200], s2[200];
				strncpy(s1, Query::GetItemName(mysql_expr[0]->GetItem()).c_str(), 199);
				strncpy(s2, Query::GetItemName(mysql_expr[mysql_expr.size() - 1]->GetItem()).c_str(), 199);
				s1[199] = '\0';
				s2[199] = '\0';
				sprintf(buf, "VC:%d.%d = CREATE_VC(T:%d,EXPR([%s,..,%s](%ld items)))", N(t1.n), N(a1.n), N(t1.n), s1, s2, mysql_expr.size());
			} else if(virt_cols.size() > 1) {
				sprintf(buf, "VC:%d.%d = CREATE_VC(T:%d,VIRT_COLS([%d,..,%d](%ld items)))", N(t1.n), N(a1.n), N(t1.n), virt_cols[0], virt_cols[virt_cols.size() - 1], virt_cols.size());
			} else if(a2.n != NULL_VALUE_32)
				sprintf(buf, "VC:%d.%d = CREATE_VC(T:%d,PHYS_COL(T:%d,A:%d))", N(t1.n), N(a1.n), N(t1.n), N(t2.n), N(a2.n));
			else
				sprintf(buf, "VC:%d.%d = CREATE_VC(T:%d,SUBQUERY(T:%d))", N(t1.n), N(a1.n), N(t1.n), N(t2.n));
			break;
		case T_MODE:
			sprintf(buf,"T:%d.MODE(%s,%lld,%lld)", N(t1.n), b_tmpar, n1, n2);
			break;
		case JOIN_T:
			sprintf(buf,"T:%d.JOIN(T:%d)",N(t1.n),N(t2.n));
			break;
		case DEFINE_PARAM:
			sprintf(buf,"P:%d = DEFINE_PARAM(%s)",p1.n,e1.ToString(b1, _countof(b1), 0));
			break;
		case FILTER_T:
			sprintf(buf,"T:%d.FILTER(F:%d)",N(t1.n),N(f1.n));
			break;
		case ADD_COLUMN:
			sprintf(buf,"A:%d = T:%d.ADD_COLUMN(%s,%s,\"%s\",\"%s\")", N(a1.n), N(t1.n),
					e1.ToString(b1, _countof(b1), t1.n), b_cop, (alias) ? alias : "null", n1 ? "DISTINCT" : "ALL");
			break;
		case ADD_ORDER:
			sprintf(buf, "T:%d.ADD_ORDER(VC:%d.%d,%s)", N(t1.n), N(t1.n), N(a1.n), n1 ? "DESC" : "ASC");
			break;
		case UNION:
			sprintf(buf, "T:%d = UNION(T:%d,T:%d,%lld)", N(t1.n), N(t2.n), N(t3.n), n1);
			break;
		case INTERSECT:
			sprintf(buf, "T:%d = INTERSECT(T:%d,T:%d,%lld)", N(t1.n), N(t2.n), N(t3.n), n1);
			break;
		case MINUS:
			sprintf(buf, "T:%d = MINUS(T:%d,T:%d,%lld)", N(t1.n), N(t2.n), N(t3.n), n1);
			break;
		case RESULT:
			sprintf(buf, "RESULT(T:%d)", N(t1.n));
			break;
		default:
			sprintf(buf, "Unsupported type of CQStep: %d", type);
	}
	if(!str || !size_left)
		rccontrol << buf << endl;
	else {
		string_add(str, size_left, "\n");
		string_add(str, size_left, buf);
	}
}

void CompiledQuery::TableAlias(TabID &t_out, TabID n, const char *name)
{
	CompiledQuery::CQStep s;
	s.type = TABLE_ALIAS;
	s.t1 = t_out=NextTabID();
	s.t2 = n;
	if(name) {
		s.alias = new char [strlen(name) + 1];
		strcpy(s.alias, name);
	}
	steps.push_back(s);
}

void CompiledQuery::TmpTable(TabID &t_out, TabID t1)
{
	CompiledQuery::CQStep s;
	assert(t1.n < 0 && NoTabs() > 0);
	s.type=TMP_TABLE;
	s.t2=t_out=NextTabID();
	s.t1=t1;
	steps.push_back(s);
}

void CompiledQuery::CreateFilter(FilterID &f_out, TabID t1, CQTerm e1, Operator pr, CQTerm e2, CQTerm e3, FilterType type)
{
	CompiledQuery::CQStep s;
	s.type = CREATE_FILTER;
	s.f1 = f_out = NextFilterID();
	s.t1 = t1;
	s.e1 = e1;
	s.op = pr;
	s.e2 = e2;
	s.e3 = e3;
	s.n1  = type;
	steps.push_back(s);
}

void CompiledQuery::And(FilterID &f_out, FilterID f1, FilterID f2, TabID t)
{
	if(f1.n == NULL_VALUE_32) {
		f_out = f2;
		return;
	}
	if(f2.n == NULL_VALUE_32) {
		f_out = f1;
		return;
	}
	CompiledQuery::CQStep s;
	s.t1 = t;
	s.type = AND_F;
	if(f_out.n != f1.n)
		s.f3 = f_out = NextFilterID();
	else
		s.f3 = f1;
	s.f1 = f1;
	s.f2 = f2;
	steps.push_back(s);
}

void CompiledQuery::Or(FilterID &f_out, FilterID f1, FilterID f2)
{
	if(f1.n == NULL_VALUE_32) {
		f_out = f2;
		return;
	}
	if(f2.n == NULL_VALUE_32) {
		f_out = f1;
		return;
	}
	CompiledQuery::CQStep s;
	s.type = OR_F;
	if(f_out.n != f1.n)
		s.f3 = f_out = NextFilterID();
	else
		s.f3 = f1;
	s.f1 = f1;
	s.f2 = f2;
	steps.push_back(s);
}

void CompiledQuery::Not(FilterID &f_out, FilterID f1)
{
	CompiledQuery::CQStep s;
	s.type=NOT_F;
	if(f_out.n!=f1.n) s.f2=f_out=NextFilterID();
	else s.f2=f1;
	s.f1=f1;
	steps.push_back(s);
}

void CompiledQuery::Copy(FilterID &f_out, FilterID f1)
{
	CompiledQuery::CQStep s;
	s.type=COPY_F;
	if(f_out.n!=f1.n) s.f2=f_out=NextFilterID();
	else s.f2=f1;
	s.f1=f1;
	steps.push_back(s);
}

void CompiledQuery::And(FilterID &f_out, FilterID f1, TabID t, CQTerm e1, Operator pr, CQTerm e2, CQTerm e3)
{
	CompiledQuery::CQStep s;
	s.type = AND_E;
	if(f_out.n != f1.n)
		s.f2 = f_out = NextFilterID();
	else
		s.f2 = f1;
	s.t1 = t;
	s.f1 = f1;
	s.e1 = e1;
	s.op = pr;
	s.e2 = e2;
	s.e3 = e3;
	steps.push_back(s);
}

void CompiledQuery::Mode(TabID t1, TMParameter mode, _int64 n1, _int64 n2)
{
	CompiledQuery::CQStep s;
	if(s.t1.n >= 0)
	{
		size_t const alias_ct(100);
		s.type = STEP_ERROR;
		s.alias = new char [alias_ct];
		strcpy(s.alias,  "T_MODE: can't be applied to RCTable");
		return;
	}
	s.type = T_MODE;
	s.t1 = t1;
	s.tmpar = mode;
	s.n1 = n1;
	s.n2 = n2;
	steps.push_back(s);
}

void CompiledQuery::Join(TabID t1, TabID t2)
{
	CompiledQuery::CQStep s;
	s.type = JOIN_T;
	s.t1 = t1;
	s.t2 = t2;
	steps.push_back(s);
}

void CompiledQuery::Define_Param(ParamID &p_out, CQTerm e1)
{
	CompiledQuery::CQStep s;
	s.type = DEFINE_PARAM;
	s.p1 = p_out = NextParamID();
	s.e1 = e1;
	steps.push_back(s);
}

void CompiledQuery::Filter(TabID t1, FilterID f1)
{
	CompiledQuery::CQStep s;
	s.type = FILTER_T;
	s.t1 = t1;
	s.f1 = f1;
	steps.push_back(s);
}

void CompiledQuery::AddColumn(AttrID &a_out, TabID t1, CQTerm e1, ColOperation op, char const alias[], bool distinct)
{
	assert(t1.n < 0 && NoTabs() > 0);
	CompiledQuery::CQStep s;
	s.type = ADD_COLUMN;
	s.a1 = a_out = NextAttrID(t1);
	s.t1 = t1;
	s.e1 = e1;
	s.cop = op;
	if(alias) {
		size_t const alias_ct(strlen(alias) + 1);
		s.alias = new char [alias_ct];
		strcpy(s.alias,  alias);
	} else
		s.alias = NULL;
	s.n1 = distinct ? 1 : 0;
	steps.push_back(s);
}


void CompiledQuery::CreateVirtualColumn(AttrID &a_out, TabID t1, MysqlExpression* expr, TabID src_tab)
{
	assert(t1.n < 0 && NoTabs() > 0);
	CompiledQuery::CQStep s;
	s.type = CREATE_VC;
	s.a1 = a_out = NextVCID(t1);
	s.t1 = t1;
	s.t2 = src_tab;
	s.mysql_expr.push_back(expr);
	steps.push_back(s);
}

void CompiledQuery::CreateVirtualColumn(AttrID &a_out, TabID t1, TabID subquery, bool on_result)
{
	assert(t1.n < 0 && NoTabs() > 0);
	CompiledQuery::CQStep s;
	s.type = CREATE_VC;
	s.a1 = a_out = NextVCID(t1);
	s.t1 = t1;
	s.t2 = subquery;
	s.n1 = on_result ? 1 : 0;
	steps.push_back(s);
}

void CompiledQuery::CreateVirtualColumn(AttrID &a_out, TabID t1, vector<int>& vcs, AttrID vc_id)
{
	assert(t1.n < 0 && NoTabs() > 0);
	CompiledQuery::CQStep s;
	s.type = CREATE_VC;
	s.a1 = a_out = NextVCID(t1);
	s.a2 = vc_id;
	s.t1 = t1;
	s.virt_cols = vcs;
	steps.push_back(s);
}

void CompiledQuery::CreateVirtualColumn(int &a_out, TabID t1, TabID table_alias, AttrID col_number)
{
	assert(t1.n < 0 && NoTabs() > 0);
	CompiledQuery::CQStep s;
	s.type = CREATE_VC;
	s.a1 = NextVCID(t1);
	a_out = s.a1.n;
	s.a2 = col_number;
	s.t1 = t1;
	s.t2 = table_alias;
	steps.push_back(s);
}

void CompiledQuery::Add_Order(TabID t1, AttrID vc, int d)			// d=1 for descending
{
	CompiledQuery::CQStep s;
	s.type = ADD_ORDER;
	s.t1 = t1;
	s.a1 = vc;
	s.n1 = d;
	steps.push_back(s);
}


void CompiledQuery::Union(TabID &t_out, TabID t2, TabID t3, int all)
{
	CompiledQuery::CQStep s;
	s.type = UNION;
	if(t_out.n != t2.n)
		s.t1 = t_out = NextTabID();
	else
		s.t1 = t2;
	s.t2 = t2;
	s.t3 = t3;
	s.n1 = all;
	steps.push_back(s);
}

void CompiledQuery::Intersect(TabID &t_out, TabID t2, TabID t3, int all)
{
	CompiledQuery::CQStep s;
	s.type = INTERSECT;
	s.t1 = t_out = NextTabID();
	s.t2 = t2;
	s.t3 = t3;
	s.n1 = all;
	steps.push_back(s);
}

void CompiledQuery::Result(TabID t1)
{
	CompiledQuery::CQStep s;
	s.type = RESULT;
	s.t1 = t1;
	steps.push_back(s);
}

void CompiledQuery::Print(char ** str, int * size_left)
{
	uint i;
	for(i = 0; i < steps.size(); i++)
		steps[i].Print(str, size_left);
}

bool CompiledQuery::CountColumnOnly(int table)
{
	CompiledQuery::CQStep step;
	bool count_only = false;
	for(int i = 0; i < NoSteps(); i++) {
		step = Step(i);
		if(step.type == CompiledQuery::ADD_COLUMN && step.t1.n == table && step.cop == COUNT && step.e1.IsNull())
			count_only = true;
		if(step.type == CompiledQuery::ADD_COLUMN && step.t1.n == table && (step.cop != COUNT || ( step.cop == COUNT && !step.e1.IsNull() )) ) {
			count_only = false;
			break;
		}
	}
	return count_only;
}

bool CompiledQuery::NoAggregationOrOrdering(int table)
{
	CompiledQuery::CQStep step;
	for(int i = 0; i < NoSteps(); i++) {
		step = Step(i);
		if(step.type == CompiledQuery::ADD_ORDER && step.t1.n == table)
			return false;			// exclude ordering
		if(step.type == CompiledQuery::ADD_COLUMN && step.t1.n == table && step.cop != LISTING)
			return false;			// exclude all kinds of aggregations
	}
	return true;
}

_int64 CompiledQuery::FindLimit(int table)
{
	CompiledQuery::CQStep step;
	for(int i = 0; i < NoSteps(); i++) {
		step = Step(i);
		if(step.type == CompiledQuery::T_MODE && step.t1.n == table && step.tmpar == TM_TOP)
			return step.n1 + step.n2;			// n1 - omitted, n2 - displayed,
												// i.e. either  ...LIMIT n2;   or  ...LIMIT n1, n2;
	}
	return -1;
}

set<int> CompiledQuery::GetUsedDims(int table_id, vector<JustATablePtr>& ta)
{
	set<int> result, local;
	CompiledQuery::CQStep step;
	for(int i = 0; i < NoSteps(); i++) {
		step = Step(i);
		if(step.type == CompiledQuery::ADD_COLUMN && step.t1.n == table_id && step.e1.vc_id != NULL_VALUE_32) {
			VirtualColumn* vc = ((TempTable*)ta[-table_id - 1].get())->GetVirtualColumn(step.e1.vc_id);
			if(vc) {
				local = vc->GetDimensions();
				result.insert(local.begin(), local.end());
			} else {
				// VC does not exist since it is created in latter compilation step
				// temporary solution is to skip optimization
				result.clear();
				result.insert(NULL_VALUE_32);
				return result;
			}
		}
	}
	return result;
}

bool CompiledQuery::IsSimpleFilter(int filter)
{
	CompiledQuery::CQStep step;
	bool is_simple = true;
	for(int i = 0; i < NoSteps(); i++) {
		step = Step(i);
		if(step.f1.n == filter && (step.type == CompiledQuery::AND_F || step.type == CompiledQuery::OR_F || step.type == CompiledQuery::NOT_F) ) {
			is_simple = false;
			break;
		} else if( step.f1.n == filter && ( // detection of condition with subqueries
			(step.e1.vc && dynamic_cast<MultiValColumn*>(step.e1.vc) != NULL) ||
			(step.e2.vc && dynamic_cast<MultiValColumn*>(step.e2.vc) != NULL) ||
			(step.e3.vc && dynamic_cast<MultiValColumn*>(step.e3.vc) != NULL)) ) {

			is_simple = false;
			break;
		}
	}
	return is_simple;
}

bool CompiledQuery::IsCopyOf(FilterID first, FilterID second)
{
	CompiledQuery::CQStep step;
	for(int i = 0; i < NoSteps(); i++) {
		step = Step(i);
		if(step.type == CompiledQuery::COPY_F && step.f2.n == first.n && step.f1.n == second.n)
			return true;
	}
	return false;
}

bool CompiledQuery::ExistsWhereFilterWithoutOr(int tab_id)
{
	CompiledQuery::CQStep step, step2;
	for(int i = 0; i < NoSteps(); i++) {
		step = Step(i);
		if(step.t1.n == tab_id && step.type == CompiledQuery::CREATE_FILTER && step.n1 == WHERE_FILTER) {
			FilterID f = step.f1;
			for(int j = i + 1; j < NoSteps(); j++) {
				step2 = Step(j);
				if(step2.type == OR_F && (step2.f1.n == f.n || IsCopyOf(step2.f1, f) || step2.f2.n == f.n || IsCopyOf(step2.f2, f)))
					return false;
			}
			return true;
		}
	}
	return false;
}

bool CompiledQuery::ExistsWhereFilter(int tab_id)
{
	CompiledQuery::CQStep step;
	for(int i = 0; i < NoSteps(); i++) {
		step = Step(i);
		if(step.t1.n == tab_id && step.type == CompiledQuery::CREATE_FILTER && step.n1 == WHERE_FILTER)
			return true;
	}
	return false;
}

bool CompiledQuery::IsGroupByQuery(TabID tab_id)
{
	CompiledQuery::CQStep step;
	for(int i = 0; i < NoSteps(); i++) {
		step = Step(i);
		if(step.type == CompiledQuery::ADD_COLUMN && step.t1 == tab_id && step.cop == GROUP_BY)
			return true;
	}
	return false;
}

bool CompiledQuery::ExistsInTempTable(TabID tab_id, TabID tmp_table)
{
	CompiledQuery::CQStep step;
	if (tab_id == tmp_table)
		return true;
	for(int i = 0; i < NoSteps(); i++) {
		step = Step(i);
		if( step.type == CompiledQuery::TMP_TABLE && step.t2 == tmp_table && step.t1 == tab_id ||
				step.type == CompiledQuery::JOIN_T && step.t1 == tmp_table && step.t2 == tab_id)
			return true;
	}
	return false;
}

TabID CompiledQuery::FindSourceOfParameter(TabID tab_id, TabID tmp_table, bool& is_group_by)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(tmp_table.n < 0);
	is_group_by = false;
	for(int x = tmp_table.n + 1; x < 0; x++) {
		if(IsTempTable(TabID(x)) && ExistsInTempTable(tab_id, TabID(x))) {
			is_group_by = IsGroupByQuery(TabID(x));
			return TabID(x);
		}
	}
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(!"Invalid parameter");
	return TabID(0);
}

vector<int> CompiledQuery::GetOnFilterIds(int tab_id)
{
	vector<int> result;
	CompiledQuery::CQStep step;
	for(int i = 0; i < NoSteps(); i++) {
		step = Step(i);
		if(step.t1.n == tab_id && step.type == CompiledQuery::CREATE_FILTER && (step.n1 == ON_LEFT_FILTER || step.n1 == ON_RIGHT_FILTER || step.n1 == ON_INNER_FILTER))
			if(IsFilterAppliedToTable(step.f1, step.t1))
				result.push_back(step.f1.n);
	}
	return result;
}

bool CompiledQuery::IsFilterAppliedToTable(FilterID f_id, TabID tab_id)
{
	CompiledQuery::CQStep step;
	for(int i = 0; i < NoSteps(); i++) {
		step = Step(i);
		if(step.type == CompiledQuery::FILTER_T && step.t1 == tab_id && step.f1.n == f_id.n)
			return true;
	}
	return false;
}

bool IsJoinOfTables(CompiledQuery::CQStep step, TabID tab_id, TabID joined_tab_id)
{
	return ((step.type == CompiledQuery::JOIN_T) && (step.t1 == tab_id) && (step.t2 == joined_tab_id));
}

void CompiledQuery::SetLOJOuterDim(TabID tab_id, TabID joined_tab_id, int joined_tab_dim)
{
	vector<CQStep>::iterator iter =
		find_if(steps.begin(), steps.end(), boost::bind(IsJoinOfTables, _1, tab_id, joined_tab_id));
	++iter;
	while(iter != steps.end()) {
		if((*iter).type == CREATE_VC || (*iter).type == OR_F) {
			++iter;
		} else if((*iter).type == CREATE_FILTER && ((*iter).n1 == ON_LEFT_FILTER || (*iter).n1 == ON_RIGHT_FILTER)) {
			iter->n2 = joined_tab_dim;
			++iter;
		} else if((*iter).type == AND_E) {
			iter->n2 = joined_tab_dim;
			++iter;
		} else
			break;
	}
}

bool CompiledQuery::IsTempTable(TabID t)
{
	for(int i = 0; i < NoSteps(); i++) {
		if(Step(i).type == CompiledQuery::TMP_TABLE && Step(i).t2 == t)
			return true;
	}
	return false;
}

