/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "Joiner.h"
#include "JoinerHash.h"
#include "JoinerSort.h"
#include "Query.h"
#include "vc/VirtualColumn.h"
#include "vc/ConstColumn.h"

using namespace std;

TwoDimensionalJoiner::TwoDimensionalJoiner(	MultiIndex *_mind, // multi-index to be updated
							RoughMultiIndex *_rmind,
							bool _count_only_now,
							_int64 _output_limit)
	:	m_conn(ConnectionInfoOnTLS.Get())
{
	mind = _mind;
	rmind = _rmind;
	count_only_now = _count_only_now;
	output_limit = _output_limit;
	why_failed = NOT_FAILED;
}

TwoDimensionalJoiner::~TwoDimensionalJoiner()
{
	// Note that mind and rmind are external pointers
}

JoinAlgType TwoDimensionalJoiner::ChooseJoinAlgorithm(MultiIndex& mind, Query* query, vector<Descriptor>& desc)
{
	JoinAlgType join_alg = JTYPE_GENERAL;
	// TODO: uncomment this section when general condition checking will be quick enough.
	// Right now, using hash join for one-row joining is still faster!
	// Test case on TPCH100: SELECT AVG(o_orderkey) FROM orders JOIN customer ON o_custkey = c_custkey WHERE c_name='Customer#001000000';
	/*
	if( desc[0].IsInner() && desc[0].IsType_JoinSimple() &&
		desc[0].val1.vc->GetDim() != -1 && mind.GetFilter(desc[0].val1.vc->GetDim()) &&
		mind.GetFilter(desc[0].val1.vc->GetDim())->NoOnes() == 1) {
		// Special case of single value
		MIIterator mit(&mind, desc[0].val1.vc->GetDim());
		VirtualColumn *prev_vc = desc[0].val1.vc;
		prev_vc->LockSourcePacks(mit);
    	desc[0].val1.vc = new ConstColumn(prev_vc->GetValue(mit), prev_vc->Type());
    	desc[0].val1.vc_id = query->AppendVC(desc[0].val1.vc);
    	return join_alg;	// i.e. the general one
	}
	*/
	if(desc[0].IsType_JoinSimple() && desc[0].op == O_EQ)
		join_alg = JTYPE_HASH;
	else  {
		if(desc[0].IsType_JoinSimple() &&
		    (desc[0].op == O_MORE_EQ || desc[0].op == O_MORE || desc[0].op == O_LESS_EQ || desc[0].op == O_LESS))
			join_alg = JTYPE_SORT;
	}
	return join_alg; 
}

JoinAlgType TwoDimensionalJoiner::ChooseJoinAlgorithm(JoinFailure join_result, size_t desc_size)
{
	// the easiest strategy: in case of any problems, use general joiner
	return JTYPE_GENERAL;
}

TwoDimsJoinerAutoPtr TwoDimensionalJoiner::CreateJoiner(JoinAlgType join_alg_type, MultiIndex& mind, RoughMultiIndex& rmind,
														bool count_only_now, _int64 limit_output, TempTable *table)
{
	switch( join_alg_type ) {
		case JTYPE_HASH:
			return TwoDimsJoinerAutoPtr(new JoinerHash(		&mind, &rmind, count_only_now, limit_output));
		case JTYPE_SORT:
			return TwoDimsJoinerAutoPtr(new JoinerSort(	 	&mind, &rmind, count_only_now, limit_output));
		case JTYPE_GENERAL:
			return TwoDimsJoinerAutoPtr(new JoinerGeneral(	&mind, &rmind, count_only_now, limit_output));
		default:
			BHERROR("Join algorithm not implemented");
	}
    return TwoDimsJoinerAutoPtr(0);
}
