/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef JOINER_H_
#define JOINER_H_

#include "Descriptor.h"
#include "MIIterator.h"

class RoughMultiIndex;
class TempTable;
class Query;

enum JoinAlgType { 	JTYPE_NONE,	JTYPE_SORT,	JTYPE_HASH,	JTYPE_MIXED,	JTYPE_GENERAL };
// MIXED   - for reporting: more than one algorithm was used

class TwoDimensionalJoiner		// abstract class for multiindex-based join algorithms
{
public:
	enum JoinFailure { NOT_FAILED, FAIL_COMPLEX, FAIL_SORTER_TOO_WIDE, FAIL_1N_TOO_HARD, FAIL_HASH };

	////////////
	TwoDimensionalJoiner(	MultiIndex *_mind, // multi-index to be updated
							RoughMultiIndex *_rmind,
							bool _count_only_now,
							_int64 _output_limit);
	virtual ~TwoDimensionalJoiner();

	virtual int ExecuteJoinConditions(std::vector<Descriptor> &desc) = 0;
								// join descriptors are concerned with the same pair of tables;
								// this method triggers the joining for the whole multiindex,
								// using as much descriptors from the list as possible
								// return: a number of descriptors processed

	JoinFailure WhyFailed()		{ return why_failed; }
								// the reason of the last failure of join operation
public:
	static JoinAlgType ChooseJoinAlgorithm(MultiIndex& mind, Query* query, std::vector<Descriptor>& desc);
	static JoinAlgType ChooseJoinAlgorithm(JoinFailure join_result, size_t desc_size);
	static std::auto_ptr<TwoDimensionalJoiner> CreateJoiner(JoinAlgType join_alg_type,
															MultiIndex& mind, RoughMultiIndex& rmind,
															bool count_only_now, _int64 limit_output, TempTable *table);
protected:
	MultiIndex *mind;
	RoughMultiIndex *rmind;
	bool count_only_now;
	_int64 output_limit;			// the joining may reduce output rows to output_limit; -1 means no limits
	ConnectionInfo& m_conn;

	int cond_processed;
	JoinFailure why_failed;
};

typedef std::auto_ptr<TwoDimensionalJoiner> TwoDimsJoinerAutoPtr;

/////////////////////////// Specializations //////////////////////////

class JoinerGeneral : public TwoDimensionalJoiner
{
/*
 * Algorithm: use MIUpdatingIterator to iterate through all dimensions involved
 * and delete tuples not matching a list of (arbitrary) conditions.
 */
public:
	JoinerGeneral( MultiIndex *_mind, RoughMultiIndex *_rmind,
			bool _count_only_now,
			_int64 _output_limit) : TwoDimensionalJoiner(_mind, _rmind, _count_only_now, _output_limit) {}
	~JoinerGeneral() {}

	int ExecuteJoinConditions(std::vector<Descriptor> &_desc);
};

#endif /*JOINER_H_*/
