/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "JustATable.h"
#ifndef __BH_COMMUNITY__
#	include "enterprise/local.h"
#else /* not __BH_COMMUNITY__ */
#	include "community/local.h"
#endif /* __BH_COMMUNITY__ */
#include "TempTable.h"
#include "Filter.h"
#include "CQTerm.h"
#include "common/bhassert.h"

using namespace std;

bool IsTimeStampZero(MYSQL_TIME& t);

bool JustATable::CheckJoinCondition(_int64 obj1,int attr1,JustATable &sec,_int64 obj2,int attr2,Operator op, ConnectionInfo *conn, _int64 factor1,_int64 factor2,_int64 vmax)	// WARNING: packs (if any) must be locked externally!
{
	// Assuming that the columns are compatible!
	if(AttrType(attr1)==RC_STRING || AttrType(attr1)==RC_VARCHAR)
	{
		if(IsNull(obj1,attr1) || sec.IsNull(obj2,attr2))
			return false;
		RCBString v1,v2;
		v1 = GetTable_S(obj1, attr1);
		v2 = sec.GetTable_S(obj2, attr2);
		switch(op)
		{
		case O_EQ:		if(v1==v2) return true;
			break;
		case O_LESS:	if(v1<v2) return true;
			break;
		case O_MORE:	if(v1>v2) return true;
			break;
		case O_LESS_EQ:	if(v1<=v2) return true;
			break;
		case O_MORE_EQ:	if(v1>=v2) return true;
			break;
		case O_NOT_EQ:	if(v1!=v2) return true;
			break;
		}
	}
	else if(ATI::IsRealType(AttrType(attr1)) || ATI::IsRealType(sec.AttrType(attr2)))
	{					// numericals incl. floating point, scale factor must be used
		if(IsNull(obj1,attr1) || sec.IsNull(obj2,attr2))
			return false;
		double v1,v2;
		if(ATI::IsRealType(AttrType(attr1)))
		{
			_int64 iv = GetTable64( obj1, attr1);
			v1 = *((double*)(&iv)) * factor1;
		}
		else
		{
			v1 = double(GetTable64( obj1, attr1)) * factor1;		// we don't need to check PLUS_INF etc., because double is larger than _int64
		}
		if(ATI::IsRealType(sec.AttrType(attr2)))
		{
			_int64 iv = sec.GetTable64( obj2, attr2);
			v2 = *((double*)(&iv)) * factor2;
		}
		else
		{
			v2 = double(sec.GetTable64( obj2, attr2)) * factor2;
		}
		switch(op)
		{
		case O_EQ:		if(v1==v2) return true;
			break;
		case O_LESS:	if(v1<v2) return true;
			break;
		case O_MORE:	if(v1>v2) return true;
			break;
		case O_LESS_EQ:	if(v1<=v2) return true;
			break;
		case O_MORE_EQ:	if(v1>=v2) return true;
			break;
		case O_NOT_EQ:	if(v1!=v2) return true;
			break;
		}
	}
	else				// numericals, scale factor must be used
	{
		_int64 v1,v2;
		v1 = GetTable64( obj1, attr1 );
		if(v1==NULL_VALUE_64) return false;
		v2 = sec.GetTable64( obj2, attr2 );
		if(v2==NULL_VALUE_64) return false;
		if(factor1>1)
		{
			if(v1 > vmax)		v1 = PLUS_INF_64;	// otherwise multiplication will be overflowed
			else if(v1 < -vmax)	v1 = MINUS_INF_64;
			else				v1 *= factor1;
		}
		if(factor2>1)
		{
			if(v2 > vmax)		v2 = PLUS_INF_64;	// otherwise multiplication will be overflowed
			else if(v2 < -vmax)	v2 = MINUS_INF_64;
			else				v2 *= factor2;
		}
		switch(op)
		{
		case O_EQ:		if(v1==v2) return true;
			break;
		case O_LESS:	if(v1<v2) return true;
			break;
		case O_MORE:	if(v1>v2) return true;
			break;
		case O_LESS_EQ:	if(v1<=v2) return true;
			break;
		case O_MORE_EQ:	if(v1>=v2) return true;
			break;
		case O_NOT_EQ:	if(v1!=v2) return true;
			break;
		}
	}
	return false;
}

int JustATable::SortPack(int pack, Filter *f, _int64 *index, int attr, ConnectionInfo *conn)
{ // WARNING: pack (for RSTable) must be locked externally!
	_int64 obj_to_sort = f->NoOnes(pack);
	if(obj_to_sort == 0)
		return 0;
	FilterOnesIterator it(f);
	it.Rewind(pack);
	for(_int64 i = 0; i < obj_to_sort; i++, ++it) {
		index[i] = *it;
	}
	if(obj_to_sort == 1)
		return 1;
	QuickSortPack(index, attr, 0, obj_to_sort-1, conn);
	return obj_to_sort;
}

void JustATable::QuickSortPack(_int64 *index, int attr, int s1, int s2, ConnectionInfo *conn)
{
	if(s2-s1==1)
	{
		if( Compare(attr, index[s1], index[s2],conn) == 1 )
		{
			_int64 ppp=index[s1];
			index[s1]=index[s2];
			index[s2]=ppp;
		}
	}
	else
	{
		int i = s1;		// QuickSort here
		int j = s2;
		_int64 mid = index[(s1+s2)/2];
		do
		{
			while(Compare(attr, index[i], mid,conn) < 0) i++;
			while(Compare(attr, index[j], mid,conn) > 0) j--;
			if(i<=j)
			{
				_int64 ppp=index[i];
				index[i]=index[j];
				index[j]=ppp;
				i++;
				j--;
			}
		}
		while(i<=j);
		if(s1<j) QuickSortPack(index,attr,s1,j,conn);
		if(i<s2) QuickSortPack(index,attr,i,s2,conn);
	}
}

int JustATable::Compare(int attr,_int64 obj1,_int64 obj2,ConnectionInfo *conn)
{
	bool n1 = IsNull(obj1,attr);
	bool n2 = IsNull(obj2,attr);
	if(n1 && n2) return 0;
	if(n1) return -1;
	if(n2) return 1;
	if(AttrType(attr)==RC_STRING || AttrType(attr)==RC_VARCHAR)
	{
		RCBString v1,v2;
		v1 = GetTable_S(obj1, attr);
		v2 = GetTable_S(obj2, attr);
		if(v1>v2) return 1;
		if(v1<v2) return -1;
	}
	else
	{
		_int64 v1 = GetTable64(obj1, attr);
		_int64 v2 = GetTable64(obj2, attr);
		if(v1>v2) return 1;
		if(v1<v2) return -1;
	}
	return 0;
}

ValueOrNull JustATable::GetComplexValue(const _uint64 obj, const int attr)
{
	if(obj==NULL_VALUE_64 || IsNull(obj, attr))
		return ValueOrNull();

	ColumnType ct = GetColumnType(attr);
	if(ct == RC_TIMESTAMP) {
		// needs to convert UTC/GMT time stored on server to time zone of client
		RCBString s = GetTable_S(obj, attr);
		MYSQL_TIME myt;
		int not_used;
		// convert UTC timestamp given in string into TIME structure
		str_to_datetime(s.Value(), s.len, &myt, TIME_DATETIME_ONLY, &not_used);
		if(!IsTimeStampZero(myt)) {
			// compute how many seconds since beg. of EPOCHE it is
			my_time_t secs = sec_since_epoch_TIME(&myt);
			// convert to local time in TIME structure
			ConnectionInfoOnTLS.Get().Thd().variables.time_zone->gmt_sec_to_TIME(&myt, secs);
		}
		return ValueOrNull(RCDateTime(myt.year, myt.month, myt.day, myt.hour, myt.minute, myt.second, RC_TIMESTAMP).GetInt64());
	}
	if(ct.IsFixed() || ct.IsFloat() || ct.IsDateTime())
		return ValueOrNull(GetTable64(obj, attr));
	if(ct.IsString()) {
		ValueOrNull val;
		RCBString s(GetTable_S(obj, attr, false));
		val.SetString(s);
		return val;
	}
	throw RCException("Unrecognized data type in JustATable::GetComplexValue");
}
