/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "MultiIndex.h"
#include "system/ConnectionInfo.h"

////////////////////////////////////////////////////////////////////////////////

MIRoughSorter::MIRoughSorter(MultiIndex* _m)
{
	mind = _m;
	rough_cols_to_sort = NULL;
}

MIRoughSorter::~MIRoughSorter()
{
	delete [] rough_cols_to_sort;
}

void MIRoughSorter::SortMultiIndexRoughly(_int64 limit)		// sort the tuples using objects' pack numbers as keys, locally in bigblocks
{
	if(mind->material_no_tuples < 2 || mind->index_type == mind->IT_COUNT)		// IT_COUNT
		return;

	no_rough_cols_to_sort = 0;
	_int64 bigblock_min_size = 100000;	// will be found below

	for(int i = 0; i < mind->no_dimensions; i++)
		if(mind->t[i] && mind->OrigSize(i) > 65536)			// sort only with more-than-one-pack dimensions
		{
			no_rough_cols_to_sort++;
			if(mind->t[i]->BlockShift()<bigblock_min_size)
				bigblock_min_size = mind->t[i]->BlockShift();
		}
	if(no_rough_cols_to_sort == 0)
		return;									// no materialized dimensions to sort
	bigblock_min_size = (_int64(1)<<bigblock_min_size);

	for(int i = 0; i < mind->no_dimensions; i++)
		if(mind->t[i])
			mind->t[i]->Lock();								// lock all materialized, even if not used in explicit sorting

	///////////////////////// start sorting from the largest table ////////////
	rough_cols_to_sort = new int [no_rough_cols_to_sort];
	int j = 0;
	for(int i = 0; i < mind->no_dimensions; i++)
		if(mind->t[i] && mind->OrigSize(i) > 65536)
			rough_cols_to_sort[j++] = i;
	for(int i = 0; i < no_rough_cols_to_sort - 1; i++)
		for(int j = i + 1; j < no_rough_cols_to_sort; j++)
			if(mind->OrigSize(rough_cols_to_sort[i]) < mind->OrigSize(rough_cols_to_sort[j]))
			{
				int p = rough_cols_to_sort[i];
				rough_cols_to_sort[i] = rough_cols_to_sort[j];
				rough_cols_to_sort[j] = p;
			}
	///////////////////////// check the actual need for sorting ///////////////
	bool sorting_needed = true;

/*	if(mind->temp_table_page_size[rough_cols_to_sort[0]] > 0) 	// TempTable: decides just a number of rows
		sorting_needed = SortingNeededTT(bigblock_min_size);
	else 													// RCTable: analyze open packs
	*/
	sorting_needed = SortingNeededRC(bigblock_min_size);

	///////////////////////// the main sorting loop through bigblocks /////////
	if(sorting_needed)	{
		rccontrol.lock(mind->m_conn->GetThreadID()) << "Sorting roughly multiindex..." << unlock;
		_int64 start_tuple = 0;
		_int64 stop_tuple = 0;
		_int64 bigblock_stop_tuple = -1;	// to ensure proper calculation of initial value
		rough_comp_bit_offset = 16;			// default - pack size
//		if(mind->temp_table_page_size[rough_cols_to_sort[0]] > 0)
//			rough_comp_bit_offset = CalculateBinSize(_uint64(mind->temp_table_page_size[rough_cols_to_sort[0]])) - 1;

		while(start_tuple < mind->material_no_tuples)		// sort with respect to bigblock boundaries
		{
			if(start_tuple > bigblock_stop_tuple) {
				bigblock_stop_tuple = start_tuple + bigblock_min_size - 1;
				if(bigblock_stop_tuple >= mind->material_no_tuples)
					bigblock_stop_tuple = mind->material_no_tuples - 1;
			}
			IndexTable *cur_t = mind->t[rough_cols_to_sort[0]];
			// define a block to be sorted: at least 80000 tuples (to balance between QSort advantages and use L2 efficiently)
			// and stop at the boundary of classes defined by the first dimension (typically it is the least changing dimension)
			stop_tuple = start_tuple + 80000;
			if(stop_tuple < bigblock_stop_tuple) {
				int pack_prev = ((cur_t->Get64(stop_tuple - 1) - 1) >> 16);
				while(stop_tuple <= bigblock_stop_tuple &&
					  pack_prev == ((cur_t->Get64(stop_tuple) - 1) >> 16)) {
					pack_prev = ((cur_t->Get64(stop_tuple) - 1) >> 16);
					stop_tuple++;
				}
				stop_tuple -= 1;	// now stop_tuple is the last tuple from the same data pack defined on the first dimension
			}
			else
				stop_tuple = bigblock_stop_tuple;
			RoughQSort(start_tuple, stop_tuple);
			start_tuple = stop_tuple + 1;

			if(limit != -1 && start_tuple > limit)			// don't sort other bigblocks
				break;
			if(mind->m_conn->killed())
				break;
		}
	}
	///////////////////////////////////////////////////////////////////////////
	for(int i = 0; i < mind->no_dimensions; i++)
		if(mind->t[i])
			mind->t[i]->Unlock();								// lock all materialized, even if not used in explicit sorting
	delete [] rough_cols_to_sort;
	rough_cols_to_sort = NULL;
}

bool MIRoughSorter::SortingNeededRC(_int64 bigblock_min_size)
{
//	return false;					// temporarily switched off
	bool sorting_needed = true;
	_int64 no_switches = 0;
	int cur_pack, prev_pack;
	_int64 stop_tuple = bigblock_min_size - 1;
	if(stop_tuple >= mind->material_no_tuples)
		stop_tuple = mind->material_no_tuples - 1;

	// STRATEGY 1: analyze a number of pack switches in just the largest table (incl. pack guardian)
/*
	const int	MAX_PACKS_TO_SORT  = 2000;		// we assume that this number of packs will match memory
	IndexTable *cur_t = t[rough_cols_to_sort[0]];
	int all_cur_packs = bh::common::NoObj2NoPacks(OrigSize(rough_cols_to_sort[0]));
	if( cur_t->Get64(0)==0 )		// i.e. null object
		prev_pack = 0;
	else
		prev_pack = (cur_t->Get64(0) - 1) >> 16;

	if( all_cur_packs < MAX_PACKS_TO_SORT )
		sorting_needed = false;
	if(sorting_needed) {
		int distinct_packs = 1;						// prev_pack is already counted
		bool *pack_found = new bool [all_cur_packs];
		for(int i = 0; i < all_cur_packs; i++)
			pack_found[i] = false;
		pack_found[prev_pack] = true;
		for(_int64 tuple = 1; tuple < stop_tuple; tuple++)	{
			if( cur_t->Get64(tuple)==0 )		// i.e. null object - no need to change pack
				cur_pack = prev_pack;
			else
				cur_pack = (cur_t->Get64(tuple) - 1) >> 16;
			if( cur_pack != prev_pack )
			{
				no_switches++;
				if( !pack_found[cur_pack] )
				{
					distinct_packs++;
					pack_found[cur_pack] = true;
				}
				prev_pack = cur_pack;
			}
		}
		no_switches -= distinct_packs;			// calculate unnecessary pack switches
		if( no_switches*3 < stop_tuple || distinct_packs < MAX_PACKS_TO_SORT )
			sorting_needed = false;		// do not sort if switches are rare (ave. size of packrow>3)
		delete [] pack_found;
	}
*/
	// STRATEGY 2: sort if on any level a pack is opened twice too often than really needed
	bool *pack_found = NULL;
	int pack_found_size = 0;
	_int64 switches_of_previous_dims = 0;
	sorting_needed = false;
	for(int sort_dim = 0; sort_dim < no_rough_cols_to_sort; sort_dim++) {
		IndexTable *cur_t = mind->t[rough_cols_to_sort[sort_dim]];
		int all_cur_packs = bh::common::NoObj2NoPacks(mind->OrigSize(rough_cols_to_sort[sort_dim]));
		_int64 tuple = 0;
		while(tuple < stop_tuple && cur_t->Get64(tuple)==0)		// i.e. null object
			tuple++;
		if(tuple == stop_tuple)
			continue;				// get the next dimension, this one is irrelevant
		prev_pack = (cur_t->Get64(tuple) - 1) >> 16;
		int distinct_packs = 1;						// prev_pack is already counted
		no_switches = 0;
		if(pack_found == NULL || pack_found_size < all_cur_packs) {
			delete [] pack_found;
			pack_found = new bool [all_cur_packs];
			pack_found_size = all_cur_packs;
		}
		for(int i = 0; i < all_cur_packs; i++)		// note: the table is recycled for next dimensions
			pack_found[i] = false;
		pack_found[prev_pack] = true;
		while(tuple < stop_tuple) {
			if(cur_t->Get64(tuple)==0)		// i.e. null object - no need to change pack
				cur_pack = prev_pack;
			else
				cur_pack = (cur_t->Get64(tuple) - 1) >> 16;
			if(cur_pack != prev_pack) {
				no_switches++;
				if( !pack_found[cur_pack] ) {
					distinct_packs++;
					pack_found[cur_pack] = true;
				}
				prev_pack = cur_pack;
			}
			tuple++;
		}
		if(no_switches > switches_of_previous_dims + distinct_packs*2) {
			sorting_needed = true;		// sort if too much switches
			break;
		}
		else
			switches_of_previous_dims += no_switches;
	}
	delete [] pack_found;
	return sorting_needed;
}
/*
bool MIRoughSorter::SortingNeededTT(_int64 bigblock_min_size)
{
	if(mind->OrigSize(rough_cols_to_sort[0]) <= mind->temp_table_page_size[rough_cols_to_sort[0]])
		return false;
	bool sorting_needed = true;
	_int64 no_switches = 0;
	rough_comp_bit_offset = CalculateBinSize(_uint64(mind->temp_table_page_size[rough_cols_to_sort[0]])) - 1;
	int cur_pack, prev_pack;
	_int64 stop_tuple = bigblock_min_size - 1;
	if(stop_tuple >= mind->material_no_tuples)
		stop_tuple = mind->material_no_tuples - 1;
	IndexTable *cur_t = mind->t[rough_cols_to_sort[0]];
	if( cur_t->Get64(0)==0 )		// i.e. null object
		prev_pack = 0;
	else
		prev_pack = (cur_t->Get64(0) - 1) >> rough_comp_bit_offset;
	_int64 limit_of_switches = 5 * (mind->OrigSize(rough_cols_to_sort[0])>>rough_comp_bit_offset);	// we allow at most 5 loads of the same block
	for(_int64 tuple = 1; tuple < stop_tuple; tuple++)
	{
		if( cur_t->Get64(tuple)==0 )		// i.e. null object - no need to change pack
			cur_pack = prev_pack;
		else
			cur_pack = (cur_t->Get64(tuple) - 1) >> rough_comp_bit_offset;
		if( cur_pack != prev_pack )
		{
			no_switches++;
			prev_pack = cur_pack;
			if(no_switches > limit_of_switches)
				break;
		}
	}
	if(no_switches <= limit_of_switches)	// all the sample is checked out and switches are not too frequent
		sorting_needed = false;
	return sorting_needed;
}
*/
void MIRoughSorter::SwitchMaterial(_uint64 tuple1, _uint64 tuple2)
{
	_uint64 val;
	IndexTable *ti;
	for(int d = 0; d < mind->no_dimensions; d++) {
		ti = mind->t[d];
		if(ti) {
			val = ti->Get64(tuple1);
			ti->Set64(tuple1, ti->Get64(tuple2));
			ti->Set64(tuple2, val);
		}
	}
}

void MIRoughSorter::RoughQSort(_int64 start_tuple, _int64 stop_tuple)
{
	if(stop_tuple - start_tuple < 20) {	// bubble sort for small buffers
		_int64 j = stop_tuple;
		_int64 pom;
		do {
			pom = -1;
			for(_int64 i = start_tuple; i < stop_tuple; i++)
				if(RoughCompare(i, i+1) > 0) {		// sorting includes stop_tuple
					SwitchMaterial(i, i+1);
					pom = i;
				}
			j = pom;
		} while(j != -1);
	}
	else
	{
		_int64 i = start_tuple;			// QuickSort here
		_int64 j = stop_tuple;
		_int64 mid = (i + j)/2;
		do {
			while(RoughCompare(i, mid) < 0)
				i++;
			while(RoughCompare(j, mid) > 0)
				j--;
			if(i <= j) {
				if(i != j) {
					SwitchMaterial(i, j);
					if(i == mid)
						mid = j;		// move the middle point, if needed
					else if(j == mid)
						mid = i;
				}
				i++;
				j--;
			}
		} while(i <= j);
		if(start_tuple < j)
			RoughQSort(start_tuple, j);
		if(i < stop_tuple)
			RoughQSort(i, stop_tuple);
	}
}

int MIRoughSorter::RoughCompare( _int64 tuple1, _int64 tuple2)
{
	for(int i = 0; i < no_rough_cols_to_sort; i++)
	{
		IndexTable *ti = mind->t[rough_cols_to_sort[i]];
		if( ((ti->Get64(tuple1)-1)>>rough_comp_bit_offset) < ((ti->Get64(tuple2)-1)>>rough_comp_bit_offset) ) return -1;
		if( ((ti->Get64(tuple1)-1)>>rough_comp_bit_offset) > ((ti->Get64(tuple2)-1)>>rough_comp_bit_offset) ) return 1;
	}
	return 0;
}
