/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "MIUpdatingIterator.h"

using namespace std;

MIUpdatingIterator::MIUpdatingIterator(MultiIndex* _mind, DimensionVector &dimensions)
	: MIIterator(_mind, dimensions), changed(false), one_dim(-1),
	one_dim_filter(NULL), multi_dim_filter(NULL),
	multi_filter_pos(0), multi_filter_pack_start(0)
{
	mind->IteratorUnlock();		// unlock a base class lock
	bool success = mind->IteratorUpdatingLock();
	assert(success);			// Multiindex was already locked for reading or updating!
	PrepareUpdater();
}

MIUpdatingIterator::MIUpdatingIterator(MultiIndex* _mind, std::set<int> dimensions)
	: MIIterator(_mind, dimensions), changed(false), one_dim(-1),
	one_dim_filter(NULL), multi_dim_filter(NULL),
	multi_filter_pos(0), multi_filter_pack_start(0)
{
	mind->IteratorUnlock();		// unlock a base class lock
	bool success = mind->IteratorUpdatingLock();
	assert(success);			// Multiindex was already locked for reading or updating!
	PrepareUpdater();
}

MIUpdatingIterator::MIUpdatingIterator(MultiIndex* _mind)
	: MIIterator(_mind), changed(false), one_dim(-1),
	one_dim_filter(NULL), multi_dim_filter(NULL),
	multi_filter_pos(0), multi_filter_pack_start(0)
{
	mind->IteratorUnlock();		// unlock a base class lock
	bool success = mind->IteratorUpdatingLock();
	assert(success);			// Multiindex was already locked for reading or updating!
	PrepareUpdater();
}

MIUpdatingIterator::MIUpdatingIterator(MultiIndex* _mind, int one_dimension)
	: MIIterator(_mind, one_dimension), changed(false), one_dim(-1),
	one_dim_filter(NULL), multi_dim_filter(NULL),
	multi_filter_pos(0), multi_filter_pack_start(0)
{
	mind->IteratorUnlock();		// unlock a base class lock
	bool success = mind->IteratorUpdatingLock();
	assert(success);			// Multiindex was already locked for reading or updating!
	PrepareUpdater();
}

MIUpdatingIterator::MIUpdatingIterator(const MIUpdatingIterator& sec)
	: MIIterator(sec), changed(false), one_dim(-1),
	one_dim_filter(NULL), multi_dim_filter(NULL),
	multi_filter_pos(0), multi_filter_pack_start(0)
{
	assert(0);			// HEAVY OBJECT - do not copy
}

MIUpdatingIterator::~MIUpdatingIterator()
{
	mind->IteratorUnlock();
	// one_dim_filter is external
	delete multi_dim_filter;
}

void MIUpdatingIterator::PrepareUpdater()
{
	assert(one_dim_filter == NULL && multi_dim_filter == NULL);

	if(dim_order.size() == 1 && dim_mode[dim_order[0]] == 1) {		// one-dim filter (special, limited case)
		one_dim = dim_order[0];
		one_dim_filter = mind->GetFilter(one_dim);
	} else {
		multi_dim_filter = new Filter(no_obj);
		multi_dim_filter->Set();
		multi_filter_pos = 0;
	}
}

void MIUpdatingIterator::ResetCurrent()
{
	changed = true;
	if(one_dim_filter)
		one_dim_filter->ResetDelayed(fi[one_dim]->GetCurrPack(), fi[one_dim]->GetCurrInPack());
	else {
		multi_dim_filter->ResetDelayed(multi_filter_pos);
	}
}

void MIUpdatingIterator::ResetCurrentPack()
{
	changed = true;
	if(one_dim_filter) {
		one_dim_filter->ResetBlock(fi[one_dim]->GetCurrPack());
	} else {
		multi_dim_filter->CommitPack(multi_filter_pack_start >> 16);
		multi_dim_filter->ResetBetween(multi_filter_pack_start, multi_filter_pos + pack_size_left - 1);
	}
}

void MIUpdatingIterator::Commit()
{
	if(!changed)
		return;
	if(one_dim_filter) {
		one_dim_filter->Commit();		// working directly on multiindex filer (special case)
		mind->UpdateNoTuples();
	} else {
		multi_dim_filter->Commit();
		mind->IteratorUnlock();			// unlock to allow creating the iterator below
		{	// create scope to ensure that mit_read will be unlocked on the end
			MIIterator mit_read(*this);		// make a traversing copy of the current iterator
			mind->MIFilterAnd(mit_read, *multi_dim_filter);		// UpdateNoTuples inside
		}
		mind->IteratorUpdatingLock();	// lock it again for consistency
	}
	changed = false;
}

void MIUpdatingIterator::Rewind()
{
	Commit();
	MIIterator::Rewind();
	multi_filter_pos = 0;
	multi_filter_pack_start = 0;
}

MIUpdatingIterator& MIUpdatingIterator::operator++()
{
	MIIterator::operator++();
	multi_filter_pos++;
	if(next_pack_started)
		multi_filter_pack_start = multi_filter_pos;
	return *this;
}

void MIUpdatingIterator::NextPackrow()
{
	multi_filter_pos += pack_size_left;
	MIIterator::NextPackrow();
}

_int64 MIUpdatingIterator::UpdatedNoTuples()
{
	Commit();
	_int64 res = Factor();
	if(one_dim_filter) {
		res *= one_dim_filter->NoOnes();
	} else {
		res *= multi_dim_filter->NoOnes();		// TODO: more accurate version for complex case (partial iterator)
	}
	return res;
}

