/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _CORE_PARAMZED_FILTER_H_
#define _CORE_PARAMZED_FILTER_H_

#include "MultiIndex.h"
#include "CQTerm.h"

#include <set>

class TempTable;
class DescTree;

/*
A class defining multidimensional filter (by means of MultiIndex) on a set of tables. It can store descriptors defining
some restrictions on particular dimensions. It can be parametrized thus the multiindex is not
materialized.
It can also store tree of conditions
*/

class ParameterizedFilter
{
public:
    ParameterizedFilter(FilterType filter_type = WHERE_FILTER);
    ParameterizedFilter(const ParameterizedFilter& );
    ParameterizedFilter(TempTable*, int alias, FilterType filter_type);
    virtual ~ParameterizedFilter(void);
    ParameterizedFilter & operator =(const ParameterizedFilter & pf);
    void AddDescriptor(CQTerm e1, Operator pr, CQTerm e2, CQTerm e3, TempTable* t, int outer_dim = -1, bool pAND = true);
    void Not();
    bool IsParametrized();
    bool IsLeftFilter() { return filter_type == ON_LEFT_FILTER; }
    bool IsRightFilter() { return filter_type == ON_RIGHT_FILTER; }
    bool IsOuterFilter() { return (IsLeftFilter() || IsRightFilter()); }
    bool IsOnFilter() { return (filter_type == ON_INNER_FILTER || filter_type == ON_LEFT_FILTER || filter_type == ON_RIGHT_FILTER); }
    bool IsWhereFilter() { return filter_type == WHERE_FILTER; }
    bool IsHavingFilter() { return filter_type == HAVING_FILTER; }
    uint NoParameters();
    uint NoUnsetParameters();
    bool IsParameterSet(uint i);
    CQTerm GetParameter(uint i);
    void ClearParamsSet();
    void ProcessParameters();
    TempTable* table;
    int table_alias;
    std::vector<Descriptor> descriptors;
    MultiIndex* m_index;
    std::vector<JoinType> join_types;
    std::vector<int> aliases;
    std::vector<Descriptor> parametrized_desc;
    std::vector<CQTerm> parameters;
    std::vector<int> params_set;
    std::vector<ParameterizedFilter*> filters;
    std::vector<LogicalOperator> log_connectors;
    DescTree* tree;
    FilterType filter_type;
private:
	bool ExtractParameters(CQTerm& e);
};
#endif //_CORE_PARAMZED_FILTER_H_

