/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef __BH_COMMUNITY__
#	include "enterprise/local.h"
#else /* not __BH_COMMUNITY__ */
#	include "community/local.h"
#endif /* __BH_COMMUNITY__ */
#include "Query.h"
#include "ParametrizedFilter.h"
#include "ValueSet.h"
#include "CompiledQuery.h"
#include "TempTable.h"
#include "MysqlExpression.h"
#include "vc/ExpressionColumn.h"
#include "vc/SingleColumn.h"
#include "vc/ConstColumn.h"
#include "vc/InSetColumn.h"
#include "vc/SubSelectColumn.h"

using namespace std;

#ifdef __GNUC__
#define _ASSERT(x) assert(x)
#endif

TempTable * Query::Execute(CompiledQuery &qu)
{
	TempTable *output_table = Preexecute(qu);

	// compute all the output data
	if(output_table)
		output_table->Materialize();
	return output_table;
}

TempTable * Query::Preexecute(CompiledQuery &qu, bool display_now)
{
	if(rccontrol.isOn() && display_now)	{
		rccontrol.lock(m_conn->GetThreadID());
		qu.Print();
		rccontrol << unlock;
	}
	bool force_mysql_execution = false;
	////////////// Set up the structures
	int no_params, no_filters; //no_virt_columns;
	int i;
	//no_taba = qu.NoTabs();		// number of table aliases in CompiledQuery q
   	no_filters = qu.NoFilters(); // number of filters in CompiledQuery q
	no_params = qu.NoParams();	// number of parameters in CompiledQuery q
	ParameterizedFilter **f = NULL;
	bool *count_only_filters = new bool[no_filters];
	if(no_filters > 0)
		f = new ParameterizedFilter * [no_filters];
	for(i = 0; i < no_filters; i++) {
		count_only_filters[i] = false;
		f[i] = NULL;
	}
	int no_f = 0;


	CQTerm **params = NULL;
	if(no_params > 0)
		params = new CQTerm * [no_params];
	for(i = 0; i < no_params; i++)
		params[i] = NULL;
	int no_p = 0;

//	if(no_virt_columns > 0)
//		virt_columns = new VirtualColumn * [no_virt_columns];
//	for(i = 0; i < no_virt_columns; i++) {
//		virt_columns[i] = NULL;
//	}

	TempTable *output_table = NULL;			// NOTE: this pointer will be returned by the function

	//if(no_taba>0)
	//	ta = new JustATable * [no_taba];
	//for(i=0;i<no_taba;i++)
	//	ta[i]=NULL;
	ta.resize(qu.NoTabs());
	int no_ta = 0;
	CompiledQuery::CQStep step;
	///////////// Preprocessing: special commands
	for(i = 0; i < qu.NoSteps(); i++) {
		// Local variables
		step = qu.Step(i);
		if(step.alias && strcmp(step.alias, "bh_force_mysql") == 0)	// magical word (passed as table alias) to force MySQL execution
		{
			force_mysql_execution = true;
			break;
		}
	}

	///////////// Execution itself
	for(i = 0; i < qu.NoSteps(); i++) {
		// Local variables
		step = qu.Step(i);
		JustATablePtr t1_ptr, t2_ptr, t3_ptr;

		if(step.t1.n != NULL_VALUE_32) {
			if(step.t1.n>=0)
				t1_ptr=Table(step.t1.n)->shared_from_this();		// normal table
			else {
//				_ASSERT(ta);
				t1_ptr=ta[-step.t1.n-1];	// TempTable
			}
		}
		if(step.t2.n != NULL_VALUE_32) {
			if(step.t2.n>=0)
				t2_ptr=Table(step.t2.n)->shared_from_this();		// normal table
			else {
//				_ASSERT(ta);
				t2_ptr=ta[-step.t2.n-1];	// TempTable
			}
		}
		if(step.t3.n != NULL_VALUE_32) {
			if(step.t3.n>=0)
				t3_ptr=Table(step.t3.n)->shared_from_this();		// normal table
			else {
//				_ASSERT(ta);
				t3_ptr=ta[-step.t3.n-1];	// TempTable
			}
		}
		//////////////////// Some technical information ///////////////////
		if(step.alias && strcmp(step.alias, "roughstats") == 0)	{
			// magical word (passed as table alias) to display statistics
			((TempTable*)ta[-step.t1.n-1].get())->DisplayRSI();
		}

		if(step.alias && strcmp(step.alias, "roughattrstats") == 0) {
			// magical word (passed as table alias) to display attr. statistics
			m_conn->SetDisplayAttrStats();
		}

		////////////////// Implementation of steps ////////////////////////
		try	{
			if( force_mysql_execution )
				throw NotImplementedRCException("MySQL execution forced by user.");

			switch(step.type) {
				case CompiledQuery::TABLE_ALIAS:
					ta[-step.t1.n-1] = t2_ptr;
					no_ta++;
					break;
				case CompiledQuery::TMP_TABLE:
					assert(step.t1.n < 0);
					ta[-step.t2.n-1] = TempTable::Create(ta[-step.t1.n-1].get(), step.t1.n, this);
					((TempTable*)ta[-step.t2.n - 1].get())->ReserveVirtColumns(qu.NoVirtualColumns(step.t2));
					//((TempTable*)ta[-step.t2.n-1])->SetConnectionInfo(m_conn);
					no_ta++;
					break;
				case CompiledQuery::CREATE_FILTER:
			 		assert(step.t1.n < 0);
					step.e1.vc = (step.e1.vc_id != NULL_VALUE_32) ? ((TempTable*)ta[-step.t1.n - 1].get())->GetVirtualColumn(step.e1.vc_id) : NULL;
					step.e2.vc = (step.e2.vc_id != NULL_VALUE_32) ? ((TempTable*)ta[-step.t1.n - 1].get())->GetVirtualColumn(step.e2.vc_id) : NULL;
					step.e3.vc = (step.e3.vc_id != NULL_VALUE_32) ? ((TempTable*)ta[-step.t1.n - 1].get())->GetVirtualColumn(step.e3.vc_id) : NULL;
					if(step.n1 != HAVING_FILTER) {// on result = false
						int outer_dim = step.n2 == NULL_VALUE_64 ? -1 : (int)step.n2;
						f[step.f1.n] = new ParameterizedFilter((TempTable *)ta[-step.t1.n-1].get(), step.t1.n, (FilterType)step.n1);
						if((FilterType)step.n1 == WHERE_FILTER && qu.GetOnFilterIds(step.t1.n).size() > 0)
							outer_dim = -2;
						f[step.f1.n]->AddDescriptor(step.e1, step.op, step.e2, step.e3, (TempTable *)ta[-step.t1.n-1].get(), outer_dim);
					} else { // on result = true
						f[step.f1.n] = new ParameterizedFilter((FilterType)step.n1);
						f[step.f1.n]->tree = new DescTree(step.e1, step.op, step.e2, step.e3, (TempTable *)ta[-step.t1.n-1].get());
					}
					no_f++;
					break;
				case CompiledQuery::AND_F:
				case CompiledQuery::OR_F:
					_ASSERT(f);
					if(!f[step.f1.n]->tree) { // on result = false
						//bool last_and_or = IsLastAndOrOperation(f[step.f1.n]->table_alias, step);
						if((f[step.f1.n]->IsOuterFilter() || f[step.f2.n]->IsOuterFilter()) && step.type == CompiledQuery::OR_F)
							throw NotImplementedRCException("OR of outer join conditions not implemented.");
						if(f[step.f1.n]->descriptors.size() != 0) {
							// optimization step: detecting unused dimensions
							if(f[step.f1.n]->aliases.size() > 1) {
								for(uint i = 0; i < f[step.f1.n]->aliases.size(); i++) {
									f[step.f1.n]->m_index->SetUsedInOutput(i);
								}
							}
							this->UpdateParamFilter(f[step.f1.n], false, -1, display_now);
							f[step.f1.n]->descriptors.clear();
						}
						if(f[step.f2.n]->descriptors.size() != 0) {
							// optimization step: detecting unused dimensions
							if(f[step.f2.n]->aliases.size() > 1) {
								for(uint i = 0; i < f[step.f2.n]->aliases.size(); i++) {
									f[step.f2.n]->m_index->SetUsedInOutput(i);
								}
							}
							this->UpdateParamFilter(f[step.f2.n], false, -1, display_now);
							f[step.f2.n]->descriptors.clear();
						}
						if(step.f3.n != step.f1.n) {
							f[step.f3.n] = new ParameterizedFilter(*f[step.f1.n]);
							no_f++;
						}

						if(f[step.f3.n]->IsParametrized() || f[step.f2.n]->IsParametrized()) {
							f[step.f3.n]->filters.push_back(new ParameterizedFilter(*f[step.f2.n]));
							if(step.type == CompiledQuery::AND_F)
								f[step.f3.n]->log_connectors.push_back(O_AND);
							else
								f[step.f3.n]->log_connectors.push_back(O_OR);
						} else if(step.type == CompiledQuery::AND_F)
							f[step.f3.n]->m_index->And(*f[step.f2.n]->m_index);
						else
							f[step.f3.n]->m_index->Or(*f[step.f2.n]->m_index);
					} else {// on result = true
						if(step.f3.n != step.f1.n) {
							f[step.f3.n] = new ParameterizedFilter(*f[step.f1.n]);
							no_f++;
						}
						if(step.type == CompiledQuery::AND_F)
							f[step.f3.n]->tree->AddTree(O_AND, f[step.f2.n]->tree);
						else
							f[step.f3.n]->tree->AddTree(O_OR, f[step.f2.n]->tree);
					}
					break;
				case CompiledQuery::NOT_F:
					throw NotImplementedRCException("NOT is not implemented in this case.");
					if(!f[step.f1.n]->tree) { // on result = false
						if(f[step.f1.n]->descriptors.size() != 0) {
							// optimization step: detecting unused dimensions
							for(uint i = 0; i < f[step.f1.n]->aliases.size(); i++)
								f[step.f1.n]->m_index->SetUsedInOutput(i);
							this->UpdateParamFilter(f[step.f1.n], false, -1, display_now);
							f[step.f1.n]->descriptors.clear();
						}
						if(step.f2.n != step.f1.n) {
							f[step.f2.n] = new ParameterizedFilter(*f[step.f1.n]);
							no_f++;
						}
						f[step.f2.n]->Not();
					} else {
						// TODO: Not on result
						throw NotImplementedRCException("NOT is not implemented in this case.");
					}
					break;
				case CompiledQuery::COPY_F:
					if(step.f2.n != step.f1.n) {
						f[step.f2.n] = new ParameterizedFilter(*f[step.f1.n]);
						no_f++;
					}
					break;
				case CompiledQuery::AND_E: {
					step.e1.vc = (step.e1.vc_id != NULL_VALUE_32) ? ((TempTable*)ta[-step.t1.n - 1].get())->GetVirtualColumn(step.e1.vc_id) : NULL;
					step.e2.vc = (step.e2.vc_id != NULL_VALUE_32) ? ((TempTable*)ta[-step.t1.n - 1].get())->GetVirtualColumn(step.e2.vc_id) : NULL;
					step.e3.vc = (step.e3.vc_id != NULL_VALUE_32) ? ((TempTable*)ta[-step.t1.n - 1].get())->GetVirtualColumn(step.e3.vc_id) : NULL;
					if(step.f2.n != step.f1.n) {
						f[step.f2.n] = new ParameterizedFilter(*f[step.f1.n]);
						no_f++;
					}
					if(!f[step.f1.n]->tree) { // on result = false
						BHASSERT_WITH_NO_PERFORMANCE_IMPACT(f && f[step.f2.n]);
						int outer_dim = step.n2 == NULL_VALUE_64 ? -1 : (int)step.n2;
						if(f[step.f2.n]->IsWhereFilter() && qu.GetOnFilterIds(step.t1.n).size() > 0)
							outer_dim = -2;
						f[step.f2.n]->AddDescriptor(step.e1, step.op, step.e2, step.e3, (TempTable *)ta[-step.t1.n-1].get(), outer_dim);
					} else  // on result = true
						f[step.f2.n]->tree->AddDescriptor(O_AND, step.e1, step.op, step.e2, step.e3, (TempTable *)ta[-step.t1.n-1].get());
					break;
				}
				case CompiledQuery::T_MODE:
					assert(step.t1.n<0 && ta[-step.t1.n-1]->TableType() == TEMP_TABLE);
					((TempTable *)ta[-step.t1.n-1].get())->SetMode(step.tmpar, step.n1, step.n2);
					break;
				case CompiledQuery::JOIN_T:
					assert(step.t1.n<0 && ta[-step.t1.n-1]->TableType() == TEMP_TABLE);
					((TempTable*)ta[-step.t1.n-1].get())->JoinT(t2_ptr.get(), step.t2.n, step.jt);
					break;
				case CompiledQuery::DEFINE_PARAM:
					params[step.p1.n] = new CQTerm(step.e1);
					no_p++;
					break;
				case CompiledQuery::FILTER_T: {
					assert(step.t1.n<0 && ta[-step.t1.n-1]->TableType() == TEMP_TABLE);
					if(step.f1.n == NULL_VALUE_32)
						break;
					if(f[step.f1.n]->IsOnFilter() && qu.ExistsWhereFilterWithoutOr(step.t1.n))
						break;
					vector<int> on_filter_ids = qu.GetOnFilterIds(step.t1.n);
					int no_on_filters = on_filter_ids.size();
					int last_on_filter_id = no_on_filters ? on_filter_ids[no_on_filters - 1] : -1;
					if(f[step.f1.n]->IsOnFilter() && step.f1.n != last_on_filter_id) // skip all but last ON filter
						break;
					if(f[step.f1.n]->IsWhereFilter() || f[step.f1.n]->IsOnFilter()) { // last ON filter or WHERE filter
						vector<Descriptor> new_descs;
						for(int x = 0; x < no_on_filters; x++) {
							int on_filter_id = on_filter_ids[x];
							if(on_filter_id == step.f1.n)
								continue;
							for(uint i = 0; i < f[on_filter_id]->descriptors.size(); i++)
								new_descs.push_back(f[on_filter_id]->descriptors[i]);
							f[on_filter_id]->descriptors.clear();
						}
						for(uint i = 0; i < f[step.f1.n]->descriptors.size(); i++)
							new_descs.push_back(f[step.f1.n]->descriptors[i]);
						f[step.f1.n]->descriptors.swap(new_descs);
					}
					if(f[step.f1.n]->descriptors.size() != 0) {
						// optimization step: detect count only
						if( qu.CountColumnOnly(step.t1.n) && !f[step.f1.n]->IsParametrized() &&
							qu.IsSimpleFilter(step.f1.n))
							count_only_filters[step.f1.n] = true;
						// optimization step: detecting unused dimensions
						if(f[step.f1.n]->aliases.size() > 1 && !f[step.f1.n]->IsParametrized()) {
							set<int> used_dims = qu.GetUsedDims(f[step.f1.n]->table_alias, ta);
							bool is_simple_filter = qu.IsSimpleFilter(step.f1.n);
							if(used_dims.size() == 1 && used_dims.find(NULL_VALUE_32) != used_dims.end()) 
								is_simple_filter = false;
							for(uint i = 0; i < f[step.f1.n]->aliases.size(); i++) {
								if(used_dims.find(i) == used_dims.end() && is_simple_filter /*&& false*/)
									f[step.f1.n]->m_index->ResetUsedInOutput(i);
								else
									f[step.f1.n]->m_index->SetUsedInOutput(i);
							}
						}
						// optimization step: detect limit without ordering/aggregating
						_int64 cur_limit = -1;
						if(qu.NoAggregationOrOrdering(step.t1.n))
							cur_limit = qu.FindLimit(step.t1.n);
						bool count_only = (qu.ExistsWhereFilterWithoutOr(step.t1.n) || !qu.ExistsWhereFilter(step.t1.n)) ? 
							count_only_filters[step.f1.n] : 
							false;
						if(cur_limit != -1 && f[step.f1.n]->parametrized_desc.size() > 0)
							cur_limit = -1;
						this->UpdateParamFilter(f[step.f1.n], count_only, cur_limit, display_now);
						f[step.f1.n]->descriptors.clear();
					}
					((TempTable*)ta[-step.t1.n-1].get())->Filter(f[step.f1.n]);
					break;
				}
				case CompiledQuery::ADD_COLUMN: {
					assert(step.t1.n<0 && ta[-step.t1.n-1]->TableType() == TEMP_TABLE);
					CQTerm e(step.e1);
					if (e.vc_id != NULL_VALUE_32)
						e.vc = ((TempTable*)ta[-step.t1.n - 1].get())->GetVirtualColumn(step.e1.vc_id); //vc must have been created
					//if(step.e1.tab_id.n < 0 && step.e1.tab_id.n != NULL_VALUE_32)
					//	e.tab_p = ta[-step.e1.tab_id.n - 1]; // needs to get pointer to the source table
					//// e must contain pointer to JustATable, because aliases are useless within TempTable
					step.a1.n = ((TempTable*)ta[-step.t1.n-1].get())->AddColumn(e, step.cop, step.alias, step.n1 ? true : false);
					break;
				}
				case CompiledQuery::CREATE_VC: {
					assert(step.t1.n < 0 && ta[-step.t1.n - 1]->TableType() == TEMP_TABLE);
					TempTable* t = (TempTable*)ta[-step.t1.n - 1].get();

					assert(t);
					if(step.mysql_expr.size() > 0) {
					// VirtualColumn for Expression
						assert(step.mysql_expr.size() == 1);
						MultiIndex *mind = (step.t2.n == step.t1.n) ? t->GetOutputMultiIndexP() : t->GetMultiIndexP();
						int c= ((TempTable*)ta[-step.t1.n - 1].get())->AddVirtColumn(CreateColumnFromExpression(step.mysql_expr, t, step.t1.n, mind), step.a1.n);
						assert(c == step.a1.n);
					} else if(step.virt_cols.size() > 0) {
					// VirtualColumn for IN
						ColumnType ct;
						if(step.a2.n != NULL_VALUE_32)
							ct = ((TempTable*)ta[-step.t1.n - 1].get())->GetVirtualColumn(step.a2.n)->Type();
						vector<VirtualColumn*> vcs;
						for(uint i = 0; i < step.virt_cols.size(); i++)
							vcs.push_back(((TempTable*)ta[-step.t1.n - 1].get())->GetVirtualColumn(step.virt_cols[i]));
						int c= ((TempTable*)ta[-step.t1.n - 1].get())->AddVirtColumn( new InSetColumn(ct, t->GetMultiIndexP(), vcs), step.a1.n);
						assert(c == step.a1.n);
					} else if(step.a2.n != NULL_VALUE_32) {
					// VirtualColumn for PhysicalColumn
						JustATable* t_src = ta[-step.t2.n - 1].get();
						PhysicalColumn* phc;
						MultiIndex *mind = (step.t2.n == step.t1.n) ? t->GetOutputMultiIndexP() : t->GetMultiIndexP();
						int dim = (step.t2.n == step.t1.n) ? 0 : t->GetDimension(step.t2.n);
						phc = (PhysicalColumn* ) t_src->GetColumn(step.a2.n >= 0 ? step.a2.n : -step.a2.n - 1);
						int c= ((TempTable*)ta[-step.t1.n - 1].get())->AddVirtColumn(  new SingleColumn(phc, mind, step.t2.n, step.a2.n, ta[-step.t2.n - 1].get(), dim), step.a1.n);
						assert(c == step.a1.n);
					} else {
					// VirtualColumn for Subquery
						assert(ta[-step.t2.n - 1]->TableType() == TEMP_TABLE);
						int c= ((TempTable*)ta[-step.t1.n - 1].get())->AddVirtColumn(
									new SubSelectColumn(dynamic_cast<TempTable*>(ta[-step.t2.n - 1].get()),
										step.n1 == 1 ? t->GetOutputMultiIndexP() : t->GetMultiIndexP(),
										t, step.t1.n),
								step.a1.n);
						assert(c == step.a1.n);
					}
					break;
				}
				case CompiledQuery::ADD_ORDER: {
						assert(step.t1.n < 0 && ta[-step.t1.n-1]->TableType() == TEMP_TABLE && step.n1 >= 0 && step.n1 < 2);
						assert(step.a1.n >= 0 && step.a1.n < qu.NoVirtualColumns(step.t1));
						TempTable *loc_t = (TempTable *)ta[-step.t1.n - 1].get();
						loc_t->AddOrder(loc_t->GetVirtualColumn(step.a1.n), (int)step.n1); // step.n1 = 0 for asc, 1 for desc
						break;
					}
				case CompiledQuery::UNION:
					assert(step.t1.n<0 && step.t2.n<0 && step.t3.n<0);
					assert(ta[-step.t2.n-1]->TableType() == TEMP_TABLE && ta[-step.t3.n-1]->TableType() == TEMP_TABLE);
					if(step.t1.n!=step.t2.n)
						ta[-step.t1.n-1] = TempTable::Create(*(TempTable *)ta[-step.t2.n-1].get());
					//((TempTable*)ta[-step.t1.n-1])->SetConnectionInfo(m_conn);
					((TempTable *)ta[-step.t1.n-1].get())->Union((TempTable *)ta[-step.t3.n-1].get(), (int)step.n1);
					break;
				case CompiledQuery::RESULT:
					assert(step.t1.n<0 && -step.t1.n-1<ta.size() && ta[-step.t1.n-1]->TableType() == TEMP_TABLE);
					output_table=(TempTable *)ta[-step.t1.n-1].get();
					break;
				case CompiledQuery::STEP_ERROR:
					rccontrol.lock(m_conn->GetThreadID()) << "ERROR in step " << step.alias << unlock;
					break;
				default:
					rccontrol.lock(m_conn->GetThreadID()) << "ERROR: unsupported type of CQStep (" << step.type << ")" << unlock;
			}
		} catch(...) {
			for(i = 0; i < no_filters; i++)
				delete f[i];
			delete [] f;
			delete [] count_only_filters;
			for(i = 0; i < no_params; i++)
				delete params[i];
			delete [] params;

			throw;
		}
	}

	///////////// Cleanup

	for(i = 0; i < no_filters; i++)
		delete f[i];
	delete [] f;
	delete [] count_only_filters;

	for(i = 0; i < no_params; i++)
		delete params[i];
	delete [] params;

	// NOTE: output_table is sent out of this function and sholud be managed elsewhere.
	// before deleting all TempTables but output_table those have to be detected
	// there are used by output_table

	//output_table->SetManageUsedTables();

	return output_table;
}

void swap(CompiledQuery::CQStep& a, CompiledQuery::CQStep& b) {
	a.swap(b);
}
