/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "types/RCItemTypes.h"
#include "shared/core/RCEngine.h"

#ifdef USE_NEW_ENGINE
	#include "engine/DataSource.h"
#endif

using namespace std;

namespace
{
void scan_fields(List<Item> &fields, uint*& buf_lens, map<int,Item*> &items_backup);
void restore_fields(List<Item> &fields, map<int,Item*> &items_backup);
void SetTableNullRow(TABLE* table);
void UnsetTableNullRow(TABLE* table);
bool IsFieldNullable(Field* field);
void SetFieldState(Field* field, bool isNull);
}

#ifdef USE_NEW_ENGINE
void RCEngine::SendResults(na::DataSource* exectree, THD* thd, select_result *res, List<Item> &fields)
{
	int error = 0;
	thd->proc_info="Sending data";
	DBUG_PRINT("info", ("%s", thd->proc_info));

	res->send_fields(fields, Protocol::SEND_NUM_ROWS | Protocol::SEND_EOF);
	thd->lex->unit.offset_limit_cnt = 0;

	Item* item;
	Field* f;
	Item_field* ifield;
	Item::Type itemType;
	List_iterator_fast<Item> li(fields);
	int is_null;
	uint col_id  = 0;

	uint* buf_lens = NULL;
	std::map<int,Item*> items_backup;
	scan_fields(fields, buf_lens, items_backup);

	vector<AValue*> bufs;
	exectree->GetBuffers(bufs);
	vector<AValue*>::iterator iterBufs;

	#ifdef LIMIT_NO_RECORDS
		uint64 no_ret_rec = 0;
	#endif

	exectree->Open();
	try {
		while(exectree->GetNext())
		{
			iterBufs = bufs.begin();
			while ((item=li++))
			{
				is_null = 0;

				itemType = item->type();
				switch(itemType)
				{
				case Item::DEFAULT_VALUE_ITEM:
				case Item::FIELD_ITEM:
					{
						ifield = (Item_field*)item;
						f = ifield->result_field;
						if(buf_lens[col_id] != 0)  // if buf_lens[col_id] is 0 means that f->ptr was not assigned
							// because it was assigned for this instance of object
						{
							RCEngine::Convert(is_null, f, f->ptr, *((AValue*)*iterBufs));
							SetFieldState(f, is_null);
						}
						break;
					}
				case Item::REF_ITEM:
					{

					}
				case Item::SUM_FUNC_ITEM:
					{

					}
				}
				col_id++;
				iterBufs++;
			}


			#ifdef LIMIT_NO_RECORDS
				if(no_ret_rec++ < RETURNED_RECORDS_LIMIT)
					error = res->send_data(fields);
			#else
				error = res->send_data(fields);
			#endif
			if (error)
			{
			}

			col_id = 0;
			li.rewind();
		}
	}
	catch(...) {
		exectree->Close();
		delete [] buf_lens;
		throw;
	}

	exectree->Close();

	delete [] buf_lens;
	restore_fields(fields, items_backup);

	res->send_eof();

	return;
}
#endif

void RCEngine::SendResults(JustATable& results, THD* thd, select_result *res, List<Item> &fields, ConnectionInfo *conn)
{
	int error = 0;
	thd->proc_info="Sending data";
    DBUG_PRINT("info", ("%s", thd->proc_info));

	res->send_fields(fields, Protocol::SEND_NUM_ROWS | Protocol::SEND_EOF);
	thd->lex->unit.offset_limit_cnt = 0;

	Item* item;
	Field* f;
	Item_field* ifield;
	Item_ref* iref;
	Item_sum* is;

	Item_sum::Sumfunctype sumType;

	List_iterator_fast<Item> li(fields);
	bool nothing_to_send = false;
	_uint64 no_objects = results.NoObj();
	uint col_id  = 0;
	_int64 value;
	int is_null;

	if(no_objects == 0)
		nothing_to_send = true;

	std::map<int,Item*> items_backup;
	uint* buf_lens = NULL;
	scan_fields(fields, buf_lens, items_backup);

	Item_sum_int_rcbase*		isum_int_rcbase;
	Item_sum_sum_rcbase*		isum_sum_rcbase;
	Item_sum_hybrid_rcbase*	isum_hybrid_rcbase;

	#ifdef LIMIT_NO_RECORDS
		uint64  no_ret_rec = 0;
	#endif

	if(!nothing_to_send) {
		int size = 0;
		//fprintf(stderr, "Argument 'fields' of SendResults():\n");
		while (item=li++) {
			//printItemTree(item);
			++size;
		}
		JustATable::RcTableLockPolicy::attr_uses_t attrUses( size );
		for ( int i = 0; i < size ; ++i )
			attrUses[ i ] = i;
		TempTable::RecordIterator iter = results.begin( conn, attrUses );

		for( ; iter != results.end(conn); ++iter)	{
			// Before proceeding, undo potential usage of SetTableNullRow() inside SetFieldState()
			// All this is needed to ensure null values (caused by outer join) which appear in not null columns
			col_id = 0;
			li.rewind();
			while(item=li++) {
				if(buf_lens[col_id] != 0) {
					RCItem objItem = (*iter)[col_id];
					switch(item->type()) {
						case Item::DEFAULT_VALUE_ITEM:
						case Item::FIELD_ITEM:
								ifield = (Item_field*)item;
								break;
						case Item::REF_ITEM:
								iref = (Item_ref*)item;
								ifield = (Item_field*)(*iref->ref);
								break;
					}
					UnsetTableNullRow(ifield->result_field->table);
				}
				col_id++;
			}

			col_id = 0;
			li.rewind();
			while(item=li++) {
				is_null = 0;
				RCItem objItem = (*iter)[col_id];
				switch(item->type()) {
					case Item::DEFAULT_VALUE_ITEM:
					case Item::FIELD_ITEM:  // regular select
						ifield = (Item_field*)item;
						f = ifield->result_field;
						// if buf_lens[col_id] is 0 means that f->ptr was not assigned
						// because it was assigned for this instance of object
						if(buf_lens[col_id] != 0) {
							RCEngine::Convert(is_null, f, f->ptr, &objItem);
							SetFieldState(f, is_null);
						}
						break;
					case Item::REF_ITEM:	// select from view
						iref = (Item_ref*)item;
						ifield = (Item_field*)(*iref->ref);
						f = ifield->result_field;
						if(buf_lens[col_id] != 0) {
							RCEngine::Convert(is_null, f, f->ptr, &objItem);
							SetFieldState(f, is_null);
						}
						break;
					case Item::SUM_FUNC_ITEM:
						is = (Item_sum*)item;
						sumType = is->sum_func();

						//used only MIN_FUNC
						if(sumType == Item_sum::MIN_FUNC || sumType == Item_sum::MAX_FUNC) {
							isum_hybrid_rcbase = (Item_sum_hybrid_rcbase*)is;
							if(isum_hybrid_rcbase->result_type() == DECIMAL_RESULT)	{
								RCEngine::Convert(is_null, isum_hybrid_rcbase->dec_value(), &objItem);
								isum_hybrid_rcbase->null_value = is_null;
							}	else if(isum_hybrid_rcbase->result_type() == INT_RESULT) {
								RCEngine::Convert(is_null, isum_hybrid_rcbase->int64_value(), &objItem);
								isum_hybrid_rcbase->null_value = is_null;
							}	else if(isum_hybrid_rcbase->result_type() == REAL_RESULT)	{
								RCEngine::Convert(is_null, isum_hybrid_rcbase->real_value(), &objItem);
								isum_hybrid_rcbase->null_value = is_null;
							}	else if(isum_hybrid_rcbase->result_type() == STRING_RESULT)	{
								RCEngine::Convert(is_null, isum_hybrid_rcbase->string_value(), &objItem, isum_hybrid_rcbase->hybrid_field_type);
								isum_hybrid_rcbase->null_value = is_null;
							}
							break;
						}
						//do not check COUNT_DISTINCT_FUNC, we use only this for both types
						if(sumType == Item_sum::COUNT_FUNC || sumType == Item_sum::SUM_BIT_FUNC) {
							isum_int_rcbase = (Item_sum_int_rcbase*)is;
							RCEngine::Convert(is_null, value, &objItem);
							if(is_null)
								value = 0;
							isum_int_rcbase->int64_value(value);
							break;
						}
						if(sumType == Item_sum::SUM_FUNC || sumType == Item_sum::STD_FUNC || sumType == Item_sum::VARIANCE_FUNC) {
							isum_sum_rcbase = (Item_sum_sum_rcbase*)is;
							if(isum_sum_rcbase->result_type() == DECIMAL_RESULT) {
								RCEngine::Convert(is_null, isum_sum_rcbase->dec_value(), &objItem);
								isum_sum_rcbase->null_value = is_null;
							}	else if(isum_sum_rcbase->result_type() == REAL_RESULT) {
								RCEngine::Convert(is_null, isum_sum_rcbase->real_value(), &objItem);
								isum_sum_rcbase->null_value = is_null;
							}
							break;
						}
				}
				col_id++;
			}

			#ifdef LIMIT_NO_RECORDS
				if(no_ret_rec++ < RETURNED_RECORDS_LIMIT)
					error = res->send_data(fields);
			#else
				error = res->send_data(fields);
			#endif

		}
	}

	delete [] buf_lens;
	restore_fields(fields, items_backup);
	res->send_eof();
	return;
}

namespace {
void scan_fields(List<Item> &fields, uint* &buf_lens, map<int,Item*> &items_backup)
{
	Item* item;
	Field* f;
	Item_field* ifield;
	Item_ref* iref;
	Item_sum* is;

	Item::Type itemType;
	Item_sum::Sumfunctype sumType;
	List_iterator<Item> li(fields);

	buf_lens = new uint[fields.elements];
	uint item_id = 0;
	uint field_length = 0;
	uint total_length = 0;

	// we don't need to check each type f->pack>length() returns correct value

	// check fileds
	stdext::hash_set<size_t> used_fields;
	stdext::hash_set<size_t>::iterator iter_uf;

	Item_sum_int_rcbase* isum_int_rcbase;
	Item_sum_sum_rcbase* isum_sum_rcbase;
	Item_sum_hybrid_rcbase* isum_hybrid_rcbase;

	while ((item=li++))
	{
		itemType = item->type();
		buf_lens[item_id] = 0;
		switch(itemType)
		{

		case Item::FIELD_ITEM:  // regular select
			{
				ifield = (Item_field*)item;
				f = ifield->result_field;
				iter_uf = used_fields.find((size_t)f);
				if(iter_uf == used_fields.end())
				{
					used_fields.insert((size_t)f);

					field_length = f->pack_length();
					buf_lens[item_id] = field_length;
					total_length += field_length;
				}
				else
				{
					buf_lens[item_id] = 0;
				}

				break;
			}
		case Item::REF_ITEM:	// select from view
			{
				Item_sum_hybrid_rcbase *tmp= new Item_sum_hybrid_rcbase();
				items_backup[item_id]=item;



				tmp->decimals    = item->decimals;
				tmp->hybrid_type = item->result_type();
				tmp->unsigned_flag = item->unsigned_flag;
				tmp->hybrid_field_type = item->field_type();


				li.replace(tmp);

				break;
			}
		case Item::FUNC_ITEM:
			{
				Item_sum_hybrid_rcbase *tmp= new Item_sum_hybrid_rcbase();
				items_backup[item_id]=item;



				tmp->decimals    = item->decimals;
				tmp->hybrid_type = item->result_type();
				tmp->unsigned_flag = item->unsigned_flag;
				tmp->hybrid_field_type = item->field_type();


				li.replace(tmp);

				break;
			}
		case Item::SUBSELECT_ITEM:
			{
				Item_sum_hybrid_rcbase *tmp= new Item_sum_hybrid_rcbase();
				items_backup[item_id]=item;



				tmp->decimals    = item->decimals;
				tmp->hybrid_type = item->result_type();
				tmp->unsigned_flag = item->unsigned_flag;
				tmp->hybrid_field_type = item->field_type();


				li.replace(tmp);

				break;
			}
		case Item::SUM_FUNC_ITEM:
			{
				is = (Item_sum*)item;
				sumType = is->sum_func();

				if(sumType == Item_sum::MIN_FUNC || sumType == Item_sum::MAX_FUNC)
				{
					isum_hybrid_rcbase = new Item_sum_hybrid_rcbase();
					items_backup[item_id]=item;

					// We have to add some knowledge to our item from original item

					isum_hybrid_rcbase->decimals    = is->decimals;
					isum_hybrid_rcbase->hybrid_type = is->result_type();
					isum_hybrid_rcbase->unsigned_flag = is->unsigned_flag;
					isum_hybrid_rcbase->hybrid_field_type = is->field_type();
					isum_hybrid_rcbase->collation.set(is->collation);
					isum_hybrid_rcbase->value.set_charset(is->collation.collation);

					li.replace(isum_hybrid_rcbase);
					buf_lens[item_id] = 0;
					break;
				}
				else if(sumType == Item_sum::COUNT_FUNC || sumType == Item_sum::COUNT_DISTINCT_FUNC
						|| sumType == Item_sum::SUM_BIT_FUNC)
				{
					isum_int_rcbase = new Item_sum_int_rcbase();
					isum_int_rcbase->unsigned_flag = is->unsigned_flag;
					items_backup[item_id] = item;
					li.replace(isum_int_rcbase);

					buf_lens[item_id] = 0;
					break;
				}
				// we can use the same type for SUM,AVG and SUM DIST, AVG DIST
				else if(sumType == Item_sum::SUM_FUNC || sumType == Item_sum::AVG_FUNC
						|| sumType == Item_sum::SUM_DISTINCT_FUNC || sumType == Item_sum::AVG_DISTINCT_FUNC
						|| sumType == Item_sum::VARIANCE_FUNC || sumType == Item_sum::STD_FUNC)
				{
					isum_sum_rcbase = new Item_sum_sum_rcbase();
					items_backup[item_id]=item;

					// We have to add some knowledge to our item from original item

					isum_sum_rcbase->decimals    = is->decimals;
					isum_sum_rcbase->hybrid_type = is->result_type();
					isum_sum_rcbase->unsigned_flag = is->unsigned_flag;

					li.replace(isum_sum_rcbase);

					buf_lens[item_id] = 0;
					break;
				}
				else
				{
					buf_lens[item_id] = 0;
					break;
				}
			}
		}
		item_id++;
	}

	item_id = 0;
	li.rewind();

	while ((item=li++))
	{
		itemType = item->type();
		switch(itemType)
		{
		case Item::FIELD_ITEM:  // regular select
			{
				ifield = (Item_field*)item;
				f = ifield->result_field;

				break;
			}
		case Item::REF_ITEM:	// select from view
			{
				iref = (Item_ref*)item;
				ifield = (Item_field*)(*iref->ref);
				f = ifield->result_field;

				if(buf_lens[item_id] != 0)
				{
				//	f->ptr = (char*) n_buffer;
				//	n_buffer += buf_lens[item_id];
				}

				break;
			}
		}
		item_id++;
	}
}

void restore_fields(List<Item> &fields, map<int,Item*> &items_backup)
{
	//nothing to restore
	if(items_backup.size() == 0) {
		return;
	}

	Item* item;
	List_iterator<Item> li(fields);

	int count = 0;
	while((item=li++))	{
		if(items_backup.find(count) != items_backup.end())
			li.replace(items_backup[count]);
		count++;
	}
}

inline void SetTableNullRow(TABLE* table)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(table);
	table->null_row = 1;
}

inline void UnsetTableNullRow(TABLE* table)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(table);
	table->null_row = 0;
}

inline bool IsFieldNullable(Field* field)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(field);
#ifdef __WIN__
	return (field->null_ptr != 0);  // WINPORT - explicit bool expression
#else
	return field->null_ptr;
#endif
}

inline void SetFieldState(Field* field, bool isNull)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(field);
	if(isNull) {
		if(IsFieldNullable(field))
			field->set_null();
		else		// not nullable, but null needed because of outer join
			SetTableNullRow(field->table);
	}	else {
		field->set_notnull();
	}
}
}

