/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef CORE_RCSYNCHRO_H_INCLUDED
#define CORE_RCSYNCHRO_H_INCLUDED

#include <vector>
#include <map>

#include "system/RCException.h"

//  Interprocess synchronization object
class RCSynchronization
{
public:
	static const unsigned MAX_NUMBER_OF_TABLES = 16384;

	RCSynchronization(void):m_hSharedMutex(NULL),m_hSharedMem(NULL),m_pSharedBuf(NULL) { };
	~RCSynchronization(void);

	static bool DatabaseInUse(const char* path) throw (DatabaseRCException);

	void Init(unsigned database_id) throw (SystemRCException);
	int DatabaseCurrent() const;
	int DatabaseNewState();
	int UsersCurrent() const;
	int UsersNewState();

	int RegisterTables(int* tabids, unsigned no_tabs);	// return schema initialized state
	int RegisterTable(int tabid);				// return new schema state
	int TableNewState(int tabid) throw (InternalRCException);
	int TableCurrent(int tabid) throw (InternalRCException);

	void EnterTable(int tabid, bool modify_mode = false) throw (SystemRCException,InternalRCException);
	bool TryEnterTable(int tabid) throw (SystemRCException);
	void LeaveTable(int tabid, bool modify_mode = false) throw (SystemRCException,InternalRCException);
	void LockTableModify(int tabid) throw (SystemRCException);
	bool TryLockTableModify(int tabid) throw (SystemRCException);
	void UnlockTableModify(int tabid) throw (SystemRCException,InternalRCException);
	void LockTableExclusive(int tabid) throw (SystemRCException,InternalRCException);
	void UnlockTableExclusive(int tabid) throw (SystemRCException,InternalRCException);
	void WaitForTableDrop(int tabid) throw (SystemRCException);
	bool CanDropTable(int tabid) throw (SystemRCException);
	void LockTableDrop(int tabid) throw (SystemRCException);
	void UnregisterTable(int tabid) throw (SystemRCException,InternalRCException);

private:
	struct TableSynchr
	{
		HANDLE m_AccessMutex;
		HANDLE m_ModifyMutex;
		HANDLE m_FreeEvent;
		HANDLE m_DropEvent;
		unsigned m_nModifyLocks;
		bool m_bExclLock;
	};
	static const DWORD ERROR_SIZE = 512;
	static const TCHAR SHARED_MUTEX[];
	static const TCHAR SHARED_BUFFER[];
	static const unsigned SHARED_SIZE = MAX_NUMBER_OF_TABLES*12+12;

	static const TCHAR TABLE_ACCESS_MUTEX_PREFIX[];
	static const TCHAR TABLE_MODIFY_MUTEX_PREFIX[];
	static const TCHAR TABLE_FREE_EVENT_PREFIX[];
	static const TCHAR TABLE_DROP_EVENT_PREFIX[];

	void Enter() throw (SystemRCException);
	void Leave() throw (SystemRCException);

	TableSynchr& GetSynchr(int tabid) throw (SystemRCException);

	unsigned m_nDatabaseId;
	char m_ErrorBuf[ERROR_SIZE];
	HANDLE m_hSharedMutex;
	HANDLE m_hSharedMem;
	int* m_pSharedBuf;
	std::map<int,TableSynchr*> m_TableHandles;
};

#endif  // not CORE_RCSYNCHRO_H_INCLUDED


