/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _CORE_RSIPACK_H_
#define _CORE_RSIPACK_H_

#include "common/CommonDefinitions.h"
#include "RSI_Framework.h"
#include "bintools.h"

/////////////////////////////////////////////////////////////////////////////
// RSIndex_PP    - pack-pack indexes for two attributes
//

// Usage:
//
//	RSIndex_Hist h;
//	h.Load(fhandle);
//	if(h.NoObj()<current_no_obj) then only a part of it may be safely used, i.e.
//		only between 0 and (h.NoObj()>>16 - 1) packs are up to date.
//
//	result = h.IsValue( value,value,pack,min_pack,max_pack);	// check if a value (or interval) is present in a pack
//
// Updating:
//
//	h.ClearPack(pack);		// clear all packs which are changed
//	h.Update(new_no_obj);	// register new number of objects
//	for(all changed packs)
//		h.PutValue(value,pack,min_pack,max_pack);
//	h.Save(fhandle);
//
class RSIndex_PP : public RSIndex
{
public:
	RSIndex_PP();				// create an empty index
	~RSIndex_PP();

	void Clear();				// make the index empty

	_int64 NoObj1()		{ return no_obj1; }
	_int64 NoObj2()		{ return no_obj2; }

	// creation, update, use

	void Create(_int64 _no_obj1, _int64 _no_obj2);		// create a new PackPack based on _no_obj1 and _no_obj2 objects
	void Verify(_int64 _no_obj1, _int64 _no_obj2);		// set the proper value for updated flags

	 bool GetValue(int pack1, int pack2)			// true: the packs may have nonempty intersection
	{
		if(pack1<no_pack1 && pack2<no_pack2 &&
			u->Get(pack1,pack2))
		{
			return t->Get(pack1,pack2);
		}
		return true;				// i.e. intersection possible, must be checked manually
	}

	 void PutValue(int pack1, int pack2, bool v)
	{
		t->Set(pack1, pack2, v);
		u->Set(pack1, pack2, true);
	}

	 void SetUpToDate(int pack1, int pack2, bool v = true)
	{
		u->Set(pack1, pack2, v);
	}

	 bool IsUpToDate(int pack1, int pack2)			// true: the information about this packs is precise and up to date
	{
		if(pack1<no_pack1 && pack2<no_pack2)
		{
			return u->Get(pack1, pack2);
		}
		return false;				// i.e. intersection possible, must be checked manually
	}
	double Density(int pack2=-1);	// a relative number of ones (i.e. pairs of intersecting packs), between 0.0 and 1.0
	double UpdateDensity() 			{ return u->Density(); }

	// Loading/saving

	int Load(IBFile* frs_index, int unused);					// fd - file descriptor, open for reading
	int Save(IBFile* frs_index, int unused);					// fd - file descriptor, open for writing
	int LoadLastOnly(IBFile* frs_index, int unused);
	void Update(_int64 _new_no_obj1, _int64 _new_no_obj2);
	bool AllUpdated(BinMatrix &not_touched);	// true if there is no more chance for updating
	bool AllUpdated();	// true if there is no more chance for updating at all
	bool AllUpdated(_int64 _new_no_obj1, _int64 _new_no_obj2)	// true if there is no more chance for updating at all
	{ return ( _new_no_obj1 == no_obj1 && _new_no_obj2 == no_obj2 && AllUpdated() ); }

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// For testing:
	void SaveAsText(char *fname);
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
private:
	_int64	no_obj1;		// timeliness signature: a number of objects in table 1 for which the index is up-to-date
	_int64	no_obj2;		// timeliness signature: a number of objects in table 2 for which the index is up-to-date
	BinMatrix *t;			// matrix of values
	BinMatrix *u;			// matrix of values

	// derived:
	int		no_pack1;		// number of packs in table 1
	int		no_pack2;		// number of packs in table 2

	const static uchar current_verion = 0;
};

#endif  //_CORE_RSIPACK_H_

