/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef __BH_COMMUNITY__
#	include "enterprise/local.h"
#else /* not __BH_COMMUNITY__ */
#	include "community/local.h"
#endif /* __BH_COMMUNITY__ */
#include "RoughJoinWatcher.h"
#include "RSI_PackPack.h"
#include "system/ConnectionInfo.h"
#include "MIIterator.h"

using namespace std;

RoughJoinWatcher::RoughJoinWatcher( MultiIndex &mind, vector<Descriptor> &desc )
{
	rs1 = NULL;
	rs2 = NULL;
	dim1 = -1;
	dim2 = -1;
	attr1 = -1;
	attr2 = -1;
	t1 = NULL;
	t2 = NULL;
	inequality_join = false;

	can_update_p2p = CheckIfUpdatePossible( desc );

	if( can_update_p2p )
		CreateRoughImage( mind );
}

RoughJoinWatcher::~RoughJoinWatcher()
{
	delete [] rs1;
	delete [] rs2;
}

bool RoughJoinWatcher::CheckIfUpdatePossible( vector<Descriptor> &desc )
{
	/////////////////////// Temporarily switch off: /////////////
	return false;
	/////////////////////// TODO: implement for VCs! ////////////////////

	//if( desc.size()>1 )
	//	return false;

	//JustATable *jt1 = desc[0].attr.tab_p;
	//JustATable *jt2 = desc[0].val1.tab_p;
	//if( !jt1 || jt1->TableType() == TEMP_TABLE || !jt2 || jt2->TableType() == TEMP_TABLE )
	//	return false;

	//if( desc[0].op != O_EQ && desc[0].op != O_MORE_EQ && desc[0].op != O_LESS_EQ )
	//	return false;
	//inequality_join = ( desc[0].op == O_MORE_EQ || desc[0].op == O_LESS_EQ );

	//dim1 = desc[0].attr.dim;
	//dim2 = desc[0].val1.dim;
	//attr1 = desc[0].attr.attr_id.n;
	//attr2 = desc[0].val1.attr_id.n;
	//t1 = (RCTable*)jt1;
	//t2 = (RCTable*)jt2;
 //   uint sess_id_t1 = t1->GetSaveSessionId();
 //   uint sess_id_t2 = t2->GetSaveSessionId();
 //   if ((sess_id_t1 != 0xFFFFFFFF) || (sess_id_t2 != 0xFFFFFFFF))
 //       return false;
	//if( t1->GetID() > t2->GetID() || ( t1->GetID() == t2->GetID() && attr1 > attr2) )	// switch order
	//{
	//	int p;					// numbers of tables/attributes sent to RSIndex_PP must be properly ordered
	//	p = attr1;
	//	attr1 = attr2;
	//	attr2 = p;
	//	p = dim1;
	//	dim1 = dim2;
	//	dim2 = p;
	//	RCTable *pt;
	//	pt = t1;
	//	t1 = t2;
	//	t2 = pt;
	//}

	//if( rsi_manager == NULL || t1->NoPacks() == 0 || t2->NoPacks() == 0 )
	//	return false;

	//bool can_be_updated = true;
	//RSIndex_PP *rsi_pp = NULL;
	//rsi_pp = (RSIndex_PP *)(rsi_manager->GetIndexForUpdate(RSIndexID(RSI_JPP, t1->GetID(), attr1, t2->GetID(), attr2),0));
	//// for now - just check it out
	//if( rsi_pp == NULL || rsi_pp->AllUpdated( t1->NoObj(), t2->NoObj() ) ) 	// cannot open or up to date
	//	can_be_updated = false;
	//if(rsi_pp) rsi_manager->UpdateIndex(rsi_pp,0,true);		// unlock without save - just for now
	//return can_be_updated;
}

void RoughJoinWatcher::CreateRoughImage(MultiIndex &mind)
{
	int no_packs1 = bh::common::NoObj2NoPacks(mind.OrigSize(dim1));
	int no_packs2 = bh::common::NoObj2NoPacks(mind.OrigSize(dim2));
	BHASSERT(no_packs1 > 0 && no_packs2 > 0, "no_packs == 0");
	rs1 = new RSValue[no_packs1];
	rs2 = new RSValue[no_packs2];
	DimensionImage(rs1, mind, dim1, no_packs1);
	DimensionImage(rs2, mind, dim2, no_packs2);
}

void RoughJoinWatcher::DimensionImage( RSValue *rs, MultiIndex &mind, int dim, int no_packs )
{
	if(mind.GetFilter(dim))
		for(int i = 0; i < no_packs; i++)
			rs[i] = ( mind.GetFilter(dim)->IsFull(i)  ? RS_ALL :
					( mind.GetFilter(dim)->IsEmpty(i) ? RS_NONE : RS_SOME ) ) ;
	else {
		for(int i = 0; i < no_packs; i++)
			rs[i] = RS_NONE;
		int packs_already_marked = 0;
		int cur_pack;
		MIIterator mit(&mind, dim);
		while(mit.IsValid()) {
			if(mit[dim] != NULL_VALUE_64) {
				cur_pack = mit.GetCurPackrow(dim);
				if(cur_pack >= 0 && rs[cur_pack] == RS_NONE) {		// in case of materialized multiindex, we cannot check completeness of a pack
					rs[cur_pack] = RS_SOME;
					packs_already_marked++;
					if(packs_already_marked == no_packs)
						break;			// no more packs to be marked
				}
				mit.NextPackrow();
			}
			else
				++mit;
		}
	}
}

void RoughJoinWatcher::UpdatePackPack( MultiIndex &mind )
{
	if( !can_update_p2p || !rsi_manager ||		// may be switched off in CheckIfUpdatePossible()
		( mind.GetFilter(dim1) == NULL && mind.GetIndTable(dim1) == NULL ) ||
		( mind.GetFilter(dim2) == NULL && mind.GetIndTable(dim2) == NULL ) )	// a dimension is not available in join result
		return;

	RSIndex_PP *rsi_pp = NULL;
	rsi_pp = (RSIndex_PP *)(rsi_manager->GetIndexForUpdate(RSIndexID(RSI_JPP, t1->GetID(), attr1, t2->GetID(), attr2),0));
	if( !rsi_pp )
		return;			// cannot update for some reason

//	rsi_pp->SaveAsText("/home/jakubw/p2p.txt");
  	rsi_pp->Update( t1->NoObj(), t2->NoObj() );		// new table sizes

	bool anything_to_update = false;
	BinMatrix potentially_excluded( t1->NoPacks(), t2->NoPacks(), false );
	_int64 pairs_to_be_updated = 0;
	for( int p1 = 0; p1 < t1->NoPacks(); p1++ )	{
		if( rs1[p1] != RS_NONE )
			for( int p2 = 0; p2 < t2->NoPacks(); p2++ )
				if( rs2[p2] != RS_NONE ) {
					if( rsi_pp->IsUpToDate( p1, p2 ) == false ) {
						anything_to_update = true;
						pairs_to_be_updated++;
						if( rs1[p1] == RS_ALL && rs2[p2] == RS_ALL )
							potentially_excluded.Set( p1, p2, true );
					}
				}
	}
	if( anything_to_update == false ) {
		rsi_manager->UpdateIndex( rsi_pp, 0, true );	// unlock without save
		return;
	}

	// update algorithm:
	//		- pair of packs found and not up to date =>
	//			- potentially_excluded is false,
	//			- if condition is equality => set as intersecting and up to date
	//		- after checking all the result:
	//			- if still potentially_excluded => set as non-intersecting and up to date

	rccontrol.lock(mind.m_conn->GetThreadID()) << "Updating P2P..." << unlock;
	anything_to_update = false;
	_int64 pairs_already_updated = 0;
	int p1, p2;

	DimensionVector dims(&mind);
	dims[dim1] = true;
	dims[dim2] = true;
	MIIterator mit(&mind, dims);
	mit.Rewind();
	while(mit.IsValid()) {
		p1 = mit.GetCurPackrow(dim1);
		p2 = mit.GetCurPackrow(dim2);
		if(p1 >= 0 && p2 >= 0 && rsi_pp->IsUpToDate(p1, p2) == false) {
			potentially_excluded.Set(p1, p2, false);
			if( !inequality_join ) {
				rsi_pp->PutValue( p1, p2, true);		// intersected
				rsi_pp->SetUpToDate( p1, p2 );
				anything_to_update = true;
				pairs_already_updated++;
				if( pairs_already_updated == pairs_to_be_updated )
					break;			// no more packs to be updated
			}
		}
		mit.NextPackrow();
	}
/*
	mind.LockForGetIndex( dim1 );
	mind.LockForGetIndex( dim2 );
	if( mind.GetIndTable(dim1) && mind.GetIndTable(dim2) ) {		// normal case: materialized result
		bool next_exists = mind.Rewind( dim1 );
		while( next_exists ) {
			_int64 loc_obj1 = mind.GetCurrent( dim1 );
			_int64 loc_obj2 = mind.GetCurrent( dim2 );	// NOTE: we iterate by dim1, but in this case dim2 is also material, so it is OK
			if( loc_obj1 != NULL_VALUE_64 && loc_obj2 != NULL_VALUE_64 ) {
				p1 = ( loc_obj1 >> 16 );
				p2 = ( loc_obj2 >> 16 );
				if( rsi_pp->IsUpToDate( p1, p2 ) == false ) {
					potentially_excluded.Set( p1, p2, false );
					if( !inequality_join ) {
						rsi_pp->PutValue( p1, p2, true);		// intersected
						rsi_pp->SetUpToDate( p1, p2 );
						anything_to_update = true;
						pairs_already_updated++;
						if( pairs_already_updated == pairs_to_be_updated )
							break;			// no more packs to be updated
					}
				}
			}
			next_exists = mind.Next( dim1 );
		}
	}
	else {
		// not implemented yet: very special cases
		rsi_manager->UpdateIndex( rsi_pp, 0, true );	// unlock without save
		return;
	}
	mind.UnlockFromGetIndex( dim1 );
	mind.UnlockFromGetIndex( dim2 );
*/

	// All intersections found. Now we may update non-intersecting pairs.

	for( int p1 = 0; p1 < t1->NoPacks(); p1++ )
		if( rs1[p1] != RS_NONE )
			for( int p2 = 0; p2 < t2->NoPacks(); p2++ )
				if( potentially_excluded.Get( p1, p2 ) ) {
					anything_to_update = true;
					rsi_pp->PutValue( p1, p2, false);		// non-intersecting pair
					rsi_pp->SetUpToDate( p1, p2 );
					pairs_already_updated++;
				}

	// Finishing

	if( anything_to_update == false )
		rsi_manager->UpdateIndex( rsi_pp, 0, true );	// unlock without save
	else {
		rccontrol.lock(mind.m_conn->GetThreadID())
			<< "RSI_PP updated by "  << int(rsi_pp->UpdateDensity()*10000)/100.0
			<< "%. Pairs possible: " << int(rsi_pp->Density()*10000)/100.0 << "%" << unlock;
		rsi_manager->UpdateIndex( rsi_pp, 0 );		// save updates and unlock
	}
}
