/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */


#include <vector>

#include "common/CommonDefinitions.h"
#include "MysqlExpression.h"
#include "Descriptor.h"
#include "ParametrizedFilter.h"
#include "CompiledQuery.h"
#include "CachedBuffer.h"
#include "CQTerm.h"
#include "JustATable.h"
#include "MultiIndex.h"
#include "system/LargeBuffer.h"
#include "shared/exporter/DataExporter.h"
#include "MIUpdatingIterator.h"
#include "PhysicalColumn.h"
#include "PackOrderer.h"

#ifndef _TEMPTABLE_H_
#define _TEMPTABLE_H_

class Filter;
class GroupTable;
class GroupByWrapper;
class DistinctWrapper;

typedef unsigned int uint;

struct CQTerm;
class Descriptor;
class SortDescriptor;
struct Attr;
struct DescTreeNode;
enum JoinType;
//////////////////////////////////////////////////////////////////////////////////
//
//  TempTable - for storing a definition or content of a temporary table (view)
//

class TempTable : public JustATable
{
public:
	typedef std::vector<int> aliases_t;
	typedef std::vector<JustATable*> tables_t;
	template<class T>
	class AttrBuffer : public CachedBuffer<T> {
	public:
		explicit AttrBuffer(uint page_size, uint elem_size = 0, ConnectionInfo *conn = NULL);
		explicit AttrBuffer(_uint64 size, T value, uint page_size, uint elem_size = 0, bool constant = false, ConnectionInfo *conn = NULL);
		~AttrBuffer();

		// Warning: read only operator!!!
		T& operator[](_uint64 idx) { return constant_value ? CachedBuffer<T>::Get(0) : CachedBuffer<T>::Get(idx); }
		void Set(_uint64 idx, T value);
	private:
		bool			constant_value;		// only one value in the whole buffer, stored in buf[0]
	};

	class Attr : public PhysicalColumn
	{
	public:
		void *			buffer; // buffer to values of attribute, if materialized
		_uint64			no_obj; // number of objects in the buffer
		uint			page_size; // size of one page of buffered values
		char *			alias;
		ColOperation	mode;
		bool			distinct; // distinct modifier for aggregations
		CQTerm			term;
		int				dim;	// dimension of a column, i.e., index of source table,
								// -1 means there is no corresponding table: constant, count(*)
		RCBString		bufS; // temporary buffer of one RCBString

		Attr();
		Attr(CQTerm t,
			ColOperation	m,
			bool			distinct = false,
			char *			alias = NULL,
			int				dim = -1,
			AttributeType	type = RC_INT,
			uint			scale = 0,
			uint			precision = 10,
			NullMode		n = NO_NULLS, 
			DTCollation collation = DTCollation()
			);
		Attr(const Attr &);
		Attr& operator=(const Attr &);
		int operator==(const Attr &);
		~Attr();

		//! Use in cases where actual string length is less than declared, before materialization of Attr
		void OverrideStringSize(int size) 	{ ct.OverrideInternalSize(size); ct.SetPrecision(size); }
		void SetPrecision(int prec) 		{ ct.SetPrecision(prec); }
		void SetScale(int sc) 				{ ct.SetScale(sc); }

		void CreateBuffer(_uint64 size, ConnectionInfo *conn = NULL);
		//void ResizeBuffer(_uint64 new_size);
		void SetValueString(_uint64 obj, RCBString & val);
		RCValueObject GetValue(_uint64 obj, bool lookup_to_num = false);
		RCBString& GetValueString(_uint64 obj);
		void GetValueString(_uint64 row, RCBString& s, bool bin_as_hex)	{ s = GetValueString(row); }
		_uint64 ApproxDistinctVals(bool incl_nulls, ::Filter* f = NULL);	// provide the best upper approximation of number of diff. values (incl. null, if flag set)
		_uint64 ExactDistinctVals(Filter* f)		{ return NULL_VALUE_64; }
		bool IsDistinct(::Filter* f) 				{ return false; }
		ushort MaxStringSize(::Filter* f = NULL);		// maximal byte string length in column
		_int64 RoughMin(::Filter* f = NULL)			{ return MINUS_INF_64; }	// for numerical: best rough approximation of min for a given filter (or global min if filter is NULL)
		_int64 RoughMax(::Filter* f = NULL)			{ return PLUS_INF_64; }		// for numerical: best rough approximation of max for a given filter (or global max if filter is NULL)
		void DisplayAttrStats(::Filter* f)	{}
		bool TryToMerge(Descriptor &d1,Descriptor &d2)	{ return false; }

		PackOntologicalStatus GetPackOntologicalStatus(int pack_no) {return NORMAL;} //not implemented properly yet

		//bool CheckCondition(_uint64 obj, Descriptor &d);

		void ApplyFilter(MultiIndex &, _uint64 offset, _uint64 no_obj);
//		bool IsNumeric();
		void DeleteBuffer();

		// Methods related to complex expressions (term.c_term)

		//void InferComplexType(MysqlExpression::TypOfVars* vartypes);
		ValueOrNull GetComplexValue(const _uint64 obj);
		void SetComplexValue(_uint64 obj, ValueOrNull val);

		// new interface
		bool IsNull(_uint64 obj) const;
		_int64 GetValueInt64(_uint64 obj) const;
		void SetValueInt64(_uint64 obj, _int64 val);
		int GetValueInt32(_uint64 row) const;
		void GetValueBin(_uint64 row, int& size, char* buf);
//		int	GetValueText(_uint64 obj, char *val_buf);
		_int64 GetMinInt64(int pack);
		_int64 GetMaxInt64(int pack);
		RCBString GetMaxString(int pack);
		RCBString GetMinString(int pack);
		_int64 GetNoNulls(int pack);
		_int64 GetSum(int pack);
		int GetValueSize(const int pack) {return ct.GetInternalSize();}
		uint AttrNo()		 						{ return (uint)-1; }
		RSValue RoughCheck(int pack, Descriptor& d, bool additional_nulls_possible) { return RS_SOME; }
		RSValue RoughCheck(int pack1, int pack2, Descriptor& d) { return RS_SOME; }
		// as far as Attr is not pack oriented the function below should not be called
		void EvaluatePack(MIUpdatingIterator &mit, int dim, Descriptor& desc) { assert(0); }
		RCBString	DecodeValue_S(_int64 code)		{ assert(0); return RCBString(); }	// RCAttr only
	};
	typedef std::vector<Attr*> attrs_t;

	struct TableMode {
		bool	distinct;
		bool	top;
		bool	exists;
		_int64		param1, param2; // e.g., TOP(5), LIMIT(2,5)
		TableMode(): distinct(false), top(false), exists(false), param1(-1), param2(-1) {}
	};

//	TempTable(): no_obj(0), materialized(true), tree(NULL), m_conn(NULL), filter(NULL), manage_used_tables(false) {}
private:
	TempTable(const TempTable &, bool for_subq = false);
	TempTable(JustATable * const , int alias, Query* q);

	TempTablePtr CreateMaterializedCopy(bool translate_order);		// move all buffers to a newly created TempTable, make this VC-pointing to it; the new table must be deleted by DeleteMaterializedCopy()
	void DeleteMaterializedCopy(TempTablePtr t_new);	// delete the external table and remove VC pointers, make this fully materialized

	map<int,vector<PackOrderer::OrderingInfo> > oi;	//for each dimension, what column should be taken into account for ordering

public:
	virtual ~TempTable();

	////////// Query execution (CompiledQuery language implementation)
	void					Filter(ParameterizedFilter *);
	void					SetMode(TMParameter mode, _int64 mode_param1, _int64 mode_param2);
	bool					GetMode(TMParameter mode);
	void					JoinT(JustATable * t, int alias, JoinType jt);
	int						AddColumn(CQTerm, ColOperation, char* alias, bool distinct);
	void					AddOrder(VirtualColumn *vc, int direction);
	void					Union(TempTable*, int);

	////////// Maintenance and low-level functions

	bool					OrderByAndMaterialize(std::vector<SortDescriptor> &ord, _uint64 limit = -1, _uint64 offset = -1); // Sort data contained in ParameterizedFilter by using some attributes (usually specified by AddOrder, but in general - arbitrary)
	void					FillMaterializedBuffers( _uint64 local_limit, _uint64 local_offset);	// just materialize as SELECT *

	void					Materialize(bool in_subq = false);

	void					MaterializeDelayedAttrs();
	// Computes the whole contents of TempTable, file buffering can be necessary.

							// just_distinct = 'select distinct' but no 'group by'
							// Return no. of rows in result ( = no. of groups).
	_int64					Aggregate(bool just_distinct, _int64 limit, _int64 offset);
	void 					AggregateFillOutput(GroupByWrapper &gbw, _int64 gt_pos, _int64 &cur_output_tuple);
	bool					AggregateRough(GroupByWrapper &gbw, MIIterator &mit, bool &packrow_done, bool &part_omitted, bool &ag_not_changeabe, _int64 &no_groups, _int64 &uniform_pos, _int64 rows_in_pack, _int64 factor, int just_one_aggr = -1);
	void 					VerifyAttrsSizes();	// verifies attr[i].field_size basing on the current multiindex contents

	_uint64					Export(const IOParameters& iop, fields_t const& fields);
protected:
	//void					InitRCDEAforExport(std::vector<RCDEA>& deas);
	_uint64					ExportFromMaterialized(DataExporter& rcde, fields_t const& fields);
	_uint64					ExportFromNonMaterialized(DataExporter& rcde, fields_t const& fields);
public:
	//void					SetConnectionInfo(ConnectionInfo* m_c)			{ m_conn = m_c; filter.m_index->m_conn = *m_c; }	// external pointer
	void 					SuspendDisplay()	{ m_conn.SuspendDisplay(); }
	void 					ResumeDisplay()		{ m_conn.ResumeDisplay(); }

	// Dummy - used actively in RCTable only
	void					LockPackForUse(unsigned attr, unsigned pack_no, ConnectionInfo& conn)	{}
	void					UnlockPackFromUse(unsigned attr, unsigned pack_no) {}

	////////// Info
	_uint64					NoObj() { return no_obj; }
	TType					TableType() const { return TEMP_TABLE; }	// type of JustATable - TempTable

	uint					NoAttrs()				{ return (uint)attrs.size(); }
	uint					NoDisplaybleAttrs()		{ return no_cols; }  // no. of columns with defined alias
	bool					IsDisplayAttr(int i)	{ return attrs[i]->alias != NULL; }

	_int64					GetTable64(_uint64 obj, int attr);
	int						GetTable32(_uint64 obj, int attr);
	RCBString*				GetTable_pS(_uint64 obj, uint attr); // Get pointer to the Attr's buffer of type RCBString
	RCBString				GetTable_S(_uint64 obj, int attr, bool bin_as_hex = true);
	void					GetTable_B(_uint64 obj, int attr, int &size, char *val_buf);
	RCValueObject			GetTable(_uint64 obj, uint attr);
	int						GetLength(_uint64 obj, int attr, ConnectionInfo *conn) { return GetTable_pS(obj, attr)->len; }
	_uint64					ApproxAnswerSize(int attr, Descriptor& d);	// provide the most probable approximation of number of objects matching the condition

	bool					IsNull(_uint64 obj, int attr); // return true if the value of attr. is null
	bool					IsAttrNumeric(uint a);
/*
	_int64					AttrSum(int n_a, ::Filter *f, _int64 &counter, _int64* quant = 0, int q_pack_start = 0, int q_pack_stop = 0, _int64 quant_factor = 1, bool distinct = false, ConnectionInfo *conn = NULL);
	_int64					AttrMin(int n_a, ::Filter *f,ConnectionInfo *conn);
	_int64					AttrMax(int n_a, ::Filter *f,ConnectionInfo *conn);
	double					AttrAvg(int n_a, ::Filter *f, _int64* quant = 0, bool distinct = false, ConnectionInfo *conn = NULL);
	_int64					AttrCountNotNull(int n_a, ::Filter *f, _int64 *quant = 0, int q_pack_start = 0, int q_pack_stop = 0,ConnectionInfo *conn=NULL);
	RCBString				AttrMinS(int n_a, ::Filter *f,ConnectionInfo *conn);
	RCBString				AttrMaxS(int n_a, ::Filter *f,ConnectionInfo *conn);
	*/
	RSValue					AttrIsDistinct(int n_a) 									{ return RS_UNKNOWN; } // three-valued answer

	_int64 					RoughMin(int n_a, ::Filter *f = NULL)					{ return MINUS_INF_64; }
	_int64 					RoughMax(int n_a, ::Filter *f = NULL)					{ return PLUS_INF_64; }
	ushort					MaxStringSize(int n_a, ::Filter *f = NULL)				{ if(n_a<0) return GetFieldSize(-n_a - 1);
																					  return GetFieldSize(n_a); }

	uint					NoDimensions() { return filter.m_index ? (uint)filter.m_index->NoDimensions() : 1; }
	int						GetDimension(int alias);
	AttributeType			AttrType(int a) { assert(a >= 0 && (uint)a < NoAttrs()); return attrs[a]->TypeName(); }
	std::vector<ATI>		GetATIs();
	int						GetAttrScale(int a) { assert(a >= 0 && (uint)a < NoAttrs()); return attrs[a]->Type().GetScale(); }
	int						GetAttrSize(int a) { assert(a >= 0 && (uint)a < NoAttrs()); return attrs[a]->Type().GetDisplaySize(); }
	int						GetFieldSize(int a) { assert(a >= 0 && (uint)a < NoAttrs()); return attrs[a]->Type().GetInternalSize() ; }
	int						GetNoDigits(int a) { assert(a >= 0 && (uint)a < NoAttrs()); return attrs[a]->Type().GetPrecision(); }
	const ColumnType&		GetColumnType(int a) { assert(a >= 0 && (uint)a < NoAttrs()); return attrs[a]->Type();}
	const PhysicalColumn* 	GetColumn(int a) { assert(a >= 0 && (uint)a < NoAttrs()); return attrs[a]; }
	const PhysicalColumn*	GetSourceColumn(int alias, int col_no);
	NullMode				GetAttrNullMode(int a) { assert(a >= 0 && (uint)a < NoAttrs()); return attrs[a]->Type().GetNullsMode(); }
	Attr*					GetAttrP(uint a) { assert(a < (uint)NoAttrs()); return attrs[a]; }
	Attr*					GetDisplayableAttrP(uint i);
	uint GetDisplayableAttrIndex(uint attr);
	MultiIndex*				GetMultiIndexP() { return filter.m_index; }
	MultiIndex*				GetOutputMultiIndexP() { return &output_mind; }
	ParameterizedFilter*		GetFilterP() { return & filter; }
	JustATable*				GetTableP(uint dim) { assert(dim<tables.size()); return tables[dim]; }
	int 					NoTables() const	{ return tables.size(); }
	std::vector<JoinType>&		GetJoinTypes() { return join_types; }
	aliases_t&				GetAliases() { return aliases; }
	tables_t&				GetTables() { return tables; }
	bool					IsMaterialized() { return materialized; }
	uint					NoParameters() { return filter.NoParameters(); }
	uint					NoUnsetParameters() { return filter.NoUnsetParameters(); }
	CQTerm					GetParameter(uint i) { assert(i<NoParameters()); return filter.GetParameter(i); }
	int						CalculatePageSize(_int64 no_obj = -1);	// computes number of output records kept in memory, updates attrs[]
								// based on no_obj and some upper limit CACHE_SIZE (100MB)

	void					DisplayRSI();	// display info about all RSI contained in RCTables from this TempTable

	// a function to check if TempTable 'what' is used by current TempTable
	bool					uses(TempTable * what);
	bool					uses(ParameterizedFilter *f, TempTable * what);

	//std::set<int>				ForeignTables() { return filter.ForeignTables(); }
	bool					HasHavingConditions() { return tree != NULL; }
	void 					SetManageUsedTables() { manage_used_tables = true; }
	void 					SetManageUsedTables(bool new_state) { manage_used_tables = new_state; }
	void 					RemoveFromManagedList(const RCTable* rct);
	int						AddVirtColumn(VirtualColumn* vc);
	int						AddVirtColumn(VirtualColumn* vc, int no);
	uint					NoVirtColumns() { return virt_cols.size(); }
	void					ReserveVirtColumns(int no) {assert(!no || no> virt_cols.size()); virt_cols.resize(no); virt_cols_for_having.resize(no); }
	VirtualColumn*			GetVirtualColumn(uint col_num) { assert(col_num < virt_cols.size()); return virt_cols[col_num]; }
	void 					ProcessParameters();
	Query*					GetQuery() {return query;}

protected:
	_uint64					no_obj;
	uint					no_cols;	// no. of output columns, i.e., with defined alias
	_uint64					buf_index;	// index of first row of TT cached in column buffers
										// e.g., 100 means that all the buffers contain values of objects starting from 101th row of TT
										// size of buffer determines index of last object
	TableMode				mode;		// based on { TM_DISTINCT, TM_TOP, TM_EXISTS }
	attrs_t					attrs; // vector of output columns, each column contains a buffer with values
	tables_t				tables; // vector of pointers to source tables
	aliases_t				aliases; // vector of aliases of source tables
	vector<VirtualColumn*>	virt_cols; // vector of virtual columns defined for TempTable
	vector<bool>			virt_cols_for_having;	//is a virt column based on output_mind (true) or source mind (false)
	std::vector<JoinType>	join_types; // vector of types of joins, one less than tables
	ParameterizedFilter		filter;		// multidimensional filter, contains multiindex, can be parametrized
	MultiIndex				output_mind; // one dimensional MultiIndex used for operations on output columns of TempTable
	DescTree*				tree;		// tree of conditions defining filter on output columns, used, e.g., by 'having' clause
	std::vector<SortDescriptor>	order_by;// indexes of order by columns
	bool					group_by;	// true if there is at least one grouping column
	bool					for_subq;	// true if this is a temptable copy for subq execution
	int						no_global_virt_cols; // keeps number of virtual columns. In case for_subq == true, all locally created
							// virtual columns will be removed

	////////////////////////////////////////////////////
	// some internal functions for low-level query execution
	static const uint		CACHE_SIZE = 100000000;	// size of memory cache for data in the materialized TempTable
												// everything else is cached on disk (CachedBuffer)
	////////////////////////////////////////////////////
	bool					IsParametrized(DescTreeNode *); //do the having conditions use parameters?
	bool					IsParametrized(); //is the temptable (select) parametrized?

	void					ApplyDescTree(MultiIndex &mind, DescTreeNode *);
	void					ApplyOffset(_int64 limit, _int64 offset); // apply limit and offset to attr buffers

	bool					materialized;
	bool 					manage_used_tables; // true if destructor should destroy all temptables stored in tables<>
	bool					has_temp_table;

	ConnectionInfo&			m_conn;					// external pointer
	Query*					query;						//query containing this tempTable

	////////////////////////////////////////////////////
	// Refactoring: extracted small methods
	bool					HasDelayedAttrs();
	ColumnType				GetUnionType(ColumnType type1, ColumnType type2);
	//void 					ExtractColumnsFromTerm(CQTerm& term, bool& is_complex, DataType& type, std::map<VarID, ValueOrNull>& columns,
	//												MysqlExpression::SetOfVars& vars, MysqlExpression::TypOfVars& types);
	void					UnlockLastPacks(unsigned* last_pack);
	_int64					MultiDimensionalGroupByScan(GroupByWrapper &gbw, _int64 upper_approx_of_groups, _int64 limit, _int64 offset);
	void					MultiDimensionalDistinctScan(GroupByWrapper &gbw, DimensionVector &dims);
	virtual RecordIterator	do_find( ConnectionInfo *conn, _uint64 pos, RcTableLockPolicy::attr_uses_t const& );
	virtual RecordIterator	do_find( ConnectionInfo *conn, _uint64 pos, ::Filter const*, RcTableLockPolicy::attr_uses_t const& );
	virtual RecordIterator	do_begin( ConnectionInfo *conn, RcTableLockPolicy::attr_uses_t const& );
	virtual RecordIterator	do_begin( ConnectionInfo *conn, ::Filter const& filter, RcTableLockPolicy::attr_uses_t const& );
	virtual RecordIterator	do_end( ConnectionInfo *conn );

public:
	void					Display(std::ostream &out = std::cout); // output to console

public:
	static TempTablePtr Create(const TempTable &, bool for_subq = false);
	static TempTablePtr Create(JustATable * const , int alias, Query* q);
};

struct OrderedListNode {
	OrderedListNode(_int64 key): next(NULL), prev(NULL), key(key), count(1) {}
	OrderedListNode *next, *prev;
	_int64 key;
	_int64 count;
};

class OrderedList {
	OrderedListNode *head, *tail, *curr;
	int size;
public:
	OrderedList(): head(NULL), tail(NULL), curr(NULL), size(0) {}
	~OrderedList();
	void Insert(_int64 key); // inserts key to a list, if exists increases count
	int Size() { return size; }
	void Head() { curr = head; }
	void Tail()	{ curr = tail; }
	bool Next() { if(!curr || !curr->next) return false; curr = curr->next; return true; }
	bool Prev() { if(!curr || !curr->prev) return false; curr = curr->prev; return true; }
	bool EOL() { return curr->next == NULL; }
	bool BOL() { return curr->prev == NULL; }
	_int64 CurrentKey() { return curr ? curr->key : NULL_VALUE_64; }
	_int64 CurrentCount() { return curr ? curr->count : NULL_VALUE_64; }
};

struct DescTreeNode {
	DescTreeNode(LogicalOperator _lop, TempTable* t) : desc(t), left(NULL), right(NULL), parent(NULL)
	{ desc.lop = _lop; }

	DescTreeNode(CQTerm e1, Operator op, CQTerm e2, CQTerm e3, TempTable* t ) : desc(t), left(NULL), right(NULL), parent(NULL)
	{ desc.attr = e1; desc.op = op; desc.val1 = e2; desc.val2 = e3; desc.Simplify(true); desc.CoerceColumnTypes(); }

	DescTreeNode(DescTreeNode &n, bool in_subq = false) : desc(n.desc), left(NULL), right(NULL), parent(NULL) { }
	bool CheckCondition(MIIterator& mit);
	void Simplify();
	Descriptor desc;
	DescTreeNode *left, *right, *parent;
};

// Tree of descriptors: internal nodes - predicates, leaves - terms
// Used to store conditions filtering output columns of TempTable
// Used, particularly, to store conditions of 'having' clause

class DescTree {
public:
	DescTreeNode *root, *curr;
	DescTreeNode * Copy(DescTreeNode * node);
	void Release(DescTreeNode * &node); // releases whole subtree starting from node

	DescTree(CQTerm e1, Operator op, CQTerm e2, CQTerm e3, TempTable* q);
	DescTree(DescTree &);
	~DescTree() { Release(); }

	void Display();
	void Display(DescTreeNode * node);

	void Release(); // releases whole subtree

	bool Left() { if(!curr || !curr->left) return false; curr = curr->left; return true; }
	bool Right() { if(!curr || !curr->right) return false; curr = curr->right; return true; }
	bool Up() { if(!curr || !curr->parent) return false; curr = curr->parent; return true; }
	void Root() { curr = root; }

	// make lop the root, make current tree the left child, make descriptor the right child
	void AddDescriptor(LogicalOperator lop, CQTerm e1, Operator op, CQTerm e2, CQTerm e3, TempTable* t);
	// make lop the root, make current tree the left child, make tree the right child
	void AddTree(LogicalOperator lop, DescTree *tree);
};

#endif

