/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "ValueSet.h"
#include "common/CommonDefinitions.h"
#include "types/RCDataTypes.h"
#include "Filter.h"
#include "RCAttrTypeInfo.h"

using namespace std;

ValueSet::ValueSet()
{
	prepared = false;
	contains_nulls = false;
	prep_type = RC_UNKNOWN;
	prep_scale = 0;
	prep_collation = DTCollation();
	min = max = 0;
	values = new RCDTptrHashSet;
	no_obj = 0;
}

ValueSet::ValueSet(const ValueSet& sec) :
	prepared(sec.prepared), prep_type(sec.prep_type), prep_scale(sec.prep_scale), prep_collation(sec.prep_collation)
{
	values = new RCDTptrHashSet;
	no_obj = sec.NoVals();
	contains_nulls = sec.contains_nulls;
	min = max = 0;
	if(sec.values) {
		for( RCDTptrHashSet::iterator it = sec.values->begin(), end = sec.values->end(); it != end; it++)
			values->insert((*it)->Clone().release());
	}

	if(sec.min)
		min = sec.min->Clone().release();

	if(sec.max)
		max = sec.max->Clone().release();
}

ValueSet::~ValueSet()
{
	Clear();
	delete values;
}

void ValueSet::Clear( void )
{
	RCDTptrHashSet::iterator iter = values->begin();
	RCDTptrHashSet::iterator tmp;
	RCDTptrHashSet::iterator end = values->end();
	while(iter != end) {
		tmp = iter++;
		delete *tmp;
	}
	values->clear();
	delete min;
	min = 0;
	delete max;
	max = 0;
	prepared = false;
	prep_type = RC_UNKNOWN;
	prep_scale = 0;
	prep_collation = DTCollation();
}

void ValueSet::Add(auto_ptr<RCDataType> rcdt)
{
	if(rcdt->IsNull()) {
		contains_nulls = true;
		return;
	}
	RCDataType* rcv = rcdt.release();
	if(prep_type != rcv->Type() || (rcv->Type() == RC_NUM && static_cast<RCNum&>(*rcv).Scale() != prep_scale))
		prepared = false;

	if(!values->insert(rcv).second)
		delete rcv;
	else if(prepared) {
		if(RequiresUTFConversions(prep_collation)) {
			//int len1 = prep_collation.collation->coll->strnxfrmlen(prep_collation.collation, rcv->ToRCString().len * prep_collation.collation->mbmaxlen);
			//int len2 = prep_collation.collation->coll->strnxfrmlen(prep_collation.collation, min->ToRCString().len * prep_collation.collation->mbmaxlen);
			//int len3 = prep_collation.collation->coll->strnxfrmlen(prep_collation.collation, max->ToRCString().len * prep_collation.collation->mbmaxlen);
			//int buf_len = len1 > len2 ? len1 : len2;
			//buf_len = buf_len > len3 ? buf_len : len3;
			//RCBString buf_v(NULL, buf_len, true);
			//RCBString buf_min(NULL, buf_len, true);
			//RCBString buf_max(NULL, buf_len, true);
			//ConvertToBinaryForm(rcv->ToRCString(), buf_v, buf_len, prep_collation.collation, false);
			//ConvertToBinaryForm(min->ToRCString(), buf_min, buf_len, prep_collation.collation, false);
			//ConvertToBinaryForm(max->ToRCString(), buf_max, buf_len, prep_collation.collation, false);
			if(min->IsNull() || CollationStrCmp(prep_collation, rcv->ToRCString(), min->ToRCString()) < 0)
				*min = *rcv;
			if(max->IsNull() || CollationStrCmp(prep_collation, rcv->ToRCString(), max->ToRCString()) > 0)
				*max = *rcv;
		} else {
			if(min->IsNull() || *rcv < *min)
				*min = *rcv;
			if(max->IsNull() || *rcv > *max)
				*max = *rcv;
		}
	}

	no_obj = (int)values->size();
}

void ValueSet::Add(const RCValueObject& rcv)
{
	if(rcv.IsNull())
		contains_nulls = true;
	else
		Add(rcv.Get()->Clone());
}

bool ValueSet::Contains(const RCDataType& v, DTCollation coll)
{
	if(!prepared)
		Prepare(v.Type(), v.Type() == RC_NUM ? ((RCNum*)&v)->Scale() : 0, coll);

	if(v.IsNull())
		return false;
	if(RequiresUTFConversions(coll)) {
		RCDTptrHashSet::iterator it = values->begin();
		RCDTptrHashSet::iterator it_end = values->end();
		//RCDataType* rcdt_it = 0;
		//RCBString buf_itv(NULL, ((RCBString*)&v)->len, true);
		while(it != it_end) {
			//ConvertToBinaryForm((*(*it)).ToRCString(), buf_itv, buf_itv.len, coll.collation, false);
			if(CollationStrCmp(coll, (*(*it)).ToRCString(), v.ToRCString()) == 0) 
				return true;
			it++;
		}
		return false;
	} else
		return (values->size() > 0 && values->find(const_cast<RCDataType*>(&v)) != values->end());
}

bool ValueSet::Contains(const RCValueObject& v, DTCollation coll)
{
	if(v.IsNull())
		return false;
	return Contains(*v.Get(), coll);
}

string ValueSet::ToText()
{
	char buf[16];
	string res("");
	res += int2str(NoVals(), buf, 10, 0);
	res += " item";
	if(NoVals()!=1)
		res+="s";
	return res;
}

void ValueSet::Prepare(AttributeType at, int scale, DTCollation coll)
{
	if(!IsPrepared(at, scale, coll)) {
		auto_ptr<RCDTptrHashSet> new_values(new RCDTptrHashSet()); // warning C6211: Leaking memory 'new_values', '*this[12]' due to an exception
		RCDTptrHashSet::iterator it = values->begin();
		RCDTptrHashSet::iterator it_end = values->end();
		RCDTptrHashSet::iterator tmp;
		RCDataType* rcdt_it = 0;


		if(ATI::IsStringType(at)) {
			is_text = true;
			delete min;
			min = new RCBString();
			delete max;
			max = new RCBString();
			it = values->begin();
			bool min_set = false, max_set = false;
			if(RequiresUTFConversions(coll)) {
				while(it != it_end) {
					RCBString* v = new RCBString((*(*it)).ToRCString());
					if(!min_set || CollationStrCmp(coll, v->ToRCString(), min->ToRCString()) < 0) {
						*min = *v;
						min_set = true;
					}
					if(!max_set || CollationStrCmp(coll, v->ToRCString(), max->ToRCString()) > 0) {
						*max = *v;
						max_set = true;
					}
					new_values->insert(v);
					tmp = it++;
					delete *tmp;
				}
			} else {
				while(it != it_end) {
					RCBString* v = new RCBString((*(*it)).ToRCString());
					if(!min_set || *v < *min) {
						*min = *v;
						min_set = true;
					}
					if(!max_set || *v > *max) {
						*max = *v;
						max_set = true;
					}
					new_values->insert(v);
					tmp = it++;
					delete *tmp;
				}
			}
			delete values;
			values = new_values.release();
		} else if(ATI::IsIntegerType(at)) {
			delete min;
			min = new RCNum();
			delete max;
			max = new RCNum();
			while(it != it_end) {
				rcdt_it = (*it);
				if(RCNum* rcn = dynamic_cast<RCNum*>(rcdt_it)) {
					if(rcn->IsInt() || rcn->IsReal()) {
						if(!(*rcn > *min))
							*min = *rcn;
						if(!(*rcn < *max))
							*max = *rcn;
						if(!new_values->insert(rcn).second) {
							tmp = it++;
							delete *tmp;
						} else
							it++;
					} else {
						tmp = it++;
						delete *tmp;
					}
				} else if(rcdt_it->Type() == RC_STRING) {
					RCNum* rcn = new RCNum();
					if(RCNum::Parse(rcdt_it->ToRCString(), *rcn, at) == BHRC_SUCCESS) {
						if(!(*rcn > *min))
							*min = *rcn;
						if(!(*rcn < *max))
							*max = *rcn;
						if(!new_values->insert(rcn).second)
							delete rcn;
					} else {
						delete rcn;
						rcn = new RCNum(0, scale, false, at);
						if(!new_values->insert(rcn).second)
							delete rcn;
					}
					tmp = it++;
					delete *tmp;
				} else {
					tmp = it++;
					delete *tmp;
				}
			}
			delete values;
			values = new_values.release();
		} else if(at == RC_NUM) {
			delete min;
			min = new RCNum();
			delete max;
			max = new RCNum();

			while(it != it_end) {
				rcdt_it = (*it);
				if(RCNum* rcn = dynamic_cast<RCNum*>(rcdt_it)) {
					if(rcn->IsDecimal(scale)) {
						RCNum* rcn_new = new RCNum();
						RCDataType::ToDecimal(*rcn, scale, *rcn_new);
						if(!(*rcn_new > *min))
							*min = *rcn_new;
						if(!(*rcn_new < *max))
							*max = *rcn_new;
						if(!new_values->insert(rcn_new).second)
							delete rcn_new;
					}
				} else if(rcdt_it->Type() == RC_STRING) {
					RCNum* rcn = new RCNum();
					if(RCNum::Parse(rcdt_it->ToRCString(), *rcn, RC_NUM) == BHRC_SUCCESS && rcn->IsDecimal(scale)) {
						if(!(*rcn > *min))
							*min = *rcn;
						if(!(*rcn < *max))
							*max = *rcn;
						if(!new_values->insert(rcn).second)
							delete rcn;
					} else {
						delete rcn;
						rcn = new RCNum(0, scale, false, at);
						if(!new_values->insert(rcn).second)
							delete rcn;
					}
				}
				tmp = it++;
				delete *tmp;
			}

			delete values;
			values = new_values.release();
		} else if(ATI::IsRealType(at)) {
			delete min;
			min = new RCNum(at);
			delete max;
			max = new RCNum(at);

			while(it != it_end) {
				rcdt_it = (*it);
				if(RCNum* rcn = dynamic_cast<RCNum*>(rcdt_it)) {
					if(ATI::IsRealType(rcn->Type())) {
						if(!(*rcn > *min))
							*min = (*rcn);
						if(!(*rcn < *max))
							*max = (*rcn);
						if(!new_values->insert(rcn).second) {
							tmp = it++;
							delete *tmp;
						} else
							it++;
					} else {
						RCNum* rcn_new = new RCNum(rcn->ToReal());
						if(!(*rcn_new > *min))
							*min = (*rcn_new);
						if(!(*rcn < *max))
							*max = (*rcn_new);
						if(!new_values->insert(rcn_new).second)
							delete rcn_new;
						tmp = it++;
						delete *tmp;
					}
				} else if(rcdt_it->Type() == RC_STRING) {
					RCNum *rcn = new RCNum();
					if(RCNum::ParseReal(*(RCBString*)rcdt_it, *rcn, at) == BHRC_SUCCESS) {
						if(!(*rcn > *min))
							*min = *rcn;
						if(!(*rcn < *max))
							*max = *rcn;
						if(!new_values->insert(rcn).second)
							delete rcn;
					} else {
						delete rcn;
						rcn = new RCNum(0, scale, true, at);
						if(!new_values->insert(rcn).second)
							delete rcn;
					}
					tmp = it++;
					delete *tmp;
				} else {
					tmp = it++;
					delete *tmp;
				}
			}
			delete values;
			values = new_values.release();
		} else if(ATI::IsDateTimeType(at)) {
			delete min;
			min = new RCDateTime();
			delete max;
			max = new RCDateTime();

			while(it != it_end) {
				rcdt_it = (*it);
				if(RCDateTime* rcdt = dynamic_cast<RCDateTime*>(rcdt_it)) {
					if(!(*rcdt > *min))
						*min = *rcdt;
					if(!(*rcdt < *max))
						*max = *rcdt;
					if(!new_values->insert(rcdt_it).second) {
						tmp = it++;
						delete *tmp;
					} else
						it++;
				} else if(rcdt_it->Type() == RC_STRING) {
					RCDateTime* rcdt = new RCDateTime();
					if(!BHReturn::IsError(RCDateTime::Parse(rcdt_it->ToRCString(), *rcdt, at))) {
						if(!(*rcdt > *min))
							*min = *rcdt;
						if(!(*rcdt < *max))
							*max = *rcdt;
						if(!new_values->insert(rcdt).second)
							delete rcdt;
					} else {
						delete rcdt;
						rcdt = static_cast<RCDateTime*>(RCDateTime::GetSpecialValue(at).Clone().release());
						if(!new_values->insert(rcdt).second)
							delete rcdt;
					}
					tmp = it++;
					delete *tmp;
				} else if(RCNum* rcn = dynamic_cast<RCNum*>(rcdt_it)) {
					try {
						RCDateTime* rcdt = new RCDateTime(*rcn, at);
						if(!(*rcdt> *min))
							*min = *rcdt;
						if(!(*rcdt < *max))
							*max = *rcdt;
						if(!new_values->insert(rcdt).second)
							delete rcdt;
					} catch (DataTypeConversionRCException) {
						delete rcdt;
						rcdt = static_cast<RCDateTime*>(RCDateTime::GetSpecialValue(at).Clone().release());
						if(!new_values->insert(rcn).second)
							delete rcdt;
					}
					tmp = it++;
					delete *tmp;
				} else {
					tmp = it++;
					delete *tmp;
				}
			}
			delete values;
			values = new_values.release();
		}
		size_t a = 0;

		this->no_obj = (int)values->size();
		prepared = true;
		prep_type = at;
		prep_scale = scale;
		prep_collation = coll;
	}
}

int ValueSet::NoVals() const
{
	if(values)
		return (int)values->size();
	return no_obj;
}

/////////////////////

ValueSet::const_iterator ValueSet::begin() const
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( values );
	if(!prepared)
		return values->end();
	return values->begin();
}

ValueSet::const_iterator ValueSet::end() const
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( values );
	return values->end();
}

_int64 ValueSet::AsInt64( ValueSet::const_iterator const& it ) const
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(values && (it != values->end()));
	_int64 cur( 0 );
	RCNum* rcn = dynamic_cast<RCNum*>(*it);
	if(rcn)
		cur = rcn->Value();
	else
		cur = (dynamic_cast<RCDateTime*>(*it))->GetInt64();
	return cur;
}
