/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef __BH_COMMUNITY__
#	include "enterprise/local.h"
#else /* not __BH_COMMUNITY__ */
#	include "community/local.h"
#endif /* __BH_COMMUNITY__ */
#include "CompiledQuery.h"
#include "compilation_tools.h"
#include "shared/core/RCEngine.h"
#include "ValueSet.h"
#include "MysqlExpression.h"

Item * UnRef(Item * item)
{
	if (!item) return 0;

	bool changed;

	do {
		changed = false;

		if (item->type() == Item::REF_ITEM) {
			item = * (((Item_ref*)item)->ref);
			changed = true;
		};

		Item_int_with_ref * int_ref = dynamic_cast<Item_int_with_ref*>(item);
		if (int_ref) {//item_int_with_ref
			item = int_ref->real_item();
			changed = true;
		};

	} while (changed);

	return item;
}

const char * TablePath(TABLE_LIST * tab)
{
	ASSERT_MYSQL_STRING(tab->table->s->path);

	return tab->table->s->path.str;
};


int TableUnmysterify(TABLE_LIST * tab, const char *& database_name, const char *& table_name, const char *& table_alias, const char *& table_path)
{
	ASSERT_MYSQL_STRING(tab->table->s->db);
	ASSERT_MYSQL_STRING(tab->table->s->table_name);
	ASSERT_MYSQL_STRING(tab->table->s->path);

	database_name = tab->table->s->db.str;
	table_name = tab->table->s->table_name.str;
	table_alias = tab->alias;
	table_path = tab->table->s->path.str;;

	return RCBASE_QUERY_ROUTE;
};


/*bool Item2ConstTerm(Item * an_arg, CQTerm *term, ValueSet *v, bool negative)
{
	//This function tries to construct a constant cqterm (or add to v) from the argument.
	//returns true on success and false on a failure.

	an_arg = UnRef(an_arg);

	Item::Type itemType = an_arg->type();

	Item_func_get_user_var *item_func = dynamic_cast<Item_func_get_user_var *>(an_arg);
	if (item_func && item_func->getVarEntry()) {
		item_func->val_int();
		switch (item_func->getVarEntry()->type) {
			case INT_RESULT:
				itemType = Item::INT_ITEM;
				break;
			case REAL_RESULT:
				itemType = Item::REAL_ITEM;
				break;
			case STRING_RESULT:
				itemType = Item::STRING_ITEM;
				break;
			case DECIMAL_RESULT:
				itemType = Item::DECIMAL_ITEM;
				break;
			default:
				itemType = Item::NULL_ITEM;
		}
	}

	if(itemType == Item::NULL_ITEM || (itemType == Item::FUNC_ITEM &&
			((Item_func*) an_arg)->functype() == Item_func::GUSERVAR_FUNC && an_arg->is_null())) {
		if(negative) //null cannot be negative !!!!!!!!!!!!!!
			return false;
		if(v)
			return true; //There is no need to use null in a valueset, cause a value cannot be equal NULL
		*term = CQTerm();
		return true;
	}

	switch (itemType) {
		case Item::INT_ITEM:
		{
			String str;
			RCNum tmp;
			an_arg->val_str(&str);
			if (!an_arg->null_value) {
				size_t len = str.length();
				if(RCNum::Parse(len > 0 ? RCBString(str.c_ptr_safe(), len) : RCBString(""), tmp, RC_BIGINT) != BHRC_SUCCESS)
					if(RCNum::ParseNum(len > 0 ? RCBString(str.c_ptr_safe(), len) : RCBString(""), tmp) != BHRC_SUCCESS)
						return false;
				if (negative) tmp.Negate();
			}
			if (v)
				v->Add(new RCNum(tmp));  //ValuSet has correct destructor and there is no memory leak, since ValueSet v is statically allocated and its destructor will be called
			else
				*term = CQTerm(&tmp);  //the CQTerm constructor makes a deep copy of a variable referenced by a pointer
			return true;
		}
		case Item::STRING_ITEM:
		{
			if (negative)
				return false;
			String str;
			String* tmp = an_arg->val_str(&str); //sets null_value
			if (! an_arg->is_null()) {
				if (v)
					v->Add(new RCBString(tmp->c_ptr_safe(), 0, true));
					//ValuSet has correct destructor and there is no memory leak, since ValueSet v is statically allocated and its destructor will be called
				else
					*term = CQTerm(tmp->c_ptr_safe());
			} else {
				if (v)
					v->Add(new RCBString());
					//ValuSet has correct destructor and there is no memory leak, since ValueSet v is statically allocated and its destructor will be called
				else
					*term = CQTerm();
			}
			return true;
		}
		case Item::VARBIN_ITEM:
		{
			if (negative)
				return false;
			String str;
			String* tmp = an_arg->val_str(&str); //sets null_value
			if (! an_arg->null_value) {
				if (v)
					v->Add(new RCBString(tmp->c_ptr_safe(), 0, true));
					//ValuSet has correct destructor and there is no memory leak, since ValueSet v is statically allocated and its destructor will be called
				else {
					if (an_arg->max_length <=8) {
						*term = CQTerm((_int64)an_arg->val_int(), RC_BIGINT);
					} else
						return false; //too large binary to be treated as BIGINT
				}
			} else {
				if (v)
					v->Add(new RCBString());
					//ValuSet has correct destructor and there is no memory leak, since ValueSet v is statically allocated and its destructor will be called
				else
					*term = CQTerm();
			}
			return true;
		}
		case Item::DECIMAL_ITEM:
		{
			String str;
			RCNum tmp;
			String* str_tmp = an_arg->val_str(&str);
			if (! an_arg->null_value) {
				if(RCNum::Parse(RCBString(str_tmp->c_ptr_safe()), tmp) != BHRC_SUCCESS)
					return false;

				if (negative) tmp.Negate();

				if (v)
					v->Add(new RCNum(tmp));  //ValuSet has correct destructor and there is no memory leak, since ValueSet v is statically allocated and its destructor will be called
				else
					*term = CQTerm(&tmp); //the CQTerm constructor makes a deep copy of a variable referenced by a pointer
			} else {
				if (v)
					v->Add(new RCNum());
				else
					*term = CQTerm();
			}
			return true;
		}
		case Item::REAL_ITEM:
		{
			double val_tmp = an_arg->val_real();  //sets null_value
			if (! an_arg->null_value) {
				if (v)
					v->Add(new RCNum(negative ? -val_tmp : val_tmp ));
				else
					*term = CQTerm(negative ? -val_tmp : val_tmp);
			} else {
				if (v)
					v->Add(new RCNum());
				else
					*term = CQTerm();
			}
			return true;
		}
		case Item::FUNC_ITEM:
		{
			Item_func_neg *item_func_neg = dynamic_cast<Item_func_neg *>(an_arg);
			if ((item_func_neg == 0) || (((Item_func *)an_arg)->arg_count != 1))   //SZN: not a minus or the argument count too large
				return false;

			Item ** args = item_func_neg->arguments();
			Item * arg = UnRef(args[0]);
			return Item2ConstTerm(arg, term, v, true);
		}
	}

	return false;
};*/

int OperationUnmysterify(Item * item, ColOperation& oper, bool& distinct, const int group_by_clause)
{
	distinct = false;

	item = UnRef(item);

	switch(static_cast<int>(item->type())) {
		case Item::NULL_ITEM:
		case Item::INT_ITEM:
		case Item::VARBIN_ITEM:
		case Item::STRING_ITEM:
		case Item::DECIMAL_ITEM:
		case Item::REAL_ITEM:
		case Item::FUNC_ITEM:
		case Item::SUBSELECT_ITEM:
		case Item_bhfield::BHFIELD_ITEM:
		case Item::FIELD_ITEM:  // regular select
			oper = LISTING;/*ColOperation::GROUP_BY : ColOperation::LISTING;*/
			break;

		case Item::SUM_FUNC_ITEM: // min(k), max(k), count(), avg(k), sum
			switch (((Item_sum*)item)->sum_func())
			{
				case Item_sum::MIN_FUNC:
					oper = MIN;/*ColOperation::MIN;*/
					break;
				case Item_sum::MAX_FUNC:
					oper = MAX;/*ColOperation::MAX;*/
					break;
				case Item_sum::SUM_DISTINCT_FUNC:
					distinct = true;
				case Item_sum::SUM_FUNC:
					oper = SUM; /*ColOperation::SUM;*/
					break;
				case Item_sum::AVG_DISTINCT_FUNC:
					distinct = true;
				case Item_sum::AVG_FUNC:
					oper = AVG;/*ColOperation::AVG;*/
					break;
				case Item_sum::COUNT_DISTINCT_FUNC:
					distinct = true;
				case Item_sum::COUNT_FUNC:
					oper = COUNT;/*ColOperation::COUNT;*/
					break;
				case Item_sum::STD_FUNC:
					if(((Item_sum_variance*)item)->sample == 1)
						oper = STD_SAMP;
					else
						oper = STD_POP;
					break;
				case Item_sum::VARIANCE_FUNC:
					if(((Item_sum_variance*)item)->sample == 1)
						oper = VAR_SAMP;
					else
						oper = VAR_POP;
					break;
				case Item_sum::SUM_BIT_FUNC:
					if(dynamic_cast<Item_sum_xor*>(item) != NULL)
						oper = BIT_XOR;
					else if(dynamic_cast<Item_sum_and*>(item) != NULL)
						oper = BIT_AND;
					else
						oper = BIT_OR;
					break;
				case Item_sum::GROUP_CONCAT_FUNC:
					return RETURN_QUERY_TO_MYSQL_ROUTE;
					// NOT IMPLEMENTED YET in the engine
					oper = GROUP_CONCAT;
					break;
				default:
					return RETURN_QUERY_TO_MYSQL_ROUTE;
			}
			break;
		default:
			return RETURN_QUERY_TO_MYSQL_ROUTE;
	}
	return RCBASE_QUERY_ROUTE;
}


//AttributeType SelectType(Query &q, CompiledQuery * cq, CQTerm *term, int how_many)
//{
///*	static CompiledQuery * last_cq = 0;
//	static map<int, int> table_index2number;
//
//	if (cq != last_cq) {
//		last_cq = cq;
//		table_index2number.clear();
//	}
//*/
//	//the most important is the type of the second and third terms, in case
//	//of between they set the type for the constant.
//	//Then the least important is the first term.
//	//We start from the least to the most important
//
//	AttributeType type = RC_UNKNOWN;
//
//	for (int i=0; i<how_many; i++)
//		if (term[i].CQTermType()==CQ_ATTR) {   //it is a column
//
//			int real_number;   //real number of this table in the query list
//			//first we try to find if we already cached this table alias
//			//if not, we must perform the scan and locate the compilation step for it
//
//			bool found = false;   //is it a table from Table list or other TempTable (e.g. in Having cluase)
//
///*			map<int, int>::iterator it=table_index2number.find(term[i].tab_id.n);
//			if (it != table_index2number.end()) {  //already cached
//				real_number = table_index2number[term[i].tab_id.n];
//				found = true;
//			}
//			else { //we need to scan all the compilation steps
//		*/
//				int step_no = cq->NoSteps();
//				for (int st =0; st < step_no; st ++) {
//					CompiledQuery::CQStep step = cq->Step(st);
//					if (step.type == CompiledQuery::TABLE_ALIAS && step.t1.n == term[i].tab_id.n) {
//						real_number = step.t2.n;
//		//				table_index2number[step.t1.n] = real_number;
//						found = true;
//					};
//				};
//		//	}
//
//			if (found) {
//				RCTable *tab = q.Table(real_number);
//				type = tab->AttrType(term[i].attr_id.n);
//			};
//
//		}
//	return type;
//}


const char* fieldType(enum_field_types ft)
{
	// For declaration of enum_field_types see mysql.../include/mysql_com.h
	static const char* lowNames[] = {
			"MYSQL_TYPE_DECIMAL", "MYSQL_TYPE_TINY",
			"MYSQL_TYPE_SHORT",  "MYSQL_TYPE_LONG",
			"MYSQL_TYPE_FLOAT",  "MYSQL_TYPE_DOUBLE",
			"MYSQL_TYPE_NULL",   "MYSQL_TYPE_TIMESTAMP",
			"MYSQL_TYPE_LONGLONG", "MYSQL_TYPE_INT24",
			"MYSQL_TYPE_DATE",   "MYSQL_TYPE_TIME",
			"MYSQL_TYPE_DATETIME", "MYSQL_TYPE_YEAR",
			"MYSQL_TYPE_NEWDATE", "MYSQL_TYPE_VARCHAR",
			"MYSQL_TYPE_BIT" };
	static const char* highNames[] = {
			"MYSQL_TYPE_NEWDECIMAL",
			"MYSQL_TYPE_ENUM",
			"MYSQL_TYPE_SET",
			"MYSQL_TYPE_TINY_BLOB",
			"MYSQL_TYPE_MEDIUM_BLOB",
			"MYSQL_TYPE_LONG_BLOB",
			"MYSQL_TYPE_BLOB",
			"MYSQL_TYPE_VAR_STRING",
			"MYSQL_TYPE_STRING",
			"MYSQL_TYPE_GEOMETRY" };

	int i = (int) ft;
	if(i < sizeof(lowNames)/sizeof(*lowNames))
		return lowNames[i];
	if((i >= 246) && (i <= 255))
		return highNames[i-246];
	return "<unknown field type>";
}

void printItemTree(Item* item, int indent)
{
	static const char* nameOf[] = {
			"FIELD_ITEM", "FUNC_ITEM", "SUM_FUNC_ITEM", "STRING_ITEM", "INT_ITEM",
			"REAL_ITEM", "NULL_ITEM", "VARBIN_ITEM", "COPY_STR_ITEM", "FIELD_AVG_ITEM",
			"DEFAULT_VALUE_ITEM", "PROC_ITEM", "COND_ITEM", "REF_ITEM", "FIELD_STD_ITEM",
			"FIELD_VARIANCE_ITEM", "INSERT_VALUE_ITEM", "SUBSELECT_ITEM", "ROW_ITEM",
			"CACHE_ITEM", "TYPE_HOLDER", "PARAM_ITEM", "TRIGGER_FIELD_ITEM", "DECIMAL_ITEM",
			"XPATH_NODESET", "XPATH_NODESET_CMP", "VIEW_FIXER_ITEM" };
	static const char* sumNameOf[] =  {
	    "COUNT_FUNC", "COUNT_DISTINCT_FUNC", "SUM_FUNC", "SUM_DISTINCT_FUNC", "AVG_FUNC",
	    "AVG_DISTINCT_FUNC", "MIN_FUNC", "MAX_FUNC", "STD_FUNC",
	    "VARIANCE_FUNC", "SUM_BIT_FUNC", "UDF_SUM_FUNC", "GROUP_CONCAT_FUNC" };


	for(int i = 0; i <= indent; i++)
		fprintf(stderr, "*");
	fprintf(stderr, " ");
	indent += 1;

	if(!item) {
		fprintf(stderr, "NULL item\n");
		return;
	}

	const char* name;
    Item::Type type = item->type();

	if((int)type < sizeof(nameOf)/sizeof(*nameOf))
		name = nameOf[type];
	else name = "BHFIELD_ITEM";

	const char* result = "<unknown result type>";
    switch (item->result_type()) {
		case STRING_RESULT: result = "STRING_RESULT"; break;
		case INT_RESULT: result = "INT_RESULT"; break;
		case REAL_RESULT: result = "REAL_RESULT"; break;
		case DECIMAL_RESULT: result = "DECIMAL_RESULT"; break;
    }
    fprintf(stderr, "%s %s @%p %s %s max_length=%d",
    		name, item->full_name(),  (void*)item,
    		fieldType(item->field_type()), result, item->max_length);

    if(item->result_type() == DECIMAL_RESULT)
    	fprintf(stderr, " [prec=%d,dec=%d,int=%d]",
    			item->decimal_precision(), (int)item->decimals, item->decimal_int_part());

    switch(type) {
		case Item::FUNC_ITEM:
		{
			Item_func* func = static_cast <Item_func*> (item);

			fprintf(stderr, " func_name=%s\n", func->func_name());
			String str;
			// GA, print function takes extra argument but do not use it in the base class.
			func->print(&str,QT_ORDINARY);
			fprintf(stderr, " f contents: %s\n", str.c_ptr_safe());

			Item** args = func->arguments();
			for(uint i = 0; i < func->argument_count(); i++)
				printItemTree(args[i], indent);
			return;
		}
		case Item::COND_ITEM:
		{
			Item_cond* cond = static_cast <Item_cond*> (item);

			fprintf(stderr, " func_name=%s\n", cond->func_name());

			List_iterator<Item> li(*cond->argument_list());
			Item* arg;
			while ((arg = li++))
				printItemTree(arg, indent);
			return;
		}
		case Item::SUM_FUNC_ITEM:
		{
			Item_sum* sum_func = static_cast <Item_sum*> (item);

			uint index = sum_func->sum_func();
			const char* sumname;
			if (index >= sizeof(sumNameOf)/sizeof(*sumNameOf))
				sumname = "<UNKNOWN>";
			else
				sumname = sumNameOf[index];

			fprintf(stderr, "  %s\n", sumname);

			Item** args = sum_func->args;
			uint args_count = sum_func->arg_count;
			for(uint i=0; i < args_count; i++)
				printItemTree(args[i], indent);
			return;
		}
		case Item::REF_ITEM:
		{
			Item_ref* ref = static_cast<Item_ref*> (item);
			Item* real = ref->real_item();
			fprintf(stderr, "\n");
			if(ref != real) printItemTree(real, indent);
			return;
		}
		case Item::INT_ITEM:
		{
			Item_int_with_ref* int_ref = dynamic_cast<Item_int_with_ref*>(item);
			if(!int_ref)
				break;
			// else item is an instance of Item_int_with_ref, not Item_int
			fprintf(stderr, " [Item_int_with_ref]\n");
			printItemTree(int_ref->real_item(), indent);
			return;
		}
		case Item::SUBSELECT_ITEM:
			Item_subselect* ss = dynamic_cast<Item_subselect*>(item);
			fprintf(stderr, " subselect type %d\n", ss->substype());
			String str;
			// GA, print function takes extra argument but do not use it.
			ss->get_unit()->print(&str, QT_ORDINARY);
			fprintf(stderr, "%s\n", str.c_ptr_safe());
    }

    fprintf(stderr, "\n");
}

void printItemTree(char const* info, Item* item)
{
	///  Comment out this line to switch printing on  ///
	return;
	///  Comment out this line to switch printing on  ///

	fprintf(stderr, "%s\n", info);
	printItemTree(item);
}

