/* Copyright (C) 2000 MySQL AB, portions Copyright (C) 2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "common/bhassert.h"
#include "export2file.h"
#include "shared/core/RCEngine.h"

void RCEngine::ExportResults(THD *thd, TempTable& results, select_bh_export *res, List<Item> &fields, const char *filename, int is_dumpfile)
{
	IOParameters iop;
	BHError bherror;

	res->send_fields(fields, Protocol::SEND_NUM_ROWS | Protocol::SEND_EOF);

	if((bherror = RCEngine::GetIOParameters(iop, *thd, *res->SqlExchange())) != BHERROR_SUCCESS)
		throw LD_Failed;

	List_iterator_fast<Item> li(fields);
	fields_t f;
	li.rewind();
  Item* item( NULL );
	while ((item=li++)) {
		fields_t::value_type t( item->field_type() );
		f.push_back( t );
	}

	_int64 no_save_rec = results.Export(iop, f);

	res->SetRowCount((ha_rows)no_save_rec);
	res->SendOk(thd);
	return;
};


select_bh_export::select_bh_export(sql_exchange *ex)
	: select_export(ex)
{
}

select_bh_export::select_bh_export(select_export* se)
	: select_export(se->get_sql_exchange()), se(se)
{
}

select_bh_export::~select_bh_export()
{
	//delete se;
}

int select_bh_export::prepare(List<Item> &list, SELECT_LEX_UNIT *u)
{
	bool blob_flag=0;
	unit= u;
	{
		List_iterator_fast<Item> li(list);
		Item *item;
		while ((item=li++))
		{
			if (item->max_length >= MAX_BLOB_WIDTH)
			{
				blob_flag=1;
				break;
			}
		}
	}
	field_term_length = exchange->field_term->length();
	if (!exchange->line_term->length())
		exchange->line_term=exchange->field_term;	// Use this if it exists
	field_sep_char= (exchange->enclosed->length() ? (*exchange->enclosed)[0] :
		field_term_length ? (*exchange->field_term)[0] : INT_MAX);
	escape_char=	(exchange->escaped->length() ? (*exchange->escaped)[0] : -1);
	line_sep_char= (exchange->line_term->length() ?
		(*exchange->line_term)[0] : INT_MAX);
	if (!field_term_length)
		exchange->opt_enclosed=0;
	if (!exchange->enclosed->length())
		exchange->opt_enclosed=1;			// A little quicker loop
	fixed_row_size= (!field_term_length && !exchange->enclosed->length() &&
		!blob_flag);
	return 0;
}

void select_bh_export::SetRowCount(ha_rows x)
{
	row_count = x;
}

void select_bh_export::SendOk(THD* thd)
{
	// GA
	::my_ok(thd, row_count);
}

sql_exchange* select_bh_export::SqlExchange()
{
	return exchange;
}

mysql_bool select_bh_export::send_data(List<Item> &items)
{
	return se->send_data(items);
}


