/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

//////////////// Ftree - dictionary for strings ////////////////////////////

#ifndef __FTREE__
#define __FTREE__

#include "bintools.h"
#include "../types/RCDataTypes.h"

#include "../system/MemoryManagement/TrackableObject.h"

class FTree
	: public TrackableObject
{
public:

	FTree();
	~FTree();

	void Init(int width);

	RCBString GetRealValue(int v);
	char *GetBuffer(int v);			// the pointer to the memory place of the real value

	int GetEncodedValue(RCBString);	// return -1 when value not found

	int Add(RCBString);				// return -1 when len=0

	int CountOfUniqueValues();
	char *Mem()						// table of (sorted or partially sorted) values
	{ return mem; }
	void SetSize(int s)				// for manual dictionary setting (e.g. loading from disc)
	{ dic_size=s; }

	// Compress/uncompress

	int ByteSize();							// number of bytes required to compress the dictionary
	void SaveData(unsigned char *&buf);
	void Init(unsigned char *&buf);

	int ValueSize()			{return value_size;}
	int ValueSize(int i)	{ return len[i]; }
	int MaxValueSize()		{ return MaxValueSize(0, dic_size - 1); }
	int MaxValueSize(int start, int end);				// max. value size for an interval of codes

	int GetSortPosition(int v);		// get the sorting reverse index of value v (the lowest string v return 0 etc.)
	int GetSortValue(int v);		// get the code on i-th position in sorting index (e.g. GetSortValue(0) is the code of the lowest string)

	TRACKABLEOBJECT_TYPE TrackableType(){return TO_FTREE;}

private:
	char *mem;			// dictionary: dic_size*value_size bytes. Values are NOT terminated by anything and may contain any character (incl. 0)
	unsigned short *len;// table of lengths of values

	void QuickSort(int s1,int s2);
	 int Compare(char *val1, int len1, char *val2, int len2);	// compare two strings, return -1 if val1<val2, 0 if equal, 1 if val1>val2
	int *index;			// index[i] is the value on i-th sorted position (at least "sorted_to" first entries are sorted)
	int sorted_to;
	int *reverse_index;	// reverse_index[v] is the position of value v in the sorted order

	int dic_size;		// number of different values
	int total_dic_size;	// may be more than dic_size, because we reserve some place for new values
	int last_code;		// last code used
	int value_size;		// max. length of one string

	// Compression section - not used yet
	int changed,compressed_size,comp_mode;

};

#endif

