/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _ENGINE_BHITEM_H_
#define _ENGINE_BHITEM_H_

#include <set>
#include <map>

#ifndef __BH_COMMUNITY__
#	include "enterprise/local.h"
#else /* not __BH_COMMUNITY__ */
#	include "community/local.h"
#endif /* __BH_COMMUNITY__ */
#include "types/AValue.h"

namespace na {

class BHItem;
class DataSource;


// Type of a functor invoked for each item of a tree, inside Walk().
// A functor returns 1 to indicate that walking through the tree should be halted
// (e.g. we are looking for one occurence of smthng and it is found).
typedef int (BHItem::*ItemProcess) (void* arg);


// Corresponds to class Item from MySQL. Base class for BHQuery, Expression and Condition classes.
class BHItem
{
public:
	virtual ~BHItem()		{}
	virtual bool IsExpression()		{ return false; }
	virtual bool IsCondition()		{ return false; }
	virtual bool IsBHQuery()		{ return false; }

	// Generic method for walking through the whole tree, in POSTORDER. As in MySQL.
	virtual int Walk(ItemProcess proc, void* par, bool subquery)	{ return (this->*proc)(par); }

protected:
	virtual int Proc_LinkVarsToDS(void*)				{ return 0; }
	virtual int Proc_GetVariables(void* vars)			{ return 0; }
	virtual int Proc_SetBuffers(void* bufs)				{ return 0; }
};


struct Variable
{
	DataSource* srctab;		// DataSource which represents source or derived table, from which a variable comes; may be NULL (when used in old engine)
	JustATable* bhtab;		// necessary for the old engine
	int index;

	bool operator< (const Variable& cd) const			{ return srctab<cd.srctab || (srctab==cd.srctab && (bhtab<cd.bhtab || (bhtab==cd.bhtab && index<cd.index))); }
	Variable(DataSource* ds, JustATable* bh, int i)		:srctab(ds), bhtab(bh), index(i) {}
	Variable()											:srctab(NULL), bhtab(NULL), index(-1) {}
};


typedef std::set<Variable>			SetOfVars;
typedef std::map<Variable,AValue*>	BufOfVars;

// Base class for Expression and Condition classes
class BHExprCond : public BHItem
{
public:

	// Translate linkage of variables from TableDescriptor to DataSource link.
	void LinkVarsToDS()						{ Walk(&BHExprCond::Proc_LinkVarsToDS, NULL, false); }

	// Add to 'vars' all column descriptors (variables, ExprVar) occuring in the subtree rooted in this node
	void GetVars(SetOfVars* vars)			{ Walk(&BHExprCond::Proc_GetVariables, vars, false); }

	// Tell each ExprVar in the subtree where is the buffer for a given variable.
	// Allocate buffers for intermediate results.
	void SetBufs(BufOfVars* bufs)			{ Walk(&BHExprCond::Proc_SetBuffers, bufs, false); }
};


}
#endif

