/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _ENGINE_DATASOURCE_H_
#define _ENGINE_DATASOURCE_H_

#include <vector>
#include "Expression.h"
#include "Condition.h"
namespace na {


class DataSource
{
public:

	// Data ITERATOR interface //
	virtual void Open()			{}
	virtual bool GetNext()		{ return false; }		// return false if there are no more records
	virtual void Close()		{}

	// Allocate buffers to store output variables; return them in the 'bufs' map.
	virtual void SetBuffers(BufOfVars* bufs)	{}

	virtual ~DataSource()		{}		// Caution: destructors of DataSource subclasses will be deleting whole subtree rooted in a given DS...
};


class DSTable : public DataSource
{
public:
	virtual void Open();
	virtual bool GetNext();
	virtual void Close();

	virtual void SetBuffers(BufOfVars* bufs);

	DSTable(JustATable* t) :
		tab(dynamic_cast<RCTable*>(t)), current(0), packlock(-1)
	{
		BHASSERT(tab, "'tab' should not be null");
	}
	virtual ~DSTable()
		{
		//TODO test this part
		long vcounts = vals.size();
		long index;
		for (index=0; index<vcounts; index++)
		{
			delete vals[index];
		}
		//for each(AValue* v in vals) delete v;
		}

private:
	AValue* CreateBuf(int col);				// allocate buffer of appropriate type for column 'col' of RCTable

	RCTable*				tab;
	std::vector<int>		columns;		// which columns of source table are used (maybe none)
	std::vector<AValue*>	vals;

	uint64					current;		// number of rows already read (or index of the next row to read)
	uint64					total;			// total no. of rows
	int64					packlock;		// index of pack being locked or -1
};

class DSSelect : public DataSource
{
public:
	virtual void Open()				{ src->Open(); }
	virtual bool GetNext()
		{
			if(src->GetNext())
			{
				//TODO test this part
				long ecounts = expr.size();
				long index;
				for (index=0; index<ecounts; index++)
					expr[index]->Evaluate();
				//for each(Expression* e in expr)
					//e->Evaluate();
				return true;
			}
			return false;
		}
	virtual void Close()			{ src->Close(); }

	void GetResultBufs(std::vector<AValue*>& bufs);		// give buffers where result rows will be stored
	void AddExpr(Expression* e)							{ expr.push_back(e); }	// from now on, DSSelect is responsible for deallocation of 'e'

	virtual void SetBuffers(BufOfVars* bufs);

	DSSelect(DataSource* s)			:src(s) {}
	virtual ~DSSelect()
		{
			delete src;
			//TODO test this part
			long ecounts = expr.size();
			long index;
			for (index=0; index<ecounts; index++)
				delete expr[index];
			//for each(Expression* e in expr)
			//	delete e;
		}

private:
	DataSource*					src;
	std::vector<Expression*>	expr;
};


}	// namespace na
#endif


