/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include <boost/assign/std/vector.hpp>

#ifndef __BH_COMMUNITY__
#	include "enterprise/local.h"
#else /* not __BH_COMMUNITY__ */
#	include "community/local.h"
#endif /* __BH_COMMUNITY__ */
#include "system/Channel.h"
#include "system/IBFileSystem.h"
#include "core/RCAttr.h"
#include "RCTableImpl.h"
#include "RCEngine.h"
#include "shared/handler/BrighthouseHandler.h"

using namespace std;
using namespace boost;
using namespace boost::assign;

extern const char *ha_rcbase_exts[];

RCTableImpl::RCTableImpl(const char *a_path, vector<DTCollation> charsets, char *nm, int t_n, uint s_id, int use_path_only) :
	waiting_for_drop(0), charsets(charsets), m_state(0)
{
	assert(t_n >= 0);
	check_integrity = true;
	tab_num = t_n;
	no_attr = 0;
	int strl = (int)strlen(nm) + 1;
	name = new char [strl];
	desc = NULL;
	strcpy(name, nm);
	a = NULL;
	path = path_no_ext = NULL;
	// current path (terminated by '/'), determining the directory where Table.ctb and TAxxxxx.ctb may be found
	if(a_path)
		SetPath(a_path, use_path_only);
	else
		SetPath("./");
	session = s_id;
	conn_mode = -1;

#ifdef __GNUC__
	//pthread_mutexattr_t mattr;
	//pthread_mutexattr_init(&mattr);
	//pthread_mutexattr_settype(&mattr, PTHREAD_MUTEX_RECURSIVE);
	//pthread_mutex_init(&synchr, &mattr);
#else
	//InitializeCriticalSection(&this->synchr);
#endif
}

/////////////////////////////////////////////////////////////////////////////////////////////////
// Table description file format ("Table.ctb"):
//
//	"RScT"				4		- file identifier
//	<format_id>			1		- "0", "1" - obsolete formats, "2" - current
//	<security>			4		- reserved for security info
//	<no_all_obj>		8		- reserved for total no. of objects (including deleted); currently not used
//	<no_cur_obj>		8		- reserved for current (excluding deleted) no. of objects; currently not used
//	<no_attr>			4		- number of attributes
//	<block_offset>		4		- offset of the first special block (0 means no such block)
//
//	<attr_mode>			no_attr	- one-byte attribute parameter; 0 for deleted attribute, !=0 otherwise
//	<tab_name>			...		- table name, null-terminated
//	<tab_desc>			...		- table description, null-terminated
//
//	<block_1>			...		- special blocks
//	<block_2>			...
//	...
//	<block_n>			...
//
//	general sctructure of a block:		<blk_len>	4	- including 6 bytes of header
//										<blk_type>	1	- reserved (0 is assumed as interpretable now)
//										<is_next>	1	- 0 for the last block, 1 otherwise
//
//
RCTableImpl::RCTableImpl(const char *a_path, vector<DTCollation> cs, int t_n, int current_state, int use_path_only, bool force_open, bool for_loader)
	:	waiting_for_drop(0), charsets(cs), m_state(current_state)
{
	check_integrity = !force_open;
	conn_mode = 0;
	session = 0;
	tab_num=t_n; // index of this table in database (-1 means we should load it from file)
	no_attr=0;
	a=NULL;
	name=NULL;
	path=path_no_ext=NULL;
	desc=NULL;

	// set path to the table directory
	int i;
	char buf[8192]= {0};
	if(a_path)
		SetPath(a_path, use_path_only);
	else
		SetPath("./");
	strcpy(buf, path);

/*********************************************
	struct _stat dir_info;
	if(_stat(buf, &dir_info)!=0) {
		rclog << "Error: " << buf << " directory status can not be obtained" << endl;
		int err;
#ifdef __GNUC__
		err = errno;
#else
		_get_errno(&err);
#endif
		if(err==ENOENT) {
			delete[] path;
			throw NoTableFolderRCException();
		}
	} else
		m_create_time = dir_info.st_ctime;
************************************************/
    int res = GetFileCreateTime(buf);
    if (res < 0)
    	throw NoTableFolderRCException(); // no need to call delete[] path
    m_create_time = (time_t) res;

	// open table definition file ("...<base>.../D000000000/Table.ctb")
	//strcpy_s(buf, 8192, path);

    strcat(buf, "/Table.ctb");
	IBFile ftbl;

	try
	{
		ftbl.OpenReadOnly(buf);

		int block_offset=0;
		ftbl.Read(buf, 8192); // warning C6031: Return value ignored: '_read'
		char *b=(char*)buf;
		// header (33 bytes):
		if(strncmp(b,"RScT",4)!=0 || b[4]!='2')
		{
			rclog << lock << "Error: unrecognized format of table definition." << unlock;
			return;
		}
		no_attr=*((int*)(b+25));
		block_offset=*((int*)(b+29));

		// attribute parameters table
		if(no_attr>0)
		{
			a=new RCAttr * [no_attr];
			if(!a) rclog << lock << "Error: out of memory (" << no_attr << " bytes failed). (43)" << unlock;
			//memcpy(a_mode,b+33,no_attr);	// copy the vector of attributes' parameters
		}

		// table name and description
		b+=33+no_attr;
		i=(int)strlen(b)+1;
		name=new char [i];
		if(!name) rclog << lock << "Error: out of memory (" << i << " bytes failed). (44)" << unlock;
		//strcpy_s(name, i, b);
		strcpy(name, b);
		b+=i;
		i=(int)strlen(b)+1;
		desc=new char [i];
		if(!desc) rclog << lock << "Error: out of memory (" << i << " bytes failed). (45)" << unlock;
		//strcpy_s(desc, i, b);
		strcpy(desc, b);

		// blocks
		while(block_offset>0)
		{
			b=(char*)buf+block_offset; // block offset from the beginning of the file
			int b_len=*(int*)b;
			uchar b_type=*(uchar*)(b+4);
			uchar b_next=*(uchar*)(b+5);

			// block types
			if(b_type==1) // MySQL special block: internal table number
			{
				//				assert(tab_num==-1);
				tab_num = *(int*)(b+6);
			}
			else if(b_type==66)
			{
				// place for block interpretation
			}
			else
			{
				rclog << lock << "Warning: special block (unknown type " << int(b_type) << ") ignored in table " << name << unlock;
			}
			if(b_next==0) block_offset=0;
			else block_offset+=b_len;
		}
		ftbl.Close();
	}
	catch(DatabaseRCException& e)
	{
		if (a)
		delete [] a;
		if (name)
		delete [] name;
		if (desc)
		delete [] desc;

		no_attr = 0;
		throw e;
	}
	//rcbuffer.BufClose();			// important: attributes will use the same buffer!
	if(tab_num<0) {
		rclog << lock << "Error: no table number found. File format may be obsolete!" << unlock;
		tab_num = 0;
	}

	// load definitions of non-deleted attributes
	for(i=0; i<no_attr; i++)
		a[i]=NULL;

	if(charsets.size() == 0)
		charsets.assign(no_attr, DTCollation());

	if(for_loader)
		return;

#ifdef __GNUC__
	//pthread_mutex_init(&synchr, NULL);
#else
	//InitializeCriticalSection(&synchr);
#endif

	if(no_attr > 0) {

		try
		{
			LockPackInfoForUse();		// refresh all column information, check metadata integrity
		}
		catch(DatabaseRCException& e)
		{
			CleanUp();
			throw e;
		}

		m_update_time = a[0]->UpdateTime();
	} else
		m_update_time = m_create_time;

}

void RCTableImpl::CleanUp()
{
#ifdef __GNUC__
	//pthread_mutex_destroy(&synchr);
#else
	//DeleteCriticalSection(&synchr);
#endif
	for(int i=0; i < no_attr; i++)
		delete a[i];
	delete [] a;
	delete [] name;
	delete [] path;
	delete [] path_no_ext;
	delete [] db_name;
	delete [] desc;
}

RCTableImpl::~RCTableImpl()
{
	CleanUp();
}

void RCTableImpl::SetPath(const char *a_path, int use_path_only)
{
	delete [] path; // path="...<a_path>.../D000000000/"
	int l = strlen(a_path) + 1;
	path = new char [l];
	strcpy(path, a_path);
	delete [] path_no_ext;
	l = strlen(a_path) - strlen(ha_rcbase_exts[0]) + 1;
	path_no_ext = new char [l];
	memcpy(path_no_ext, a_path, l);
	path_no_ext[l - 1] = 0;

	char* end = path_no_ext + strlen(path_no_ext);
	while(end >= path_no_ext && *end != '/' && *end != '\\')
		end--;
	char* beg = end - 1;
	while(beg >= path_no_ext && *beg != '/' && *beg != '\\') {
		beg--;
	}
	beg++;
	db_name = new char[end - beg + 1];
	memcpy(db_name, beg, end - beg);
	db_name[end - beg] = '\0';

//	int p_size=p_len+12;
//	if(a_path[p_len-1]!='/' && a_path[p_len-1]!='\\')
//		p_size++; // add '/' if not present
//	path=new char [p_size];
//	//strcpy_s(path, p_size, a_path);
//	strcpy(path, a_path);
//	if(a_path[p_len-1]!='/' && a_path[p_len-1]!='\\')
//		strcat(path, "/");
//	if(!use_path_only) {
//		//strcat_s(path, p_size, "D");
//		strcat(path, "D");
//		char bufnr[10];
//		int curnr=tab_num;
//		for(int i=8; i>=0; i--) {
//			bufnr[i]=char('0'+curnr%10);
//			curnr/=10;
//		}
//		bufnr[9]='\0';
//		/*strcat_s(path, p_size, bufnr);
//		 strcat_s(path, p_size, "/");*/
//		strcat(path, bufnr);
//		strcat(path, "/");
//	}
}

void RCTableImpl::Refresh(int current_state)
{
	if(m_state != current_state) {
		for(int i=0; i<no_attr; i++) {
			delete a[i];
			a[i] = 0;
		}
		m_state = current_state;
	}
}

void RCTableImpl::LockPackInfoForUse()
{
	for(int att=0; att<no_attr; att++) {
		LoadAttribute(att);
	}
}

void RCTableImpl::UnlockPackInfoFromUse()
{
}

void RCTableImpl::LockPackForUse(unsigned attr, unsigned pack_no, ConnectionInfo& conn)
{
	if( pack_no == 0xFFFFFFFF )		// "not used" pack number
		return;
	LoadAttribute(attr);			// TODO: some kind of locking between these two lines? (But performance...)
	a[attr]->LockPackForUse(pack_no, conn);
}
void RCTableImpl::UnlockPackFromUse(unsigned attr, unsigned pack_no)
{
	if( pack_no == 0xFFFFFFFF )
		return;
	LoadAttribute(attr);
	a[attr]->UnlockPackFromUse(pack_no);
}

void RCTableImpl::AddAttribute(char* nm, AttributeType a_type, int a_field_size, int a_dec_places /*= 0*/, unsigned int param /*= 0*/, DTCollation collation /*= DTCollation()*/)
{
	no_attr++;
	RCAttr** a_new = new RCAttr * [no_attr]; // warning C6211: Leaking memory 'a_new' due to an exception.
	for(int i = 0; i < no_attr - 1; i++)
		a_new[i] = a[i];
	delete [] a;
	a = a_new;
	a[no_attr - 1] = new RCAttr(nm, no_attr - 1, tab_num, a_type, a_field_size, a_dec_places, path, param, session, collation);
}

void RCTableImpl::SaveHeader()
{
	char buf[8192] = {0};
//	mkdir(path, S_IRWXU);
    try {
    	if(IsAbsolutePath(path))
    		CreateDir( path ); //temporary table
    	else
    		CreateDir( infobright_data_dir + path );
    } catch(DatabaseRCException& e) {
        rclog << lock << e.what() << " DatabaseRCException" << unlock;
        throw;
    }

	//strcpy_s(buf, 8192, path);
	strcpy(buf, path);
	//buf[strlen(buf)-1] = 0;

    /**************************************
	struct _stat dir_info;
	if(_stat(buf, &dir_info)!=0)
		rclog << "Error: " << buf << " directory status can not be obtained." << endl;
	else
		m_create_time = dir_info.st_ctime;
    ***************************************/
    int res = GetFileCreateTime(buf);
    if (res < 0) {
        rclog << "Error: " << buf << " directory status can not be obtained." << endl;
    } else {
        m_create_time = (time_t) res;
    }

	/*strcpy_s(buf, 8192, path);
	 strcat_s(buf, 8192, "Table.ctb");*/
	strcpy(buf, path);
	strcat(buf, "/");
	strcat(buf, "Table.ctb");

	int i;
	char* b = 0;
	IBFile ftbl;
	ftbl.OpenCreateEmpty(buf);
	b = (char*)buf;
	memset(b, 0, 33);
	//strcpy_s(b, (8192 - (int)(b-buf)), "RScT2");
	strcpy(b, "RScT2");
	*(uint*)(b+5)=0; //	<security>, for the future use. 0 means no security encoding.
	_int64 no_obj=0;
	for(i=0; i<no_attr; i++)
		if(a[i]) {
			no_obj=a[i]->NoObj();
			break;
		}
	*(_int64*)(b+9)=no_obj;
	*(int*)(b+25)=no_attr;
	*(int*)(b+29)=0; // block offset
	b+=33;

	// attr. parameters
	//memcpy(b,a_mode,no_attr);
	memset(b, 1, no_attr);
	b+=no_attr;

	// name and description
	//strcpy_s(b, (8192 - (int)(b-buf)), name);
	strcpy(b, name);
	b+=strlen(name)+1;
	if(desc) {
		//strcpy_s(b, (8192 - (uint)(b-buf)), desc);
		strcpy(b, desc);
		b+=strlen(desc)+1;
	} else {
		*b='\0';
		b++;
	}
	//////////////// special blocks ////////////////////

	*(int*)(buf+29)=int(b-buf); // offset of the first special block (from the beginning of the file)

	// MySQL special block: internal table number

	*(int*)b = 10; // length: 6 for header, 4 for table number
	*(b+4) = 1; // block type
	*(b+5) = 0; // next?
	*(int*)(b+6) = tab_num;
	b+=10;

	////////////////////////////////////////////////////
	ftbl.Write(buf, (uint)(b - buf));
	ftbl.Close();
}

void RCTableImpl::Save()
{
	SaveHeader();
	for(int i=0; i<no_attr; i++)
		if(a[i])
			a[i]->SaveHeader();
	if(no_attr > 0) {
		LoadAttribute(0);
		m_update_time = a[0]->UpdateTime();
	} else
		m_update_time = m_create_time;
}

//void RCTableImpl::Drop()
//{
//	if(rsi_manager) {
//		rsi_manager->DeleteIndex(tab_num); // delete all indexes connected with this table
//	}
//	for(int i = 0; i < no_attr; i++) {
//		delete a[i];
//		a[i] = 0;
//	}
//	DeleteDirectory(path);
//}

void RCTableImpl::Drop()
{
	if(rsi_manager) {
		for(int j = 0; j < no_attr; j++) {
			rsi_manager->DeleteIndex(tab_num, j); // delete all indexes connected with this table
		}
	} else {
		// KNs may have been created when KNLevel was > 0, they cannot stay
		InitRSIManager(infobright_data_dir, true, true);
		for(int j = 0; j < no_attr; j++)
			rsi_manager->DeleteIndex(tab_num, j); // delete all indexes connected with this table
		delete rsi_manager;
		rsi_manager = NULL;
	}


	for(int i = 0; i < no_attr; i++) {
		delete a[i];
		a[i] = 0;
	}
	DeleteDirectory(path);
}

void RCTableImpl::LoadAttribute(int attr_no)
{
	if(!a[attr_no]) {
		//pthread_mutex_lock(&synchr);
        table_impl_mutex.Lock();
		if(!a[attr_no]) {
			try {	
				a[attr_no] = new RCAttr(attr_no, tab_num, path, conn_mode, session, charsets[attr_no]);
				if(a[attr_no]->OldFormat()) {
					a[attr_no]->UpgradeFormat();
					delete a[attr_no];
					a[attr_no] = new RCAttr(attr_no, tab_num, path, conn_mode, session, charsets[attr_no]);
				}
				// Checking consistency
				if(process_type != ProcessType::DIM && check_integrity) {
					for(int i = 0; i < no_attr; i++)
						if(i != attr_no && a[i]) {
							if(a[i]->NoObj() != a[attr_no]->NoObj()) {
								rccontrol.lock(-1) << "Data integrity error." << unlock;
								throw DatabaseRCException("Data integrity error.");
							}
							break;
						}
				}
				// next, needs to extract character set form MySql structures
				if(ATI::IsTxtType(a[attr_no]->TypeName()) && ConnectionInfoOnTLS.IsValid()) {
					if(attr_no < charsets.size())
						a[attr_no]->SetCollation(charsets[attr_no]);
					//THD& thd = ConnectionInfoOnTLS.Get().Thd();
					//TABLE_LIST* tables = thd.lex->query_tables;
					//if(tables && thd.killed != THD::KILL_CONNECTION) {
					//	while(tables) {
					//		if(tables->schema_table && tables->schema_table_name) {
					//			if(strcmp(tables->schema_table_name, name) == 0 && strcmp(tables->db, db_name) == 0)
					//				break;
					//		} else if(tables->table_name && strcmp(tables->table_name, name) == 0 && strcmp(tables->db, db_name) == 0)
					//			break;
					//		else if(tables->table && tables->table->file) {
					//			char* p = static_cast<BrighthouseHandler*>(tables->table->file)->GetTablePath();
					//			if(strstr(p, name))
					//				break;
					//		}
					//		tables = tables->next_global;
					//	}
					//	TABLE_SHARE* share = NULL;
					//	if(tables)
					//		share = get_cached_table_share(tables->db, name);
					//	//int err;
					//	//char key[NAME_LEN*2+2];
					//	//uint key_length = create_table_def_key(&thd, key, tables, 1);
					//	//TABLE_SHARE* share = get_table_share(&thd, tables, key, key_length, OPEN_VIEW, &err);
					//	if(share) {
					//		AttributeTypeInfo ati = RCEngine::GetAttrTypeInfo(*(share->field[attr_no]));
					//		a[attr_no]->SetCollation(ati.GetCollation());
					//	}  
					//}
				}
			} catch(DatabaseRCException& e) {
                table_impl_mutex.Unlock();
				//pthread_mutex_unlock(&synchr);
				throw e;
			}
		}
        table_impl_mutex.Unlock();
		//pthread_mutex_unlock(&synchr);
	}
}

AttrInfo* RCTableImpl::GetAttributes()
{
	AttrInfo* info = new AttrInfo[(int)NoAttrs()];
	uint s_id=0;
	Verify();
	for(int j=0; j < (int)NoAttrs(); j++) {
		LoadAttribute(j);
		info[j].name = AttrName(j);
		info[j].desc = a[j]->Description();
		info[j].type = a[j]->TypeName();
		info[j].size = a[j]->Type().GetPrecision();
		info[j].precision = a[j]->Type().GetScale();
		if(a[j]->Type().GetNullsMode() == NO_NULLS)
			info[j].no_nulls = true;
		else
			info[j].no_nulls = false;
		info[j].lookup = a[j]->Type().IsLookup();
		//info[j].unique = a[j]->IsDeclaredUnique();
		info[j].actually_unique = (a[j]->PhysicalColumn::IsDistinct()==RS_ALL);
		//info[j].primary = a[j]->IsPrimary();
		info[j].current_trans = a[j]->GetSaveSessionId();
		if(info[j].current_trans==0xFFFFFFFF)
			info[j].current_trans = a[j]->GetSessionId();
		info[j].uncomp_size=a[j]->NaturalSize();
		info[j].comp_size=a[j]->CompressedSize();
	}
	return info;
}

std::vector<ATI> RCTableImpl::GetATIs()
{
	vector<ATI> deas;
	for(uint j = 0; j < NoAttrs(); j++) {
		LoadAttribute(j);
		deas += ATI(a[j]->TypeName(), a[j]->Type().GetNullsMode() == NO_NULLS, a[j]->Type().GetPrecision(), a[j]->Type().GetScale(), false, false, a[j]->Type().GetCollation());
	}
	return deas;
}

int RCTableImpl::AttrNum(const char *a_name) // return the number of an attribute (given by name) or -1 if attribute not found
{
	if(a_name==NULL)
		return -1;
	int res;
	for(res=0; res<no_attr; res++) {
		LoadAttribute(res);
		if(a[res] && strcasecmp(a_name,a[res]->Name())==0)
			return res;
	}
	return -1;
}

const char* RCTableImpl::AttrName(int n_a)
{
	LoadAttribute(n_a);
	return a[n_a]->Name();
}
AttributeType RCTableImpl::AttrType(int n_a)
{
	LoadAttribute(n_a);
	return a[n_a]->TypeName();
}

int RCTableImpl::GetAttrSize(int n_a)
{
	LoadAttribute(n_a);
	return a[n_a]->Type().GetDisplaySize();
}

int RCTableImpl::GetFieldSize(int n_a)
{
	LoadAttribute(n_a);
	return a[n_a]->Type().GetPrecision();
}

int RCTableImpl::GetInternalSize(int n_a)
{
	LoadAttribute(n_a);
	return a[n_a]->Type().GetInternalSize();
}

NullMode RCTableImpl::GetAttrNullMode(int n_a)
{
	LoadAttribute(n_a);
	return a[n_a]->Type().GetNullsMode();
}

int RCTableImpl::GetAttrScale(int n_a)
{
	LoadAttribute(n_a);
	return a[n_a]->Type().GetScale();
}

bool RCTableImpl::IsLookup(int n_a)
{
	LoadAttribute(n_a);
	return a[n_a]->Type().IsLookup();
}

const ColumnType& RCTableImpl::GetColumnType(int n_a)
{
	LoadAttribute(n_a);
	return a[n_a]->Type();
}


RCAttr* RCTableImpl::GetAttr(int n_a)
{
	LoadAttribute(n_a);
	return a[n_a];
}

int RCTableImpl::Verify()
{
	int res=1;
	uint s_id=0xFFFFFFFF;
	int i;
	_int64 n_o = 0xFFFFFFFF;
	for(i=0; i<no_attr; i++)
		if(a[i]) {
			LoadAttribute(i);
			if(n_o==0xFFFFFFFF)
				n_o=a[i]->NoObj();
			if(s_id==0xFFFFFFFF)
				s_id=a[i]->GetSessionId();
			if(n_o!=a[i]->NoObj())
				res=0;
			if(s_id!=a[i]->GetSessionId())
				res=0;
		}
	if(res==0) {
		rclog << lock << "Error:  columns in table " << name << " are inconsistent:" << unlock;
		for(i=0; i<no_attr; i++)
			if(a[i]) {
				rclog << "  - " << a[i]->Name() << " \t no.obj.= " << a[i]->NoObj() << " \t sess.id= "
						<< a[i]->GetSessionId() << endl;
			}
		rclog << unlock;
	}
	return res;
}
bool RCTableImpl::LockedBy(uint currentTransaction)
{
	bool res = false;

	//if( !pthread_mutex_trylock(&synchr) ) {
	if( table_impl_mutex.TryLock() ) {
		if(a[0])
			if(GetSaveSessionId()==currentTransaction)
				res = true;
        table_impl_mutex.Unlock();
		//pthread_mutex_unlock(&synchr);
	}
	return res;
}

uint RCTableImpl::CommitSaveSession(uint s_id, int new_state)
{
	uint result=INVALID_TRANSACTION_ID;

	//////////////// Commit data ////////////////////

    _int64 row_count = -1;
	for(int i=0; i<no_attr; i++) {
		LoadAttribute(i);
		if(a[i]) {
			result=a[i]->GetSaveSessionId();
			if(result==INVALID_TRANSACTION_ID || (s_id!=0 && s_id!=result))
				return INVALID_TRANSACTION_ID; // result;  AB Switch change
		}

        // Check data integrity
//        if (i == 0) {
//            row_count = a[i]->NoObj();
//        } else {
//            if (row_count != a[i]->NoObj()) {
//			    rccontrol.lock(-1) << "RCTableImpl::CommitSaveSession  Error in data integrity" << unlock;
//    			throw DatabaseRCException("RCTableImpl::CommitSaveSession Error in data integrity");
//            }
//        }
	}

	for(int i=0; i<no_attr; i++)
		if(a[i])
			a[i]->CommitSaveSession();

	//////////////// Finishing ////////////////////

    if (result != INVALID_TRANSACTION_ID) {
        // If there were changes in columns than session id is not equal to INVALID_TRANSACTION_ID.
        // It means RCAttr created TS* file and renamed it to TA*/TB* file in CommitSaveSession() method above.
        // In this case the system flips AB Switch, which changes "current set" of TA*/TB* files.
        ABSwitcher absw;
        int res = absw.FlipState(path);
        if (res) {
            // throw here
        }
    }

	m_state = new_state;
	conn_mode = 0;
	return result;
}

bool RCTableImpl::Committed()
{
	for(int i = 0; i < no_attr; i++)
		if(a[i] && a[i]->GetSaveSessionId()!=INVALID_TRANSACTION_ID)
			return false;
	return true;
}

uint RCTableImpl::GetSaveSessionId()
{
	uint result = INVALID_TRANSACTION_ID;
	for(int i = 0; i < no_attr; i++) {
		LoadAttribute(i);
		if(a[i])
			return a[i]->GetSaveSessionId();
	}
	return result;
}

uint RCTableImpl::GetLastSaveSessionId()
{
	uint result = INVALID_TRANSACTION_ID;
	for(int i = 0; i < no_attr; i++) {
		LoadAttribute(i);
		if(a[i])
			return a[i]->GetLastSaveSessionId();
	}
	return result;
}

void RCTableImpl::Rollback(uint s_id)
{
	for(int i = 0; i < no_attr; i++) {
		LoadAttribute(i);
		if(a[i]) {
			a[i]->Rollback(s_id);
		}
	}

	SetMode(0, s_id);
	if(no_attr > 0) {
		LoadAttribute(0);
		m_update_time = a[0]->UpdateTime();
	}
}

//void RCTableImpl::RollbackLastCommitted(uint s_id, int new_state)
//{
//	for(int i = 0; i < no_attr; i++)
//	{
//		LoadAttribute(i);
//		if(a[i])
//		a[i]->RollbackLastCommitted();
//	}
//	SetMode(0, s_id);
//	m_state = new_state;
//}

//bool RCTableImpl::IsRLCPossible()
//{
//	for(int i = 0; i < no_attr; i++)
//	{
//		LoadAttribute(i);
//		if(!a[i]->IsRLCPossible())
//			return false;
//	}
//	return true;
//}

/*void RCTableImpl::RollbackSession(int s_id)
 {
 if(no_attr >= 0)
 {
 LoadAttribute(0);
 if(a[0]->GetSaveSessionId() == 0xFFFFFFFF)
 {
 if(a[0]->GetLastSaveSessionId() == (uint)s_id)
 RollbackLastCommitted();
 }
 }
 }*/

void RCTableImpl::SetMode(int conn_mode, uint s_id)
{
	session = s_id;

	this->conn_mode = conn_mode;
	for(int i=0; i<no_attr; i++) {
		delete a[i];
		a[i] = 0;
	}
}

void RCTableImpl::LoadDirtyData(unsigned int session_id)
{
	DTCollation collation;
	for(int i = 0; i < no_attr; i++) {
		collation = DTCollation();
		if(a[i]) {
			collation = a[i]->GetCollation();
			delete a[i];
		}
		a[i] = new RCAttr(i, tab_num, path, 1, session_id, collation);
	}
	if(no_attr > 0)
		m_update_time = a[0]->UpdateTime();
}

void RCTableImpl::DisplayRSI()
{
	rccontrol << lock << "--- RSIndices for " << name << " (tab.no. " << tab_num << ", " << NoPacks() << " packs) ---"
			<< unlock;
	rccontrol << lock << "Name               Triv.packs\tSpan\tHist.dens." << unlock;
	rccontrol << lock << "-----------------------------------------------------------" << unlock;
	for(int i=0; i<no_attr; i++) {
		if(!a[i])
			LoadAttribute(i);
		if(a[i]) {
			double hd;
			double span;
			int trivials;
//			bool repaired = !a[i]->VerifyRSI();
			a[i]->RoughStats(hd, trivials, span);
			rccontrol << lock;
			rccontrol << a[i]->Name();
			int display_limit = int(20) - strlen(a[i]->Name());
			for(int j=0; j<display_limit; j++)
				rccontrol << " ";
			rccontrol << trivials << "\t\t";
			if(span!=-1)
				rccontrol << int(span*10000)/100.0 << "%\t";
			else
				rccontrol << "\t";
			if(hd!=-1)
				rccontrol << int(hd*10000)/100.0 << "%\t";
			else
				rccontrol << "\t";
//			if( repaired )
//				rccontrol << "\t(just recommitted)";
			rccontrol << unlock;
		}
	}
	rccontrol << lock << "-----------------------------------------------------------" << unlock;
	if(rsi_manager) {
		vector<RSIndexID> rslist = rsi_manager->GetAll(tab_num);
		bool any_found = false;
		for(uint i=0; i<rslist.size(); i++)
			if(rslist[i].IsType2()) {
				RSIndex_PP *res = (RSIndex_PP *)(rsi_manager->GetIndex(rslist[i], 0));
				if(res) {
					any_found = true;
					int a1, a2, t2;
					if(rslist[i].tab == tab_num) {
						t2 = rslist[i].tab2;
						a1 = rslist[i].col;
						a2 = rslist[i].col2;
					} else {
						t2 = rslist[i].tab;
						a1 = rslist[i].col2;
						a2 = rslist[i].col;
					}
					rccontrol << lock << "RSI_PP: col. " << a[a1]->Name();
					if(t2 == tab_num)
						rccontrol << " \tself-joined with col. " << a[a2]->Name();
					else
						rccontrol << " \tjoined with tab. " << t2 << ", col. " << a2;
					rccontrol << "\tdens. " << int(res->Density()*10000)/100.0 << "%" << unlock;
					rsi_manager->ReleaseIndex(res);
				}
			}
		if(any_found)
			rccontrol << lock << "----------------------------------------------------------------" << unlock;
	}
}

#ifdef USE_NEW_ENGINE

void RCTableImpl::GetAValue(uint64 obj, int attr, AValue& value, ConnectionInfo *conn)
{
	assert(attr <= no_attr && (int64)obj <= a[attr]->NoObj());
	if(!a[attr])
	LoadAttribute(attr);
	return a[attr]->GetAValue(obj, value, conn);
}

//AValuePtr RCTableImpl::GetMinAValuePtr(int attr, ConnectionInfo *conn)
//{
//	assert(attr <= no_attr);
//	if(!a[attr])
//		LoadAttribute(attr);
//	return a[attr]->GetMinAValuePtr(conn);
//}
//
//AValuePtr RCTableImpl::GetMaxAValuePtr(int attr, ConnectionInfo *conn)
//{
//	assert(attr <= no_attr);
//	if(!a[attr])
//		LoadAttribute(attr);
//	return a[attr]->GetMaxAValuePtr(conn);
//}

void RCTableImpl::GetMinMaxBounds(int attr, VInteger& min, VInteger& max, ConnectionInfo *conn)
{
	assert(attr <= no_attr);
	if(!a[attr])
		LoadAttribute(attr);
	return a[attr]->GetMinMaxBounds(min, max, conn);
}

#endif

template<>
JustATable::RecordIteratorImplementation<JustATable::RcTableLockPolicy>::RecordIteratorImplementation( JustATable::RcTableLockPolicy::initializer_t const& init, _uint64 pos )
	: _data( JustATable::RcTableLockPolicy::data_t( init.table ) ), _currentRNo( pos ), _conn( init.conn )
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( ! _data.table.expired() );
	JustATablePtr t(_data.table);
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(_currentRNo <= t->NoObj() );
	if(init.filter && ( t->NoObj() != 0) && (pos != t->NoObj()) && !init.filter->IsEmpty()) {
		_data.attr_uses = init.attr_uses;
		_data.filter = init.filter;
		_data.it = _data.filter->find( pos, Filter::iterator::TYPE::D_SET);
		if (_data.it != _data.filter->end(Filter::iterator::TYPE::D_SET)) {
			_currentRNo = (*_data.it).GetIndex();
		} else
			_currentRNo = t->NoObj();
	} else
		_currentRNo = t->NoObj();
}

template<>
void JustATable::RecordIteratorImplementation<JustATable::RcTableLockPolicy>::LockPack( _uint64 pos )
{
	JustATable::RcTableLockPolicy::LockPacks( *this, pos );
}

int uuu = 0;
template<>
RCValueObject JustATable::RecordImplementation<JustATable::RecordIteratorImplementation<JustATable::RcTableLockPolicy> >::do_get_item(uint i)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( ! _it._data.table.expired() );
	JustATablePtr t(_it._data.table);
	_it.LockPack(_atRow);
	return RCValueObject( (static_cast<JustATable::RcTableLockPolicy::table_t&>(*t)).GetValue(_atRow, i, _it._conn) );
}

void JustATable::RcTableLockPolicy::LockPacks(iterator_t& it, _uint64 const& obj_no)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( ! it._data.table.expired() );
	JustATablePtr t(it._data.table);
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( ! it._data.locked || ( it._currentRNo == obj_no ) );
	if(!it._data.locked) {
		uint pack_no = JustATable::PackIndex(obj_no);
		for(attr_uses_t::const_iterator use = it._data.attr_uses.begin(); use != it._data.attr_uses.end(); ++use)
			t->LockPackForUse(*use, pack_no, ConnectionInfoOnTLS.Get());		// TODO: should use ConnectionInfo for statistics.
		it._data.locked = true;
	}
}

void JustATable::RcTableLockPolicy::UnlockPacks(iterator_t& it, _uint64 const& obj_no)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( ! it._data.table.expired() );
	JustATablePtr t(it._data.table);
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( it._data.locked );
	uint pack_no = JustATable::PackIndex(obj_no);
	for(attr_uses_t::const_iterator use = it._data.attr_uses.begin(); use != it._data.attr_uses.end(); ++use) {
		t->UnlockPackFromUse(*use, pack_no);
	}
	it._data.locked = false;
}

void JustATable::RcTableLockPolicy::Destroy(iterator_t& it)
{
	if(it._data.locked && !it._data.table.expired())
		UnlockPacks(it, it._currentRNo);
	it._currentRNo = 0;
	it._data.table.reset();
	it._data.filter.reset();
}

JustATable::RcTableLockPolicy::iterator_t& JustATable::RcTableLockPolicy::Assign( JustATable::RcTableLockPolicy::iterator_t& to, JustATable::RcTableLockPolicy::iterator_t const& from )
{
	Destroy(to);
	to._data.table = from._data.table;
	to._currentRNo = from._currentRNo;
	to._conn = from._conn;
	if(from._data.filter.get()) {
		to._data.attr_uses = from._data.attr_uses;
		to._data.filter = from._data.filter;
		to._data.it = from._data.it;
	}
	return to;
}

void JustATable::RcTableLockPolicy::Next(iterator_t& it)
{
	JustATablePtr t(it._data.table);
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( t.get() );
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(it._currentRNo < t->NoObj());
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(it._data.filter->NoObj() == t->NoObj());

	uint64 currently_locked_row = 0;
	if(it._data.locked)
		currently_locked_row = it._currentRNo;
	++it._data.it;
	if(it._data.it == it._data.filter->end(Filter::iterator::TYPE::D_SET)) {
		it._currentRNo = t->NoObj();
		if(it._data.locked)
			UnlockPacks(it, currently_locked_row);
	} else {
		it._currentRNo = (*it._data.it).GetIndex();
		if(it._data.locked && JustATable::PackIndex(currently_locked_row) != JustATable::PackIndex(it._currentRNo))
			UnlockPacks(it, currently_locked_row);
	}

	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(it._currentRNo <= t->NoObj());
}

JustATable::RecordIterator RCTableImpl::do_find(ConnectionInfo* conn, _uint64 pos, JustATable::RcTableLockPolicy::attr_uses_t const& a)
{
	return (JustATable::RecordIterator(
		JustATable::RecordIteratorInterface::iterator_ptr_t(
			new JustATable::RecordIteratorImplementation<JustATable::RcTableLockPolicy>(
				JustATable::RcTableLockPolicy::initializer_t(shared_from_this(), conn,
					FilterPtr(new Filter(NoObj(), true)), a), pos))));
}

JustATable::RecordIterator RCTableImpl::do_find(ConnectionInfo* conn, _uint64 pos, ::Filter const* f, JustATable::RcTableLockPolicy::attr_uses_t const& a)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(f == 0 || f->NoObj() == NoObj());
	return (JustATable::RecordIterator(
		JustATable::RecordIteratorInterface::iterator_ptr_t(
			new JustATable::RecordIteratorImplementation<JustATable::RcTableLockPolicy>(
				JustATable::RcTableLockPolicy::initializer_t(shared_from_this(), conn,
					f ? FilterPtr(new Filter(*f)) : FilterPtr(new Filter(NoObj(), true)), a), pos))));
}

JustATable::RecordIterator RCTableImpl::do_begin(ConnectionInfo* conn, JustATable::RcTableLockPolicy::attr_uses_t const& a)
{
	return (JustATable::RecordIterator(
		JustATable::RecordIteratorInterface::iterator_ptr_t(
			new JustATable::RecordIteratorImplementation<JustATable::RcTableLockPolicy>(
				JustATable::RcTableLockPolicy::initializer_t(shared_from_this(), conn,
					FilterPtr(new Filter(NoObj(), true)), a)))));
}

JustATable::RecordIterator RCTableImpl::do_begin(ConnectionInfo* conn, const ::Filter& filter, JustATable::RcTableLockPolicy::attr_uses_t const& a)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(filter.NoObj() == NoObj());
	return (JustATable::RecordIterator(
		JustATable::RecordIteratorInterface::iterator_ptr_t(
			new JustATable::RecordIteratorImplementation<JustATable::RcTableLockPolicy>(
				JustATable::RcTableLockPolicy::initializer_t(shared_from_this(), conn,
					FilterPtr(new Filter(filter)), a)))));
}

JustATable::RecordIterator RCTableImpl::do_end(ConnectionInfo* conn)
{
	return (JustATable::RecordIterator(
		JustATable::RecordIteratorInterface::iterator_ptr_t(
			new JustATable::RecordIteratorImplementation<JustATable::RcTableLockPolicy>(
				JustATable::RcTableLockPolicy::initializer_t(shared_from_this(), conn,
					FilterPtr(), JustATable::RcTableLockPolicy::attr_uses_t()), NoObj()))));
}

uint JustATable::RcTableLockPolicy::DataTag::Size()
{
	JustATablePtr t(table);
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(t.get());
	return t->NoDisplaybleAttrs();
}

