/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "system/LargeBuffer.h"
#include "system/IOParameters.h"
#include "DataLoader.h"
#include "NewValueSet.h"
#ifndef __BH_COMMUNITY__
#	include "enterprise/loader/RCAttr_load.h"
#	include "enterprise/loader/DataParserFactory.h"
#else /* not __BH_COMMUNITY__ */
#	include "community/loader/RCAttr_load.h"
#	include "community/loader/DataParserFactory.h"
#endif /* __BH_COMMUNITY__ */
#include "DataParser.h"
#include "system/LargeBuffer.h"
#include "system/IOParameters.h"

using namespace std;

DataLoader::DataLoader(vector<RCAttrLoad*> attrs, Buffer& buffer, const IOParameters& iop)
	:	attrs(attrs), buffer(&buffer), iop(iop)
{
	no_loaded_rows = 0;
	no_prepared = 0;
}

DataLoader::~DataLoader()
{
}

_uint64 DataLoader::Proceed()
{
	no_loaded_rows = 0;
	int to_prepare = MAX_NO_OBJ - (int)(attrs[0]->NoObj() % MAX_NO_OBJ);
	try {
		while(true)	{
			if((no_prepared = parser->Prepare(to_prepare)) == 0) {
				if(LoadData() == 0)
					break;
				else if((no_prepared = parser->Prepare(to_prepare)) == 0)
					throw FormatRCException(BHERROR_DATA_ERROR, parser->error_value.first + no_loaded_rows, parser->error_value.second);
			}
			for(uint i = 0; i < attrs.size(); i++) {
					parser->PrepareNextCol();
				NewValuesSet nvs(parser.get(), i, attrs[i]->Type().GetCollation());
#ifndef __BH_COMMUNITY__
				attrs[i]->LoadData(&nvs, process_type != ProcessType::BHLOADER);
#else
				attrs[i]->LoadData(&nvs);
#endif
			}
			no_loaded_rows += no_prepared;
			to_prepare = MAX_NO_OBJ - (int)(attrs[0]->NoObj() % MAX_NO_OBJ);
		}
	} catch(FormatRCException& frce) {
		Abort();
		frce.m_row_no += no_loaded_rows;
		throw frce;
	} catch(DataTypeConversionRCException) {
		Abort();
		throw FormatRCException(BHERROR_DATA_ERROR, parser->error_value.first + no_loaded_rows, parser->error_value.second);
	} catch(...) {
		Abort();
		throw InternalRCException(BHERROR_UNKNOWN);
	}

	if(no_loaded_rows == 0)
		throw FormatRCException(BHERROR_DATA_ERROR, parser->error_value.first + no_loaded_rows, parser->error_value.second);

	for_each(attrs.begin(), attrs.end(), boost::bind(&RCAttrLoad::Save, _1));
	return no_loaded_rows;
}

int DataLoader::LoadData()
{
	int ret = 0;
	try	{
		ret = buffer->BufFetch(int(parser->buf_end - parser->buf_ptr));
	} catch (FileRCException& frce) {
		Abort();
		throw frce;
	}

	if(ret)	{
		parser->buf_ptr = parser->buf_start = buffer->Buf(0);
		parser->buf_end = parser->buf_start + buffer->BufSize();
	}
	return ret;
}


void DataLoader::Abort()
{
}

auto_ptr<DataLoader>
DataLoader::CreateDataLoader(vector<RCAttrLoad*> attrs, Buffer& buffer, const IOParameters& iop)
{
	auto_ptr<DataLoader> data_loader(new DataLoader(attrs, buffer, iop));
        data_loader->parser = DataParserFactory::CreateDataParser(attrs, buffer, iop);
	return data_loader;
}
