/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "types/DataConverter.h"
#include "core/tools.h"
#include "core/bintools.h"
#include "shared/loader/NewValueSet.h"
#ifndef __BH_COMMUNITY__
#	include "enterprise/loader/RCAttr_load.h"
#else /* not __BH_COMMUNITY__ */
#	include "community/loader/RCAttr_load.h"
#endif /* __BH_COMMUNITY__ */
#include "system/IOParameters.h"
#include "DataParserForText.h"
#include "types/ValueParserForText.h"

using namespace std;
using namespace boost;
using namespace boost::assign;

DataParserForText::DataParserForText(vector<RCAttrLoad*> attrs, Buffer& buffer, const IOParameters& iop)
	:	DataParser(attrs, buffer, iop), delimiter(iop.Delimiter()),
	string_qualifier(iop.StringQualifier()), escape_char(iop.EscapeCharacter())
{
	ShiftBufferByBOM();
	qualifier_present = new bool * [no_attr];
	for(int i = 0; i < no_attr; i++)
		qualifier_present[i] = new bool [65536];

	prepared = false;
	crlf = 0;
	value_parser = DataFormat::GetDataFormat(iop.GetEDF())->CreateValueParser();

	for(int i = 0; i < no_attr; i++)
		if(atis[i].Type() != RC_STRING)
			value_sizes[i] = new uint [65536];
}

DataParserForText::~DataParserForText()
{
	for(int i = 0; i < no_attr; i++) {
		delete [] qualifier_present[i];
		if(atis[i].Type() != RC_STRING) {
			delete [] value_sizes[i];
			value_sizes[i] = 0;
		}
	}
	delete [] qualifier_present;
}

void DataParserForText::PrepareValuesPointers()
{
	int opsbs = (delimiter ? 1 : 0);
	for(int i = 0; i < no_prepared; i++) {
		if (cur_attr != 0 ) {
			values_ptr[i] += objs_sizes[cur_attr-1][i] + opsbs ;
			if(qualifier_present[cur_attr - 1][i]) //previous column had string qualifier
				values_ptr[i]--; //undoing the following ++
		}
		if(string_qualifier && values_ptr[i][0] == string_qualifier) {
			values_ptr[i]++;
			qualifier_present[cur_attr][i]=true;
		} else
			qualifier_present[cur_attr][i]=false;
	}
}

void DataParserForText::FormatSpecificProcessing()
{
	if(ATI::IsBinType(cur_attr_type)) {
		for(int i = 0; i < no_prepared; i++) {
			int p = 0;
			uint os = GetObjSize(i);
			for(uint l = 0; l < os; l++) {
				int a = isalpha((uchar)values_ptr[i][l]) ? tolower(values_ptr[i][l]) - 'a' + 10 : values_ptr[i][l] - '0';
				int b = isalpha((uchar)values_ptr[i][l+1]) ? tolower(values_ptr[i][l+1]) - 'a' + 10 : values_ptr[i][l+1] - '0';
				values_ptr[i][p++] =  a * 16 + b;
				l++;
			}
		}
	}
}

void DataParserForText::GetCRLF()
{
	bool inside_string = false;
	char* ptr = buf_start;
	while(ptr != buf_end) {
		if(!inside_string) {
			if(string_qualifier && *ptr == string_qualifier)
				inside_string = 1;
			else if(*ptr == '\r' && *(ptr+1) == '\n') {
				crlf = 2;
				break;
			} else if(*ptr == '\n') {
				crlf = 1;
				break;
			}
		} else if(inside_string && string_qualifier && *ptr == string_qualifier) {
			if(*(ptr + 1) == delimiter ||  *(ptr + 1) == '\n')
				inside_string = 0;
			else if((ptr + 2) <= buf_end && *(ptr + 1) == '\r' && *(ptr + 2) == '\n')
				inside_string = 0;
		}
		ptr++;
	}
}

inline int DataParserForText::GetRowSize(char* row_ptr, int rno)
{
	if(!prepared) {
		GetCRLF();
		prepared = true;
	}

	if(row_ptr != buf_end) {
		char* ptr = row_ptr;
		char* v_ptr = row_ptr;
		bool stop = false;
		bool inside_string = false;
		int no_attrs_tmp = 0;
		int tan = no_attr - 1;
		bool esc_char = false;
		bool new_value_beg = true;

		while(!stop) {
			if( !esc_char &&
			    escape_char &&
			    *ptr == escape_char &&
			    ATI::IsTxtType(atis[no_attrs_tmp].Type()) &&
			    !(escape_char == string_qualifier && new_value_beg) &&
			    (   !(escape_char == string_qualifier && !new_value_beg &&
				(((no_attrs_tmp != tan) && (*(ptr+1) == delimiter)) ||
				   ((no_attrs_tmp == tan) && IsEOR(ptr+1))
				))
                            ))
			{
				new_value_beg = false;
				esc_char = true;
			} else if(string_qualifier && *ptr == string_qualifier && !esc_char /*&&
					ATI::IsTxtType(atis[no_attrs_tmp].Type())*/
			) {
				new_value_beg = false;
				esc_char = false;
				if(inside_string)
					inside_string = false;
				else
					inside_string = true;
			} else if(!inside_string && !esc_char) {
				if(
					((no_attrs_tmp != tan) && (*ptr == delimiter)) ||
					((no_attrs_tmp == tan) && IsEOR(ptr))
				) {
					objs_sizes[no_attrs_tmp][rno] = (uint)(ptr - v_ptr);
					if(no_attrs_tmp != 0)
						objs_sizes[no_attrs_tmp][rno]--;
					value_sizes[no_attrs_tmp][rno] = objs_sizes[no_attrs_tmp][rno];
					//no_esc_char[no_attrs_tmp][rno] = 0;
					v_ptr = ptr;
					no_attrs_tmp++;
					new_value_beg = true;
					if(no_attrs_tmp == no_attr)
						break;
				} else if(IsEOR(ptr)) {
					if(string_qualifier || (no_attrs_tmp != tan ))
						throw FormatRCException(BHERROR_DATA_ERROR, rno + 1, no_attrs_tmp + 1 );
				}
				esc_char = false;
			} else {
				new_value_beg = false;
				esc_char = false;
			}
			if((ptr + 1) == buf_end) {
				error_value.first = rno + 1;
				error_value.second =  no_attrs_tmp + 1;
				break;
			}
			ptr++;
		}
		if(no_attrs_tmp == no_attr)
			return (int)(ptr - row_ptr) + crlf;
	}
 	return -1;
}

void DataParserForText::PrepareObjsSizes()
{
	objs_sizes_ptr = objs_sizes[cur_attr];
	for(int i = 0; i < no_prepared; i++) {
		if (string_qualifier ) {
			if(objs_sizes_ptr[i] && qualifier_present[cur_attr][i])
				value_sizes[cur_attr][i] -= 2;
		}
		//process escape chars

		// eliminate trailing spaces
		if(ATI::IsCharType(cur_attr_type)) {
			for (char* c = values_ptr[i] + value_sizes[cur_attr][i]- 1; c >= values_ptr[i]; c--) {
				if (*c == ' ')
					value_sizes[cur_attr][i]--;
				else
					break;
			}
		}

	}
}

void DataParserForText::ProcessEscChar(int i)
{
	if(!IsNull(i)) {
		uint s = GetObjSize(i);
		int p = 0;
		for(uint j = 0; j < s; j++)	{
			if(values_ptr[i][j] == escape_char && !(escape_char == string_qualifier && j == 0) && (j != s - 1)){
				values_ptr[i][p] = TranslateEscapedChar(values_ptr[i][j+1]);
				j++;
				value_sizes[cur_attr][i]--;
			} else
				values_ptr[i][p] = values_ptr[i][j];
			p++;
		}
	}
}

void DataParserForText::PrepareNulls()
{
	if(!IsLastColumn()) {
		for(int i = 0; i < no_prepared; i++) {
			int rs = row_sizes[i];
			if(
				*values_ptr[i] == delimiter ||
				(((values_ptr[i] + 5) < (rows_ptr[i] + rs)) && strncasecmp(values_ptr[i], "NULL", 4) == 0
						&& (*(values_ptr[i] + 4) == delimiter)) ||
				(((values_ptr[i] + 3) < (rows_ptr[i] + rs)) && strncasecmp(values_ptr[i], "\\N", 2) == 0
						&& (*(values_ptr[i] + 2) == delimiter))
			)
				nulls[i] = 1;
			else
				nulls[i] = 0;
		}
	} else {
		for(int i = 0; i < no_prepared; i++) {
			int rs = row_sizes[i];
			if(
				IsEOR(values_ptr[i]) ||
				(((values_ptr[i] + 4 + crlf) == (rows_ptr[i] + rs)) && strncasecmp(values_ptr[i], "NULL", 4) == 0
						&& IsEOR(values_ptr[i] + 4)) ||
				(((values_ptr[i] + 2 + crlf) == (rows_ptr[i] + rs)) && strncasecmp(values_ptr[i], "\\N", 2) == 0
						&& IsEOR(values_ptr[i] + 2))
			)
				nulls[i] = 1;
			else
				nulls[i] = 0;
		}
	}
}

bool DataParserForText::FormatSpecificDataCheck()
{
//	if(ATI::IsTxtType(cur_attr_type)) {
//		if(string_qualifier)
//			for(int i = 0; i < no_prepared; i++) {
//				if(!IsNull(i) && *(values_ptr[i] - 1) != string_qualifier) {
//					error_value.first = no_loaded_rows + i + 1;
//					error_value.second = cur_attr + 1;
//					return false;
//				}
//			}
//	} else
	if(ATI::IsTxtType(cur_attr_type)) {
		for(int i = 0; i < no_prepared; i++)
			ProcessEscChar(i);
	} else if(ATI::IsBinType(cur_attr_type)) {
		for(int i = 0; i < no_prepared; i++) {
			if(!IsNull(i) && (GetObjSize(i)%2)) {
				error_value.first = i + 1;
				error_value.second = cur_attr + 1;
				return false;
			}
		}
	}
	return true;
}

uint DataParserForText::GetValueSize(int ono)
{
	if(ATI::IsBinType(CurrentAttributeType()))
		return GetObjSize(ono) / 2;
	else
		return value_sizes[cur_attr][ono];
		//- no_esc_char[cur_attr][ono];
}

char DataParserForText::TranslateEscapedChar(char c)
{
	static char in[] = {'0',  'b',  'n',  'r',  't',  char(26)};
	static char out[]= {'\0', '\b', '\n', '\r', '\t', char(26)};
	for(int i = 0; i < 6; i++)
		if(in[i] == c)
			return out[i];
	return c;
}

bool DataParserForText::IsEOR(const char* ptr) const
{
	return (crlf == 2 && *ptr == '\r' && *(ptr + 1) == '\n') ||	(crlf == 1 && *ptr == '\n');
}

void DataParserForText::ShiftBufferByBOM()
{
	// utf-8
	if((uchar)buf_start[0] == 0xEF && (uchar)buf_start[1] == 0xBB && (uchar)buf_start[2] == 0xBF) {
		buf_start += 3;
		buf_ptr += 3;
	} else if((uchar)buf_start[0] == 0xFE && (uchar)buf_start[1] == 0xFF || // UTF-16-BE
		(uchar)buf_start[0] == 0xFF && (uchar)buf_start[1] == 0xFE) {  // UTF-16-LE
		buf_start += 2;
		buf_ptr += 2;
	}
}