/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "system/RCSystem.h"
#include "Loader.h"
#include "core/RSI_Framework.h"

using namespace std;
using namespace bh;

// This is mysqld server id.
extern ulong server_id;

int Loader::LoaderSaveThreadNumber = 16;

Loader::Loader()
  : rs_owner(-1)
{
    memset (& params, 0, sizeof(params));
}

Loader::~Loader()
{
	shared_file.Close();
}

void Loader::Init(string bhload_app, IOParameters* io_params)
{
	if(!io_params || bhload_app.length() == 0)
		throw InternalRCException("Unable to start Loader.");

	this->bhload_app = bhload_app;
    iop = *io_params;

	string err_msg= "";
	mode= false;
	try {
		Configuration::LoadSettings();
	} catch (const Configuration::Exception& e) {
		rclog << lock << e.what() << " Configuration::Exception" << __FILE__ << __LINE__ << unlock;
	}

	try {
		Loader::LoaderSaveThreadNumber = Configuration::GetProperty(Configuration::LoaderSaveThreadNumber);
	} catch(...) {
		Loader::LoaderSaveThreadNumber = 16;
	}

	shared_obj_name= "bhload-info-";
	Loader::BuidId(shared_obj_name);

	rs_owner= RS_CREATOR;
}

void Loader::Init(string shared_obj_name, string shared_obj_path)
{
	string err_msg = "";

	mode= false;
	try {
		Configuration::LoadSettings();
	} catch (const Configuration::Exception& e) {
		rclog << lock << e.what() << " Configuration::Exception" << __FILE__ << __LINE__ << unlock;
	}

	try {
		Loader::LoaderSaveThreadNumber = Configuration::GetProperty(Configuration::LoaderSaveThreadNumber);
	} catch(...) {
		Loader::LoaderSaveThreadNumber = 16;
	}
	try {
		use_charset =  Configuration::GetProperty(Configuration::UseCharset);
	} catch(...) {
		use_charset = true;
	}

	mode= true;

	string shared_obj;

	shared_obj = shared_obj_path;
	shared_obj += shared_obj_name;
	try {
		shared_file.OpenReadWrite(shared_obj_path.c_str());
	} catch(DatabaseRCException& e)	{
		err_msg = "Unable to start Loader. Trying to open a memory mapped file but it says - ";
		err_msg += e.what();
		throw InternalRCException(err_msg.c_str());
	}

    ReLoadParameters();
	rs_owner= RS_USER;
}

void Loader::FlushParameters()
{
	shared_file.Seek(0, SEEK_SET);
	shared_file.WriteExact(& params, sizeof(params));
    const int buf_size = 4096;
    char* buf = (char*) malloc(buf_size);
    if (buf == 0) throw InternalRCException("Loader::FlushParameters: Failed to allocate memory");
    iop.PutParameters (buf, buf_size);
    int len = iop.GetParamsBufferLen();
	try {
		shared_file.WriteExact(&len, sizeof(len));
		shared_file.WriteExact(buf, len);
	} catch(DatabaseRCException& )	{
		free(buf);
		throw InternalRCException("Loader::FlushParameters: Failed to write");
	}
	free (buf);
}

void Loader::ReLoadParameters()
{
	shared_file.Seek(0, SEEK_SET);
	shared_file.ReadExact(&params, sizeof(params));
    int len;
	shared_file.ReadExact(&len, sizeof(len));
    if (len > 4096) throw InternalRCException("Loader::FlushParameters: Invalid iop length");
    char* buf = (char*) malloc(len);
	try {
		shared_file.ReadExact(buf, len);
	} catch(DatabaseRCException& )	{
        free(buf);
        throw InternalRCException("Loader::FlushParameters: Failed to read");
    }
    iop.GetParameters (buf);
    free(buf);
}

void Loader::BuidId(string& id)
{
        char uid[64] = "";
        static ulong cur_localid = 0;
        ulong cur_serverid = server_id;

        time_t cur_time = time(NULL);
        srand (cur_time);
        ulong cur_rand = rand();

        cur_localid ++;

        sprintf(uid, "%ld-%ld-%ld-%ld", cur_serverid, cur_time, cur_rand, cur_localid);
        id += uid;
        //rclog << lock << id <<unlock;
}

void Loader::DeInit(bool cleanup)
{
	shared_file.Close();
	if (cleanup) {
		string shared_obj;
		shared_obj = GetSystemTempPath();
		shared_obj += shared_obj_name;
		RemoveFile(shared_obj.c_str(), false);
	}
}
