/* Copyright (C) 2005-2008 Infobright Inc. */

#include <sstream>
#include <boost/shared_ptr.hpp>

#include "common/CommonDefinitions.h"
#include "core/RCAttrPack.h"
#include "core/tools.h"
#include "core/WinTools.h"
#include "RCAttrLoadBase.h"
#include "shared/loader/NewValueSet.h"
#include "shared/loader/BHLoader.h"
#include "core/DPN.h"
#include "core/DataPackImpl.h"
#include "types/RCDataTypes.h"
#include "core/DataCache.h"
#include "core/Transaction.h"

using namespace std;
using namespace boost;

int RCAttrLoadBase::count_spt = 0;
IBCond  RCAttrLoadBase::threads_count_cond;

RCAttrLoadBase::RCAttrLoadBase(int a_num,int t_num,char *a_path,int conn_mode,unsigned int s_id, DTCollation collation) throw(DatabaseRCException)
	: RCAttr(a_num, t_num, a_path, conn_mode, s_id, collation), illegal_nulls(false)
{
}

RCAttrLoadBase::~RCAttrLoadBase()
{
	LogWarnigs();
}

void RCAttrLoadBase::LoadPackInfoForLoader()
{
	if(!GetPackInfoCollaped())
		return;

	IBGuard guard( attr_mutex);
	WaitForSaveThreads();
	BHASSERT(FileFormat()==10, "should be 'file_format==10'");
	if (GetDictOffset() != 0)
	{
		IBFile fattr;
		try {
			fattr.OpenReadOnly(AttrFileName(fname_buf, 1024, file));
			int fin_size = fattr.Seek(0, SEEK_END);
			char* buf_ptr = new char [(fin_size - GetPackOffset()) + 1];
			fattr.Seek(GetPackOffset(), SEEK_SET);
			fattr.ReadExact(buf_ptr, fin_size - GetPackOffset());
			if(GetDictOffset() != 0 && (fin_size - GetDictOffset())) LoadDictionaries(buf_ptr);
			delete [] buf_ptr;
		}
		catch (DatabaseRCException& e) {
			rclog << lock << "Internal error: unable to load column dictionary from " << fname_buf << ". " << GetErrorMessage(errno) << unlock;
			throw e;
		}
	}
	// load last pack description (file of a pack, localization in file, statistics, ...)
	if (NoPack() > 0) packs_omitted = NoPack()-1;
	else packs_omitted = 0;
	if (NoPack() > 0)
	{
		ExpandDPNArrays();
		char buf[37];
		IBFile fdpn;
		try {
			fdpn.OpenReadOnly(DPNFileName(fname_buf,1024));
			if (GetCurReadLocation() == 1)
				fdpn.Seek(37, SEEK_CUR);
			fdpn.ReadExact(buf, 37);
		}
		catch (DatabaseRCException& e) {
			rclog << lock << "Internal error: unable to load data pack nodes from " << fname_buf << ". " << GetErrorMessage(errno) << unlock;
			throw e;
		}
		RestoreDPN(buf, 0);
	}
	SetPackInfoCollaped(false);
}

void RCAttrLoadBase::ExpandDPNArrays()
{
	if((NoPack() - packs_omitted + 1) >= GetNoPackPtr()) {
		// Create new package table (enlarged by one) and copy old values
		uint old_no_pack_ptr = GetNoPackPtr();
		uint n_pack_ptrs = (NoPack() - packs_omitted + 1) + 500;
		SetNoPackPtr(n_pack_ptrs);
		WaitForSaveThreads();

		pack_mode = (char*) realloc(pack_mode, GetNoPackPtr() * sizeof(char));
		packs = (AttrPack**) realloc(packs, GetNoPackPtr() * sizeof(AttrPack*));
		is_stored = (bool*) realloc(is_stored, GetNoPackPtr() * sizeof(bool));
		pack_file = (int*) rc_realloc(pack_file, GetNoPackPtr() * sizeof(int), BLOCK_TEMPORARY);
		pack_addrs = (uint*) rc_realloc(pack_addrs, GetNoPackPtr() * sizeof(uint), BLOCK_TEMPORARY);
		local_min = (_int64*) rc_realloc(local_min, GetNoPackPtr() * sizeof(_int64), BLOCK_TEMPORARY);
		local_max = (_int64*) rc_realloc(local_max, GetNoPackPtr() * sizeof(_int64), BLOCK_TEMPORARY);
		local_sum = (_int64*) rc_realloc(local_sum, GetNoPackPtr() * sizeof(_int64), BLOCK_TEMPORARY);
		local_size = (ushort*) rc_realloc(local_size, GetNoPackPtr() * sizeof(ushort), BLOCK_TEMPORARY);
		loc_no_obj = (ushort*) rc_realloc(loc_no_obj, GetNoPackPtr() * sizeof(ushort), BLOCK_TEMPORARY);
		loc_no_nulls = (uint*) rc_realloc(loc_no_nulls, GetNoPackPtr() * sizeof(uint), BLOCK_TEMPORARY);

		//		delete [] spThreads;
		//		spThreads = new ThreadWrapper [no_pack_ptr];

		if(!local_min || !local_max || !local_sum || !local_size || !loc_no_obj || !loc_no_nulls || !pack_mode || !packs || !pack_file
				|| !pack_addrs) {
			rclog << lock << "Error: out of memory (" << GetNoPackPtr() - old_no_pack_ptr << " new packs failed). (25)"
					<< unlock;
			return;
		}

		int units = GetNoPackPtr() - old_no_pack_ptr;
		memset(packs + old_no_pack_ptr, 0, units * sizeof(AttrPack*));
		memset(is_stored + old_no_pack_ptr, 0, units * sizeof(bool));
		memset(pack_addrs + old_no_pack_ptr, 0, units * sizeof(uint));
		memset(local_sum + old_no_pack_ptr, 0, units * sizeof(_int64));
		memset(local_size + old_no_pack_ptr, 0, units * sizeof(ushort));
		memset(loc_no_obj + old_no_pack_ptr, 0, units * sizeof(ushort));
		memset(loc_no_nulls + old_no_pack_ptr, 0, units * sizeof(uint));
		memset(local_min + old_no_pack_ptr, 0, units * sizeof(_int64));
		memset(local_max + old_no_pack_ptr, 0, units * sizeof(_int64));

		for(uint i = old_no_pack_ptr; i < GetNoPackPtr(); i++) {
			pack_mode[i] = 3;
			is_stored[i] = false;
			pack_file[i] = PF_NO_OBJ;
		}
	}
}

void RCAttrLoadBase::UnlockPackFromUse(unsigned pack_no)
{
	if(process_type == ProcessType::BHLOADER) { //loader
		MEMORY_LOCK_GUARD;
		SynchronizedDataCache::GlobalDataCache().ReleasePack(table_number, attr_number, pack_no + packs_omitted);
		packs[pack_no] = 0;
		if(pack_mode[pack_no] == 2)
			pack_mode[pack_no] = 1;
	} else
		RCAttr::UnlockPackFromUse(pack_no);
}

void RCAttrLoadBase::LockPackForUse(unsigned pack_no)
{
	assert((int)pack_no < NoPack());
	MEMORY_LOCK_GUARD;
	if((process_type == ProcessType::BHLOADER) || !ConnectionInfoOnTLS.IsValid())
		packs[pack_no] = SynchronizedDataCache::GlobalDataCache().GetAttrPackRO(table_number, attr_number, pack_no + packs_omitted);
	else
		packs[pack_no] = ConnectionInfoOnTLS.Get().GetTransaction()->GetAttrPackRO(table_number, attr_number, pack_no + packs_omitted);
	if(packs[pack_no]) {
		if(pack_mode[pack_no] == 1)
			pack_mode[pack_no] = 2;
		packs[pack_no]->Lock();
	} else {
		if(PackType() == PackS)
			packs[pack_no] = new AttrPackS(*this, pack_no);
		else
			packs[pack_no] = new AttrPackN(*this, pack_no);
		packs[pack_no]->Lock();
	}
	LoadPack(pack_no);
}

void RCAttrLoadBase::Append(shared_ptr<DPN> dpn)
{
	LoadPackInfoForLoader();
	ExpandDPNArrays();

	pack_file[NoPack() - packs_omitted] = dpn->GetPackFile();
	pack_addrs[NoPack() - packs_omitted] = dpn->GetPackAddress();
	local_min[NoPack() - packs_omitted] = dpn->GetMin();
	local_max[NoPack() - packs_omitted] = dpn->GetMax();
	local_sum[NoPack() - packs_omitted] = dpn->GetSum();
	local_size[NoPack() - packs_omitted] = dpn->GetSize();
	loc_no_obj[NoPack() - packs_omitted] = dpn->GetNoObjInternal();
	loc_no_nulls[NoPack() - packs_omitted] = dpn->GetNoNulls();

	BHASSERT(!packs[NoPack()-packs_omitted], "'packs[no_pack-packs_omitted]' should be null");
//		this->LockMemoryManager();
	packs[NoPack()-packs_omitted] = NULL; //here we do not need the actual data pack
//		packs[NoPack()-packs_omitted]->Lock();
//		this->UnlockMemoryManager();

	SetNoObj(NoObj() + (DP_SIZE - GetNoValues(NoPack() - packs_omitted - 1)) + dpn->GetNoObj());
	SetNoNulls(NoNulls() + dpn->GetNoNulls());
	SetNoPack(NoPack() + 1);

	if(PackType() == PackN && rsi_manager && NoObj()) {
		if(rsi_manager->IndexExists(RSIndexID(RSI_HIST, table_number, attr_number))) {
			LoadPackInfo();
			RSIndex_Hist* hist = (RSIndex_Hist*)rsi_manager->GetIndex(RSIndexID(RSI_HIST, table_number, attr_number), GetCurReadLocation());

			if(rsi_hist_update == NULL)
				rsi_hist_update = (RSIndex_Hist*) rsi_manager->GetIndexForUpdate(RSIndexID(RSI_HIST, table_number,
						attr_number), GetCurReadLocation());

			if(rsi_hist_update == NULL) {
				rccontrol << lock << "Warning: cannot access RSI_Hist" << unlock;
				return;
			}

			rsi_hist_update->AppendKN(dpn->GetDataPackId(), hist);
			rsi_manager->ReleaseIndex(hist);
		}
	} else if(PackType() == PackS && rsi_manager && NoObj()) {
		if(rsi_manager->IndexExists(RSIndexID(RSI_CMAP, table_number, attr_number))) {

			LoadPackInfo();
			RSIndex_CMap* cmap = (RSIndex_CMap*)rsi_manager->GetIndex(RSIndexID(RSI_CMAP, table_number, attr_number), GetCurReadLocation());

			if(rsi_cmap_update == NULL)
				rsi_cmap_update = (RSIndex_CMap*)rsi_manager->GetIndexForUpdate(RSIndexID(RSI_CMAP, table_number,
						attr_number), GetCurReadLocation());

			if(rsi_cmap_update == NULL) {
				rccontrol << lock << "Warning: cannot access RSI_Hist" << unlock;
				return;
			}
			rsi_cmap_update->AppendKN( dpn->GetDataPackId(), cmap);
			rsi_manager->ReleaseIndex(cmap);
		}
	}
}

void RCAttrLoadBase::CreateNewPackage()
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	stringstream s1;
	s1 << "a[" << attr_number << "].CreateNewPackage()";
	FETOperator feto1(s1.str());
	FETOperator feto2("CreateNewPackage()");
#endif
	ExpandDPNArrays();
	if(PackType() == PackN)
	{
		local_min[NoPack()-packs_omitted]=PLUS_INF_64;
		local_max[NoPack()-packs_omitted]=MINUS_INF_64;

		BHASSERT(!packs[NoPack()-packs_omitted], "'packs[no_pack-packs_omitted]' should be null");
		{
            MEMORY_LOCK_GUARD;
		    packs[NoPack()-packs_omitted] = new AttrPackN(*this, NoPack()-packs_omitted);
		    packs[NoPack()-packs_omitted]->Lock();
		}
	}
	else
	{
		local_min[NoPack()-packs_omitted]=0;
		local_max[NoPack()-packs_omitted]=-1;
		{
            MEMORY_LOCK_GUARD;
		    packs[NoPack()-packs_omitted] = new AttrPackS(*this, NoPack()-packs_omitted);
		    packs[NoPack()-packs_omitted]->Lock();
		}
	}
	/*loc_no_obj[no_pack-packs_omitted-1]=0;		// WARNING: this is not a good indicator of empty pack, because "0" means 1 object, and "0xFFFF" mens 65536 objects
	loc_no_nulls[no_pack-packs_omitted-1]=0;*/
	SetNoPack(NoPack()+1);
}

void RCAttrLoadBase::LoadPack(uint n)
{
	BHASSERT(n < NoPack()-packs_omitted, "should be 'n < no_pack-packs_omitted'");
	if(pack_mode[n]!=1 || !ShouldExist(n))
		return;
	WaitForSaveThreads();
	LoadPackInherited(n);
}


int RCAttrLoadBase::Save()
{
	if(current_state != 1) {
		rclog << lock << "Error: cannot save. It is read only session." << unlock;
		throw ;
	}
	LoadPackInfo();
	WaitForSaveThreads();
	int npack=NoPack();
	for(uint i=0;i<npack-packs_omitted;i++)
		SavePack(i);
	WaitForSaveThreads();
	if(rsi_hist_update || rsi_cmap_update)
		SaveRSI();
	BHASSERT(FileFormat()==10, "should be 'file_format==10'");
	SaveDPN();
	SaveHeader();
	return 0;
}

void RCAttrLoadBase::SaveDPN()
{
	if (NoPack()==0) return;
	const int buf_size = 10000;
	char buf[buf_size];
	IBFile fdpn;
	char* fn = DPNFileName(fname_buf, 1024);

	try
	{
		fdpn.OpenCreate(fn);
		ushort buffer_pos = 0;
		int npack = NoPack();
		for(uint p = 0; p < npack-packs_omitted; p++)
		{
			if(buffer_pos + 37 > buf_size)
			{
				fdpn.Write(buf, buffer_pos);
				buffer_pos = 0;
			}
			if (p==0) StoreDPN(npack-packs_omitted-1, buf+buffer_pos);	// last pack is saved first
			else StoreDPN(p-1, buf+buffer_pos);
			if (p==0)
			{
				// last pack is saved at the beginning of the file
				if (GetCurSaveLocation()==1) fdpn.Seek(37, SEEK_CUR);
				fdpn.Write(buf, 37);
				// the rest is appended to the end of the file
				if (p < npack-packs_omitted-1) fdpn.Seek((packs_omitted+2)*37, SEEK_SET);
			}
			else buffer_pos += 37;
		}
		if (buffer_pos>0) fdpn.Write(buf, buffer_pos);
		fdpn.Close();
	}
	catch(DatabaseRCException& e)
	{
		rclog << lock << "Internal error: unable to write data pack nodes to " << fn << ". " << GetErrorMessage(errno) << unlock;
		throw e;
	}
}


void RCAttrLoadBase::CompareAndSetCurrentMin(RCBString tstmp, RCBString & min, bool set)
{
    bool res;
    if(RequiresUTFConversions(Type().GetCollation())){
        res = CollationStrCmp(Type().GetCollation(), tstmp, min) < 0;
    }else
        res = strcmp(tstmp, min) < 0;

    if(!set || res){
        min = tstmp;
        min.MakePersistent();
        set = true;
    }
}

void RCAttrLoadBase::CompareAndSetCurrentMax(RCBString tstmp, RCBString & max)
{
    bool res;
    if(RequiresUTFConversions(Type().GetCollation())){
        res = CollationStrCmp(Type().GetCollation(), tstmp, max) > 0;
    }else
        res = strcmp(tstmp, max) > 0;

    if(res){
        max = tstmp;
        max.MakePersistent();
    }
}

uint RCAttrLoadBase::RoundUpTo8Bytes(RCBString& s)
{
	int useful_len = 0;
	if(Type().GetCollation().collation->mbmaxlen > 1) {
	  int next_char_len;
	  while(true) {
		  if(useful_len >= s.len)
			  break;
		  next_char_len = Type().GetCollation().collation->cset->mbcharlen(Type().GetCollation().collation, (uchar)s.val[useful_len+s.pos]);
		  assert("wide character unrecognized" && next_char_len > 0);
		  if(useful_len + next_char_len > 8)
			  break;
		  useful_len += next_char_len;
	  }
	} else
	  useful_len = s.len > 8 ? 8 : s.len;
	return useful_len;
}

void RCAttrLoadBase::SetPackMax(uint pack, RCBString& max_s)
{
	  if(RequiresUTFConversions(Type().GetCollation())) {
		  int useful_len = RoundUpTo8Bytes(max_s);

		  //deal with ...ae -> ...

		  strncpy((uchar*)(&local_max[pack]), max_s, useful_len);
		  if(useful_len < 8)
			  ((uchar*)&local_max[pack])[useful_len] = 0;
	  } else
		  strncpy((uchar*)(&local_max[pack]), max_s, min(max_s.size(), sizeof(_int64)));
}

void RCAttrLoadBase::SetPackMin(uint pack, RCBString& min_s)
{
	  if(RequiresUTFConversions(Type().GetCollation())) {
		  int useful_len = RoundUpTo8Bytes(min_s);

		  //deal with ...ae -> ...

		  strncpy((uchar*)(&local_min[pack]), min_s, useful_len);
		  if(useful_len < 8)
			  ((uchar*)&local_min[pack])[useful_len] = 0;
	  } else
		  strncpy((uchar*)(&local_min[pack]), min_s, sizeof(_int64));
}


RCBString RCAttrLoadBase::MinS(Filter* f)
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator feto("RCAttr::MinS(...)");
#endif
	if(f->IsEmpty() || !ATI::IsStringType(TypeName()) || NoObj() == 0 || NoObj() == NoNulls())
		return RCBString();
	RCBString min;
	bool set = false;
	if(f->NoBlocks() + packs_omitted != NoPack())
		throw DatabaseRCException("Data integrity error, query cannot be evaluated (MinS).");
	else
	{
		RCBString tstmp;
		LoadPackInfo();
        for(uint b = 0; b < NoPack() - packs_omitted; b++)
		{
        	if(f->IsEmpty(b))
        		continue;
			if(PackType() == PackN &&
					(GetPackOntologicalStatus(b) == UNIFORM ||
					(GetPackOntologicalStatus(b) == UNIFORM_AND_NULLS && f->IsFull(b)))
			)
			{
				tstmp = DecodeValue_S(local_min[b]);
				bool res;
			    CompareAndSetCurrentMin(tstmp, min, set);
			}
			else if(!(pack_file[b] == PF_NULLS_ONLY || pack_file[b] == PF_NO_OBJ))
			{
				f->Rewind(b);
				int block = 0;
				ushort n = 0;
				f->GetNext1(block, n);
				while(b == block)
				{
					if(PackType() == PackS && packs[b]->IsNull(n) == 0)
					{
						int len = ((AttrPackS*)packs[b])->GetSize(n);
						tstmp = len ? RCBString(((AttrPackS*)packs[b])->GetVal(n), len) : RCBString("");
					    CompareAndSetCurrentMin(tstmp, min, set);
					}
					else if(PackType() == PackN && !packs[b]->IsNull(n))
					{
						tstmp = DecodeValue_S(((AttrPackN*)(packs[b]))->GetVal(n) + local_min[b]);
					    CompareAndSetCurrentMin(tstmp, min, set);
					}
					f->GetNext1(block, n);
				}
			}
		}
	}
	return min;
}

RCBString RCAttrLoadBase::MaxS(Filter* f)
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator feto("RCAttr::MaxS(...)");
#endif
	if(f->IsEmpty() || !ATI::IsStringType(TypeName()) || NoObj() == 0 || NoObj() == NoNulls())
		return RCBString();

	RCBString max;
	if(f->NoBlocks() + packs_omitted != NoPack())
		throw DatabaseRCException("Data integrity error, query cannot be evaluated (MaxS).");
	else
	{
		RCBString tstmp;
		LoadPackInfo();
        for(uint b = 0; b < NoPack() - packs_omitted; b++)
		{
        	if(f->IsEmpty(b))
        		continue;
			if(PackType() == PackN &&
					(GetPackOntologicalStatus(b) == UNIFORM ||
					(GetPackOntologicalStatus(b) == UNIFORM_AND_NULLS && f->IsFull(b)))
			)
			{
				tstmp = DecodeValue_S(local_min[b]);
				CompareAndSetCurrentMax(tstmp, max);
			}
			else if(!(pack_file[b] == PF_NULLS_ONLY || pack_file[b] == PF_NO_OBJ))
			{
				f->Rewind(b);
				int block = 0;
				ushort n = 0;
				f->GetNext1(block, n);
				while(b == block)
				{
					if(PackType() == PackS && packs[b]->IsNull(n) == 0)
					{
						int len = ((AttrPackS*)packs[b])->GetSize(n);
						tstmp = len ? RCBString(((AttrPackS*)packs[b])->GetVal(n), len) : RCBString("");
						CompareAndSetCurrentMax(tstmp, max);
					}
					else if(PackType() == PackN && !packs[b]->IsNull(n))
					{
						tstmp = RCBString(DecodeValue_S(((AttrPackN*)packs[b])->GetVal(n) + local_min[b]));
						CompareAndSetCurrentMax(tstmp, max);
					}
					f->GetNext1(block, n);
				}
			}
		}
	}
	return max;
}

RCDataTypePtr RCAttrLoadBase::GetMinValuePtr(int pack)		// needed for loader?
{
	AttributeType a_type = TypeName();
	RCDataType* ret = 0;
	if(this->NoNulls() != NoObj() && NoObj() != 0)
	{
		LoadPackInfo();
		if(ATI::IsIntegerType(a_type))
			ret = new RCNum(local_min[pack]);
		else if(ATI::IsDateTimeType(a_type))
			ret = new RCDateTime(local_min[pack], a_type);
		else if(ATI::IsRealType(a_type))
			ret = new RCNum(local_min[pack], 0, true);
		else if(a_type == RC_NUM)
			ret = new RCNum(local_min[pack], Type().GetScale());
		else
		{
			Filter f(NoObj() - (static_cast<_uint64>(packs_omitted) << 16));
			f.SetBlock(pack);
			ret = new RCBString(MinS(&f));
		}
	}
	return RCDataTypePtr(ret);
}

RCDataTypePtr RCAttrLoadBase::GetMaxValuePtr(int pack)		// needed for loader?
{
	AttributeType a_type = TypeName();
	RCDataType* ret = 0;
	if(this->NoNulls() != NoObj() && NoObj() != 0)
	{
		LoadPackInfo();
		if(ATI::IsIntegerType(a_type))
			ret = new RCNum(local_max[pack]);
		else if(ATI::IsDateTimeType(a_type))
			ret = new RCDateTime(local_max[pack], a_type);
		else if(ATI::IsRealType(a_type))
			ret = new RCNum(local_max[pack], 0, true);
		else if(a_type == RC_NUM)
			ret = new RCNum(local_max[pack], Type().GetScale());
		else
		{
			Filter f(NoObj() - (static_cast<_uint64>(packs_omitted) << 16));
			f.SetBlock(pack);
			ret = new RCBString(MaxS(&f));
		}
	}
	return RCDataTypePtr(ret);
}

void RCAttrLoadBase::GetMinMaxValuesPtrs(int pack, RCDataTypePtr& out_min, RCDataTypePtr& out_max)		// needed for loader?
{
	out_min = GetMinValuePtr(pack);
	out_max = GetMaxValuePtr(pack);
}


void RCAttrLoadBase::LogWarnigs()
{
	if(illegal_nulls) {
		stringstream ss;
		ss << "WARNING: There was an attempt to insert NULL to column " << this->Name() << " that is defined as NOT NULL.";
		rclog << lock << ss.str() << unlock;

		ss.str("");
		ss << "         NULLs were changed to ";

		if(ATI::IsStringType(TypeName()))
			ss << "empty string.";
		else if(ATI::IsNumericType(TypeName()))
			ss << RCNum(0, Type().GetScale(), ATI::IsRealType(TypeName()), TypeName()).ToRCString() << ".";
		else
			ss << RCDateTime(0, TypeName()).ToRCString() << ".";

		rclog << lock << ss.str() << unlock;
	}
}

void RCAttrLoadBase::UpdateRSI_Hist(int pack)		// make the index up to date for the selected pack
{
	if(PackType()!=PackN || rsi_manager==NULL || NoObj()==0) return;
	// Note that GetIndxForUpdate will create a new index, if it does not exist.
	// Assuming that if UpdateRSI_Hist() is executed, then we really should have an index for this attr.
	// Allows updating the histogram in the random order.

	if(rsi_hist_update==NULL)
		rsi_hist_update = (RSIndex_Hist*)rsi_manager->GetIndexForUpdate(RSIndexID(RSI_HIST,table_number,attr_number),GetCurReadLocation());

	if(rsi_hist_update==NULL)
	{
		rccontrol << lock << "Warning: cannot access RSI_Hist" << unlock;
		return;
	}
	LoadPackInfo();
	if(rsi_hist_update->NoObj()==0)
	{
		rsi_hist_update->Create(NoObj(),(ATI::IsRealType(TypeName()) ? false : true));	// the new index
	}
	else
	{
		rsi_hist_update->Update(NoObj());		// safe also if no_obj is the same as a while ago
	}
	rsi_hist_update->ClearPack(pack+packs_omitted);		// invalidate the previous content of the pack
	_int64 pmin = local_min[pack];
	_int64 pmax = local_max[pack];
	BHASSERT(pmin != PLUS_INF_64, "should be 'pmin != PLUS_INF_64'");
	if(pmin==pmax || GetPackOntologicalStatus(pack) == NULLS_ONLY) return;

	//LockPackForUse(pack);
//	if(pack_mode[pack] != 2) LoadPack(pack);	// do not mark this pack access, as it is optional
	_uint64 obj_start = (_uint64(pack)<<16);
	_uint64 obj_stop = obj_start+_uint64(loc_no_obj[pack])+1;
	for(_uint64 obj=obj_start ; obj<obj_stop ; obj++)
		rsi_hist_update->PutValue(GetValueInt64(obj),pack+packs_omitted,pmin,pmax);
	//UnlockPackFromUse(pack);
	// NOTE: we must update and delete index after each load, because someone may use it in a query before commit.
	// Updating should be performed externally (after load) to avoid too frequent disk access.
}

void RCAttrLoadBase::UpdateRSI_CMap(int pack, bool new_prefix)		// make the index up to date for the selected pack
{
	if(PackType()!=PackS || rsi_manager==NULL || NoObj()==0 || RequiresUTFConversions(Type().GetCollation()) )
		return;
	if(rsi_cmap_update==NULL)
		rsi_cmap_update = (RSIndex_CMap*)rsi_manager->GetIndexForUpdate(RSIndexID(RSI_CMAP,table_number,attr_number),GetCurReadLocation());

	if(rsi_cmap_update==NULL)
	{
		rccontrol << lock << "Warning: cannot access RSI_Hist" << unlock;
		return;
	}
	LoadPackInfo();
	if(rsi_cmap_update->NoObj()==0)
	{
		rsi_cmap_update->Create(NoObj(),(Type().GetPrecision()>64 ? 64 : Type().GetPrecision()));	// the new index
	}
	else
	{
		rsi_cmap_update->Update(NoObj());		// safe also if no_obj is the same as a while ago
	}
	rsi_cmap_update->ClearPack(pack+packs_omitted);		// invalidate the previous content of the pack

	if(GetPackOntologicalStatus(pack) == NULLS_ONLY)
		return;

	//LockPackForUse(pack);
	_uint64 obj_start = (_uint64(pack)<<16);
	_uint64 obj_stop = obj_start+_uint64(loc_no_obj[pack])+1;
	_uint64 obj = /*new_prefix ? 0 :*/ obj_start;
	int prefix_len = (int)GetPrefix(pack).size();
	for(; obj < obj_stop ; obj++)
		if(!IsNull(obj))
			rsi_cmap_update->PutValue(GetValueString(obj,false) += prefix_len, pack+packs_omitted);			// false - because we don't want to see binary as hex
	// NOTE: we must update and delete index after each load, because someone may use it in a query before commit.
	// Updating should be performed externally (after load) to avoid too frequent disk access.
}


