/* Copyright (C) 2005-2008 Infobright Inc. */

#ifndef LOADER_RCATTRLOADBASE_H_INCLUDED
#define LOADER_RCATTRLOADBASE_H_INCLUDED

#include "common/mysql_gate.h"
#include "core/RCAttr.h"
#include "system/ib_system.h"

class RCAttrLoadBase : public RCAttr
{
public:
	RCAttrLoadBase(int a_num,int t_num,char *a_path,int conn_mode=0,unsigned int s_id=0, DTCollation collation = DTCollation()) throw(DatabaseRCException);
	~RCAttrLoadBase();
	void LoadPackInfoForLoader();
	void ExpandDPNArrays();
	void SaveDPN();
	template <class T> int LoadData(T* nvs, bool copy_forced);

	void UnlockPackFromUse(unsigned pack_no);
	void LockPackForUse(unsigned pack_no);
	void LoadPack(uint n);
	int Save();
	void Append(boost::shared_ptr<DPN> dp);
	void Append(boost::shared_ptr<DataPack> dp);
	void CreateNewPackage();
    static DWORD WINAPI WINAPI SavePackThread(void *params);
    void UpdateRSI_Hist(int pack);
    void UpdateRSI_CMap(int pack, bool new_prefix = true);
    void threadJoin(int n);
    void threadCancel(int n);
    inline void WaitForSaveThreads()
    {
        DoWaitForSaveThreads();
    }

    inline int SavePack(int n)
    {
        return DoSavePack(n);
    }

protected:
    virtual int DoSavePack(int n) =0;
    virtual void DoWaitForSaveThreads() =0;
    void LogWarnigs();
    RCDataTypePtr GetMinValuePtr(int pack);
    RCDataTypePtr GetMaxValuePtr(int pack);
    void GetMinMaxValuesPtrs(int pack, RCDataTypePtr & out_min, RCDataTypePtr & out_max);
    void SetPackMax(uint pack, RCBString& s);
    void SetPackMin(uint pack, RCBString& s);
    inline uint RoundUpTo8Bytes(RCBString& s);
private:
    RCBString MinS(Filter *f);
    RCBString MaxS(Filter *f);
    inline void CompareAndSetCurrentMin(RCBString tstmp, RCBString & min, bool set);
    inline void CompareAndSetCurrentMax(RCBString tstmp, RCBString & min);

private:
    IBMutex rsi_update_mutex;
    static IBCond threads_count_cond;
protected:
    static int count_spt;
    bool illegal_nulls;
};

#endif /* not LOADER_RCATTRLOADBASE_H_INCLUDED */
