/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include <boost/bind.hpp>

#ifndef __BH_COMMUNITY__
#	include "enterprise/local.h"
#	include "enterprise/loader/RCAttr_load.h"
#else /* not __BH_COMMUNITY__ */
#	include "community/local.h"
#	include "community/loader/RCAttr_load.h"
#endif /* __BH_COMMUNITY__ */
#include "DataLoader.h"
#include "RCTable_load.h"
#include "system/Buffer.h"
#include "core/tools.h"

using namespace std;

RCTableLoad::RCTableLoad(const char *a_path, int t_n, int current_state, int use_path_only, vector<DTCollation> charsets) throw(DatabaseRCException)
	: RCTableImpl(a_path, charsets, t_n, current_state, use_path_only, false, true), no_loaded_rows(0)
{

	//InitializeCriticalSection(&synchr);
	//pthread_mutexattr_t mattr;
	//pthread_mutexattr_init(&mattr);
	//pthread_mutexattr_settype(&mattr, PTHREAD_MUTEX_RECURSIVE);
	//pthread_mutex_init(&synchr, &mattr);
	if(no_attr > 0) {
		try {
			LoadAttribute(0);
		} catch(DatabaseRCException& e) {
			throw e;
		}

		m_update_time = a[0]->UpdateTime();
	} else 
		m_update_time = m_create_time;
}

void RCTableLoad::WaitForSaveThreads()
{
	for(int at = 0; at < no_attr; at++) {
		if(a[at])
			((RCAttrLoad*)a[at])->WaitForSaveThreads();
	}
}

void RCTableLoad::LoadData(IOParameters& iop)
{
	LargeBuffer rclb;
	vector<RCAttrLoad*> attrs;
	for(uint i = 0; i < NoAttrs(); i++)	{
		LoadAttribute(i);
		((RCAttrLoad*)a[i])->SetLargeBuffer(&rclb);
		((RCAttrLoad*)a[i])->LoadPackInfoForLoader();
		attrs.push_back((RCAttrLoad*)a[i]);
	}
	LockPackInfoForUse();
	_uint64 prev_no_obj = NoObj();
	try {
		rclb.BufOpen(iop, READ);
		if((rclb.BufStatus() != 1 && rclb.BufStatus() != 4 && rclb.BufStatus() != 5))
			throw FileRCException("Unable to open " + (LargeBuffer::IsPipe(iop.Path()) ? string("pipe ") : string("file ")) + string(iop.Path()) + string("."));
		if(rclb.BufSize()) {
			auto_ptr<DataLoader> loader = DataLoader::CreateDataLoader(attrs, rclb, iop);
			no_loaded_rows = loader->Proceed();
		}
		rclb.BufClose();
	} catch(...) {
		WaitForSaveThreads();
		rclb.BufClose();
		UnlockPackInfoFromUse();
		for_each(attrs.begin(), attrs.end(), boost::bind(&RCAttrLoad::DeleteRSIs, _1));
		throw;
	}
	WaitForSaveThreads();
	UnlockPackInfoFromUse();
}

void RCTableLoad::LoadData(IOParameters& iop, Buffer& buffer)
{
	vector<RCAttrLoad*> attrs;
	for(uint i = 0; i < NoAttrs(); i++)	{
		LoadAttribute(i);
		((RCAttrLoad*)a[i])->LoadPackInfoForLoader();
		attrs.push_back((RCAttrLoad*)a[i]);
	}
	LockPackInfoForUse();
	_uint64 prev_no_obj = NoObj();
	try {
		auto_ptr<DataLoader> loader = DataLoader::CreateDataLoader(attrs, buffer, iop);
		loader->Proceed();
	} catch(...) {
		UnlockPackInfoFromUse();
		for_each(attrs.begin(), attrs.end(), boost::bind(&RCAttrLoad::DeleteRSIs, _1));
		throw;
	}
	UnlockPackInfoFromUse();
}

RCTableLoad::~RCTableLoad()
{
	// Note:: there was not critical section destroy function in the windows version of this destructor. ???
	// we added the destroy in the linux port.
	//pthread_mutex_destroy(&synchr);
}

inline void RCTableLoad::LoadAttribute(int attr_no)
{
	if(!a[attr_no])
	{
		//EnterCriticalSection(&synchr);
		//pthread_mutex_lock(&synchr);
        table_impl_mutex.Lock();
		if(!a[attr_no]) {
			try {
				a[attr_no] = new RCAttrLoad(attr_no, tab_num, path, conn_mode, session, charsets[attr_no]);
				if (a[attr_no]->OldFormat()) {
					a[attr_no]->UpgradeFormat();
					delete a[attr_no];
					a[attr_no] = new RCAttrLoad(attr_no, tab_num, path, conn_mode, session, charsets[attr_no]);
				}
			} catch(DatabaseRCException& e) {
				throw e;
			}
		}
		//LeaveCriticalSection(&synchr);
        table_impl_mutex.Unlock();
		//pthread_mutex_unlock(&synchr);
	}
}

_int64 RCTableLoad::NoRecordsLoaded()
{
	return no_loaded_rows;
}
