/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _MEMORY_BLOCK_H_E5FF7A96_7174_421b_9FE2_043859D08D15
#define _MEMORY_BLOCK_H_E5FF7A96_7174_421b_9FE2_043859D08D15

#include <cstddef>

#ifdef __GNUC__
enum BLOCK_TYPE
{
	BLOCK_FREE,
	BLOCK_COMPRESSED,
	BLOCK_UNCOMPRESSED,
	BLOCK_TEMPORARY,
	BLOCK_FIXED //Used in rc_realloc when pointer != NULL
};
#else
enum BLOCK_TYPE: char
{
	BLOCK_FREE,
	BLOCK_COMPRESSED,
	BLOCK_UNCOMPRESSED,
	BLOCK_TEMPORARY,
	BLOCK_FIXED //Used in rc_realloc when pointer != NULL
};
#endif

class TrackableObject;

template<typename MEM_HANDLE>
class MemoryBlock
{
public:
	MemoryBlock();
	virtual ~MemoryBlock();

	/*
		split this block into two blocks:
		next block = [this block].size - size
		[this block].size = new_size
	*/
	MemoryBlock<MEM_HANDLE>* Split(size_t new_size);

	/*
		Merge this block with given block if possible
	*/
	bool Merge(MemoryBlock<MEM_HANDLE>* blck);

	void setSize(	size_t size){ this->size = size;}
	void setAddress(MEM_HANDLE addr){this->address = addr;}
	void setPrev(	MemoryBlock* blck){this->prev = blck;}
	void setNext(	MemoryBlock* blck){this->next = blck;}
	void setType(	BLOCK_TYPE b_type){this->type = b_type;}
	void setOwner( TrackableObject* owner){ this->owner = owner;}

	size_t getSize(){return this->size;}
	MEM_HANDLE getAddress(){return address;}
	MemoryBlock<MEM_HANDLE>* getPrev(){return this->prev;}
	MemoryBlock<MEM_HANDLE>* getNext(){return this->next;}
	BLOCK_TYPE getType(){return this->type;}
	TrackableObject* getOwner(){return owner;}

private:

	size_t size; //Block's size
	MEM_HANDLE address; //starting address
	MemoryBlock<MEM_HANDLE> *prev, *next; //previous and next memory block in the memory
	BLOCK_TYPE type;
	TrackableObject* owner; //Owner of the memory block
};

#endif

