/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _MEMORY_HANDLING_POLICY_13AA8DDA_9EC9_4396_8DCB_40910E9FC771
#define _MEMORY_HANDLING_POLICY_13AA8DDA_9EC9_4396_8DCB_40910E9FC771

#include <map>

#ifdef __GNUC__
#include <ext/hash_map>
#include <ext/hash_set>
#include "system/linux/hash_set_ext.h"
#ifndef stdext
#define stdext __gnu_cxx
#endif
#else
#include <hash_map>
#include <hash_set>
#endif

#include "MemoryBlock.h"
#include "BasicHeapPolicy.h"
#include "common/bhassert.h"
#include "system/ib_system.h"

typedef void* MEM_HANDLE_MP;

class system_heap_element
{
public:
	system_heap_element(void* pointer, size_t size) :
		pointer(pointer), size(size)
	{
	}

	bool operator==(const system_heap_element& anotherElement) const
	{
		return this->pointer == anotherElement.pointer;
	}

	bool operator<(const system_heap_element& anotherElement) const
	{
		return this->pointer < anotherElement.pointer;
	}

	void* GetPointer() const
	{
		return pointer;
	}

	size_t GetSize() const
	{
		return size;
	}

private:
	void*	pointer;
	size_t	size;
};

typedef system_heap_element MEM_HANDLE_MP_WITH_SIZE;

typedef BasicHeap HeapPolicy_MP;
int cmp_eval_func(unsigned short cmp, unsigned short free);
int uncmp_eval_func(unsigned short uncmp, unsigned short uptodate, unsigned short free);

class CountingCriticalSection
{
public:

	CountingCriticalSection():counter(0)
	{
		//pthread_mutexattr_t mattr;
		//pthread_mutexattr_init(&mattr);
		//pthread_mutexattr_settype(&mattr, PTHREAD_MUTEX_RECURSIVE);
		//pthread_mutex_init(&cs, &mattr);
	}
	virtual ~CountingCriticalSection(void)
	{
		//pthread_mutex_destroy(&cs);
	}

	void EnterCountingCriticalSection()
	{
		//pthread_mutex_lock(&cs);
        counting_mutex.Lock();
	}

	void LeaveCountingCriticalSection()
	{
		//pthread_mutex_unlock(&cs);
        counting_mutex.Unlock();
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(counter >= 0);
	}

	unsigned int counter;

    IBMutex counting_mutex;
	//pthread_mutex_t cs; // pointer to the memory manager's critical section
};

class MemoryHandling
{
public:

	MemoryHandling(size_t comp_heap_size, size_t uncomp_heap_size, size_t cmp_rls, size_t uncomp_rls, std::string hugedir = "");
	virtual ~MemoryHandling(void);

	MEM_HANDLE_MP	alloc(	size_t size, BLOCK_TYPE type, TrackableObject* owner);
	void		dealloc(MEM_HANDLE_MP mh, TrackableObject* owner);
	MEM_HANDLE_MP	rc_realloc(MEM_HANDLE_MP mh, size_t size, TrackableObject* owner, BLOCK_TYPE type);
	size_t		rc_msize(MEM_HANDLE_MP mh, TrackableObject* owner); //returns size of a memory block represented by [mh]

#ifdef __GNUC__
	enum RELEASE_HEAP
	{
		RH_UNCOMPRESSED,
		RH_COMPRESSED
	};
#else
	enum RELEASE_HEAP:char
	{
		RH_UNCOMPRESSED,
		RH_COMPRESSED
	};
#endif

	bool ReleaseMemory(size_t, RELEASE_HEAP, TrackableObject* untouchable); //Release given amount of memory
	void ReleaseUnlockedDataPacks();

	void CleanPacks();
	void CompressPacks();

	void ReportLeaks();

	int MemorySettingsScale()						// Memory scale factors
	{
		if(main_heap_MB<500)	return 0;		// up to 0.5 GB
		if(main_heap_MB<1200)	return 1;		// up to 1.2 GB
		if(main_heap_MB<2500)	return 2;		// up to 2.5 GB
		if(main_heap_MB<5000)	return 3;		// up to 5 GB
		if(main_heap_MB<10000)	return 4;		// up to 10 GB
		if(main_heap_MB<20000)	return 5;		// up to 20 GB
		return 99;
	};

private:

	size_t	comp_heap_release;
	size_t	uncomp_heap_release;
	int		main_heap_MB;				// ServerMainHeapSize, in MB

	/*
	auxiliary class used in memory release algorithm
	to represent selected region
	*/
	struct MemoryRegion
	{
		MemoryBlock<MEM_HANDLE_MP>* start, *end;

		union HeapUnits
		{
			struct compressed_heap_units
			{
				unsigned short free_units;
				unsigned short compressed_units;
			};
			struct uncompressed_heap_units
			{
				unsigned short free_units;
				unsigned short uncompressed_units; //To be compressed units
				unsigned short uncompressed_up_to_date_units; //Compression was done = free units
			};

			compressed_heap_units cmp_units;
			uncompressed_heap_units uncmp_units;
		}heapunits;

//		union HeapUnits ;
		size_t size;
	};

	template<class T> class rc_hash_compare_for_mm
	{
	private:
		typedef T Key;
	public:
		static const size_t bucket_size = 4;
		static const size_t min_buckets = 1024;//1048576;
		size_t operator()(const Key& k) const
		{
			return (size_t)(k.GetPointer());
		};

#ifdef __GNUC__
		bool operator()(const Key& k1, const Key& k2) const
		{
			return k1 == k2;
		};
#else
		bool operator()(const Key& k1, const Key& k2) const
		{
			return k1 < k2;
		};
#endif
	};
#ifdef __GNUC__
	typedef stdext::hash_set<MEM_HANDLE_MP_WITH_SIZE, rc_hash_compare_for_mm<MEM_HANDLE_MP_WITH_SIZE>, rc_hash_compare_for_mm<MEM_HANDLE_MP_WITH_SIZE> > hash_set_for_system_mem_heaps;
#else
	// In windows hash_set function parameters differ in order and in number.
	typedef stdext::hash_set<MEM_HANDLE_MP_WITH_SIZE, rc_hash_compare_for_mm<MEM_HANDLE_MP_WITH_SIZE>> hash_set_for_system_mem_heaps;
#endif

	typedef std::pair<MEM_HANDLE_MP, BLOCK_TYPE> alloc_blck_pair;
	typedef std::map<MEM_HANDLE_MP, BLOCK_TYPE> AllocBlckMap;
	typedef std::map<TrackableObject*, stdext::hash_set<MEM_HANDLE_MP>* > OwnerMemoryMap;
	typedef std::map<TrackableObject*, hash_set_for_system_mem_heaps* > OwnerMemoryMapSystem;
	typedef std::pair<TrackableObject*, stdext::hash_set<MEM_HANDLE_MP>* > mbrshp_pair;

	HeapPolicy_MP uncompressed_heap;
	HeapPolicy_MP compressed_heap;

	MemoryBlock<MEM_HANDLE_MP>* uncomp_heap_mem;
	MemoryBlock<MEM_HANDLE_MP>* comp_heap_mem;

	AllocBlckMap alloc_blck_type; //Type of blocks allocated on heaps
	OwnerMemoryMap cmp_blck_mbrshp; // Compressed Memory blocks membership (used in memory release algorithm)
	OwnerMemoryMap uncmp_blck_mbrshp; // Uncompressed memory blocks membership (used in memory release algorithm)

	int (*cmp_eval) (unsigned short, unsigned short);
	int (*uncmp_eval) (unsigned short, unsigned short, unsigned short);

	/*
		Map of pointers that were allocated using system heap
		This is a last resort procedure when there is no possibility to
		release memory from heaps but memory is needed
		Objects from the map will be released (try ti be) every time ReleseMemory function is invoked
	*/
	std::vector<MEM_HANDLE_MP_WITH_SIZE> system_heap_alloc;
	OwnerMemoryMapSystem	system_heap_map;

	/*
		Code won't compile if heap definition is not consistent with memory handler definition
	*/
	MEM_HANDLE_MP check_consistencyA(){MEM_HANDLE_MP* mh = 0; return uncompressed_heap.check_consistency(mh);}
	MEM_HANDLE_MP check_consistencyB(){MEM_HANDLE_MP* mh = 0; return compressed_heap.check_consistency(mh);}


	//Memory leak test function
	void UncmpMemoryLeaks(TrackableObject*);
	void CmpMemoryLeaks(TrackableObject*);

public:
	CountingCriticalSection ccs;
	CountingCriticalSection* GetCriticalSection(){return &ccs;}

private:
	static bool FindAndErase(std::vector<MEM_HANDLE_MP_WITH_SIZE>& vec, MEM_HANDLE_MP_WITH_SIZE element);
};

#ifdef __GNUC__
namespace __gnu_cxx
{

template<>
  struct hash<TrackableObject*>
  {
    size_t
    operator()(const TrackableObject* __x) const
    { return hash<size_t>()( (size_t) __x); }
  };

}
#endif

#endif

