/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "SplayTree.h"
#include "system/RCSystem.h"
#include "MemoryBlock.h"

template <typename Key, typename Obj>
SplayTree<Key, Obj>::SplayTree()
{
	this->root = NULL;
}

template <typename Key, typename Obj>
SplayTree<Key, Obj>::SplayTree(SplayTree<Key,Obj> &splyTree)
{
}

template <typename Key, typename Item>
SplayTree<Key, Item>::~SplayTree(void)
{
	if(root)
		delete root;
}

template <typename Key, typename Item>
void SplayTree<Key, Item>::insert(Key key, Item item)
{

	if(!root)
	{
		SplayNode<Key, Item>* new_node = new SplayNode<Key, Item>();
		new_node->key = key;
		new_node->items->push_front(item);
		new_node->parent = NULL;
		new_node->left = NULL;
		new_node->right = NULL;

		root = new_node;
		return;
	}

	SplayNode<Key, Item>* node = root;
	SplayNode<Key, Item>* parent = NULL;

	while(node)
	{
		parent = node;
		if(node->key < key) node = node->right;
		else if(node->key > key) node = node->left;
		else if(node->key == key) break; //insert Item into existing node's list
		else assert(0);
	}

	if (parent->key != key)
	{
		SplayNode<Key, Item>* new_node = new SplayNode<Key, Item>();
		new_node->key = key;
		new_node->items->push_front(item);
		new_node->parent = NULL;
		new_node->left = NULL;
		new_node->right = NULL;

		if(new_node->key <= parent->key)
			parent->left = new_node;
		else
			parent->right = new_node;
		new_node->parent = parent;
	}
	else
	{
		parent->items->push_front(item);
	}

}

template <typename Key, typename Item>
Item SplayTree<Key, Item>::remove(Key key)
{
	SplayNode<Key, Item>* node = root;
	while(node)
	{
		if(node->key == key) break;
		else if(node->key < key) node = node->right;
		else node = node->left;
	}
	if(!node) // Node was not found
		return NULL;

	assert(node->items->size() == 1);
	return deleteNode(node);
}
template<typename Key, typename Item>
void SplayTree<Key, Item>::removeEx(Key key, Item item)
{
	SplayNode<Key, Item>* node = root;
	while(node)
	{
		if(node->key == key) break;
		else if(node->key < key) node = node->right;
		else node = node->left;
	}
	if(!node) // Node was not found
		return /*NULL*/;
	else
	{
		if(node->items->size() == 1)
		{
			assert(item == node->items->front());
			/*return*/ deleteNode(node);
		}
		else
			node->items->remove(item);
		//return item;
	}
}
template<typename Key, typename Item>
Item SplayTree<Key, Item>::deleteNode(SplayNode<Key, Item>* node)
{
	if(!node)
		return NULL;

	Item item = NULL;
	if(node->items)
	{
		item = node->items->front();
		node->items->pop_front();
		if (node->items->size() > 0)
		{
			return item;
		}
	}

	//Remove node if last item was removed from the list

	SplayNode<Key, Item>* temp = node;
	if(node->left == NULL && node->right == NULL)
	{
		if(node == root)
		{
			delete root;
			root = NULL;
		}
		else
		{
			delete node;
			node = NULL;
		}
	}
	else if (node->left == NULL && node->right != NULL)
	{
		node->right->parent = node->parent;
		if(!node->parent)
		{
			root = node->right;
		}
		else
		{
			if(node->parent->key >= node->right->key)
				node->right->parent->left = node->right;
			else
				node->right->parent->right = node->right;
		}

		node->left = NULL;
		node->right = NULL;
		node->parent = NULL;
		delete node;
		node = NULL;

	}
	else if (node->right == NULL && node->left != NULL)
	{
		node->left->parent = node->parent;

		if(!node->parent)
		{
			root = node->left;
		}
		else
		{
			if(node->parent->key >= node->left->key)
				node->left->parent->left = node->left;
			else
				node->left->parent->right = node->left;
		}

		node->left = NULL;
		node->right = NULL;
		node->parent = NULL;
		delete node;
		node = NULL;
	}
	else
	{
		// Node has two children - get max of left subtree
		temp = node->left;
		while (temp->right != NULL)
			temp = temp->right;

		delete node->items;
		node->items = temp->items;
		node->key = temp->key;
		temp->items = NULL;

		deleteNode(temp);
	}
	return item;
}

template <typename Key, typename Item>
bool SplayNode<Key, Item>::operator ==(const SplayNode<Key, Item>& node)
{
	return (node.key == this->key ? true : false);
}
template <typename Key, typename Item>
bool SplayNode<Key, Item>::operator !=(const SplayNode<Key, Item>& node)
{
	return (node.key == this->key ? false : true);
}

template<typename Key, typename Item>
bool SplayTree<Key, Item>::areTreesEqual(SplayNode<Key, Item>* nodeA, SplayNode<Key, Item>* nodeB)
{
	if(nodeA != nodeB)
		return false;
	else
	{
		bool left = areTreesEqual(nodeA->left, nodeB->left);
		bool right = areTreesEqual(nodeA->right, nodeB->right);

		return (left && right);
	}
}

template <typename Key, typename Item>
bool SplayTree<Key, Item>::operator ==(const SplayTree<Key, Item>& tree)
{
	return areTreesEqual(this->root, tree.root);
}

template <typename Key, typename Obj>
void SplayTree<Key, Obj>::zig(SplayNode<Key, Obj> *node)
{
	//rccontrol << lock << "zig" << unlock;
	/*
	If parent(node) is the root, rotate the edge joining node
	with parent(node)
	*/

	if(root->key < node->key)
	{
		assert(root->right == node);

		root->right = node->left;
		if(root->right)
			root->right->parent = root;
		node->left = root;
		root->parent = node;
		node->parent = NULL;
	}
	else
	{
		assert(root->left == node);

		root->left = node->right;
		if(root->left)
			root->left->parent = root;
		node->right = root;
		root->parent = node;
		node->parent = NULL;
	}
	root = node;
}

template <typename Key, typename Obj>
void SplayTree<Key, Obj>::zigzig(SplayNode<Key, Obj> *node)
{
	/*
	If parent(node) is not the root and node and parent(node)
	are both left or both right children, rotate the edge joining
	parent(node) with its grandparent and then rotate the edge
	joining node with parent(node)
	*/

	//rccontrol << lock << "zigzig" << unlock;

	SplayNode<Key, Obj>* parent = node->parent;
	SplayNode<Key, Obj>* grandparent = parent->parent;

	if(parent->key < node->key)
	{
		parent->right = node->left;
		if(parent->right)
			parent->right->parent = parent;

		grandparent->right = parent->left;
		if(grandparent->right)
			grandparent->right->parent = grandparent;

		parent->left = grandparent;
		node->parent = grandparent->parent;

		if(grandparent->parent)
		{
			if (grandparent->parent->key >= node->key)
				grandparent->parent->left = node;
			else
				grandparent->parent->right = node;
		}

		grandparent->parent = parent;
		parent->parent = node;
		node->left = parent;

	}
	else
	{
		parent->left = node->right;
		if(parent->left)
			parent->left->parent = parent;

		grandparent->left = parent->right;
		if(grandparent->left)
			grandparent->left->parent = grandparent;

		parent->right = grandparent;
		node->parent = grandparent->parent;

		if(grandparent->parent)
		{
			if (grandparent->parent->key >= node->key)
				grandparent->parent->left = node;
			else
				grandparent->parent->right = node;
		}

		grandparent->parent = parent;
		parent->parent = node;
		node->right = parent;
	}
	if(grandparent == root)
		root = node;
}

template <typename Key, typename Obj>
void SplayTree<Key, Obj>::zigzag(SplayNode<Key, Obj> *node)
{
	/*
	If parent(node) is not the root and node is a left and
	parent(node) a right child, or vice versa, rotate the edge
	joining node with parent(node)and then rotate the edge
	joining node with the new parent(node)
	*/

	//rccontrol << lock << "zigzag" << unlock;

	SplayNode<Key, Obj>* parent = node->parent;
	SplayNode<Key, Obj>* grandparent = parent->parent;

	if(parent->key >= node->key && parent->key > grandparent->key)
	{
		parent->left = node->right;
		if(parent->left)
			parent->left->parent = parent;
		node->right = parent;
		node->parent = grandparent->parent;

		if(grandparent->parent)
		{
			if (grandparent->parent->key >= node->key)
				grandparent->parent->left = node;
			else
				grandparent->parent->right = node;
		}

		grandparent->parent = node;
		parent->parent = node;
		grandparent->right = node->left;
		if(grandparent->right)
			grandparent->right->parent = grandparent;
		node->left = grandparent;
	}
	else if(parent->key < node->key && parent->key <= grandparent->key)
	{
		parent->right = node->left;
		if(parent->right)
			parent->right->parent = parent;
		node->left = parent;
		node->parent = grandparent->parent;

		if(grandparent->parent)
		{
			if (grandparent->parent->key >= node->key)
				grandparent->parent->left = node;
			else
				grandparent->parent->right = node;
		}

		grandparent->parent = node;
		parent->parent = node;
		grandparent->left = node->right;
		if(grandparent->left)
			grandparent->left->parent = grandparent;
		node->right = grandparent;
	}
	else
	{
		assert(0);
	}
	if(grandparent == root)
		root = node;

}

/*
Find node with node.key == key or with the smallest key
bigger than key.
*/
template <typename Key, typename Obj>
void SplayTree<Key, Obj>::splay(Key key)
{
	SplayNode<Key, Obj>* node = root;
	SplayNode<Key, Obj>* last = NULL;
	while(node)
	{
		//Take first node that contains the smallest values which satisfies the condition
		if(node->key > key)
		{
			if(!last)
				last = node;
			else if (last->key > node->key)
				last = node;
		}
		if(node->key == key) break;
		else if(node->key < key) node = node->right;
		else node = node->left;
	}
	if(!node)
	{
		node = last;
		if(!node)
		{
			//assert(0);
			return; //There was no node that has key bigger than [key]
		}
	}
	if(node == root)
		return; //node is root = no splaying is needed

	/*
	SplayNode<Key, Obj>* parent = NULL;
	do
	{
		parent = node->parent;
		if(node->parent)
		{
			if( (node->parent) == root )
				zig(node);
			else if( node->parent != root &&
				(node->parent->key >= node->key && node->parent->parent->key >= node->parent->key) ||
				(node->parent->key < node->key && node->parent->parent->key < node->parent->key))
				zigzig(node);
			else if( node->parent != root &&
				(parent->key < node->key && parent->key <= parent->parent->key)||
				(parent->key >= node->key && parent->key > parent->parent->key))
				zigzag(node);
			else
			{
				assert(0);
			}
		}
	}while(node->parent);
	*/
	SplayNode<Key, Obj>* parent = NULL;
	do
	{
		parent = node->parent;
		if(parent)
		{
			if( parent == root )
				zig(node);
			else if( parent != root &&
				(parent->key >= node->key && parent->parent->key >= parent->key) ||
				(parent->key < node->key && parent->parent->key < parent->key))
				zigzig(node);
			else if( node->parent != root &&
				(parent->key < node->key && parent->key <= parent->parent->key)||
				(parent->key >= node->key && parent->key > parent->parent->key))
				zigzag(node);
			else
			{
				assert(0);
			}
		}
	}while(node->parent);
}

template<typename Key, typename Obj>
SplayNode<Key, Obj>* SplayTree<Key, Obj>::strictSplay(Key key)
{
	SplayNode<Key, Obj>* node = root;
	while(node)
	{
		if(node->key == key) break;
		else if(node->key < key) node = node->right;
		else node = node->left;
	}
	if(!node)
	{
		return NULL;
	}
	if(node == root)
		return root; //node is root = no splaying is needed

	/*
	SplayNode<Key, Obj>* parent = NULL;
	do
	{
		parent = node->parent;
		if(node->parent)
		{
			if( (node->parent) == root )
				zig(node);
			else if( node->parent != root &&
				(node->parent->key >= node->key && node->parent->parent->key >= node->parent->key) ||
				(node->parent->key < node->key && node->parent->parent->key < node->parent->key))
				zigzig(node);
			else if( node->parent != root &&
				(parent->key < node->key && parent->key <= parent->parent->key)||
				(parent->key >= node->key && parent->key > parent->parent->key))
				zigzag(node);
			else
			{
				assert(0);
			}
		}
	}while(node->parent);
	*/
	SplayNode<Key, Obj>* parent = NULL;
	do
	{
		parent = node->parent;
		if(parent)
		{
			if( (parent) == root )
				zig(node);
			else if( parent != root &&
				(parent->key >= node->key && parent->parent->key >= parent->key) ||
				(parent->key < node->key && parent->parent->key < parent->key))
				zigzig(node);
			else if( node->parent != root &&
				(parent->key < node->key && parent->key <= parent->parent->key)||
				(parent->key >= node->key && parent->key > parent->parent->key))
				zigzag(node);
			else
			{
				assert(0);
			}
		}
	}while(node->parent);

	return node;
}

template<typename Key, typename Obj>
SplayNode<Key, Obj>* SplayTree<Key, Obj>::predecessor(SplayNode<Key, Obj>* node)
{
	if(node->left)
	{
		if (node->left->right)
		{
			SplayNode<Key, Obj>* most_right = node->left->right;
			while (most_right->right != NULL)
			{
				most_right = most_right->right;
			}
			return most_right;
		}
		else
			return node->left;
	}
	else
		return NULL;
}

template<typename Key, typename Obj>
Key SplayTree<Key, Obj>::getMostRightKey()
{
	if(root)
	{
		if(root->right)
		{
			SplayNode<Key, Obj>* most_right = root->right;
			while (most_right->right != NULL)
			{
				most_right = most_right->right;
			}
			return most_right->key;
		}
		else
			return root->key;
	}
	return NULL;
}

template class SplayNode<void*, MemoryBlock<void*>* >;
template class SplayNode<size_t, MemoryBlock<void*>* >;

template class SplayTree<void*, MemoryBlock<void*>* >;
template class SplayTree<size_t, MemoryBlock<void*>* >;

