/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "TrackableObject.h"
#include "system/RCSystem.h"
#include "core/RCAttrPack.h"
#include "common/bhassert.h"

#include "core/tools.h"

using namespace std;

//MemoryHandling<void*, BasicHeap<void*> >* TrackableObject::m_MemHandling = NULL;
MemoryHandling* TrackableObject::m_MemHandling = NULL;

void* TrackableObject::alloc(size_t size, BLOCK_TYPE type)
{
	ccs->EnterCountingCriticalSection();
	void* addr = Instance()->alloc(size, type, this);
	ccs->LeaveCountingCriticalSection();
	return addr;
}

void TrackableObject::dealloc(void* ptr)
{
	ccs->EnterCountingCriticalSection();
	Instance()->dealloc(ptr, this);
	ccs->LeaveCountingCriticalSection();
}

void* TrackableObject::rc_realloc(void* ptr, size_t size, BLOCK_TYPE type)
{
	ccs->EnterCountingCriticalSection();
	void* addr = Instance()->rc_realloc(ptr, size, this, type);
	ccs->LeaveCountingCriticalSection();
	return addr;
}
//
//void* TrackableObject::alloc(size_t size, BLOCK_TYPE type, TrackableObject* parent_owner)
//{
//	BHASSERT(NULL != parent_owner, "Wrong TrackableObject::alloc called (parent_owner is NULL)");
//	pthread_mutex_lock(&m_mutex);
//	void* addr = Instance()->alloc(size, type, parent_owner);
//	pthread_mutex_unlock(&m_mutex);
//	return addr;
//}
//void* TrackableObject::rc_realloc(void* ptr, size_t size, TrackableObject* parent_owner, BLOCK_TYPE type)
//{
//	BHASSERT(NULL != parent_owner, "Wrong TrackableObject::rc_realloc called (parent_owner is NULL)");
//	pthread_mutex_lock(&m_mutex);
//	void* addr = Instance()->rc_realloc(ptr, size, parent_owner, type);
//	pthread_mutex_unlock(&m_mutex);
//	return addr;
//}
//void TrackableObject::dealloc(void* ptr, TrackableObject* parent_owner)
//{
//	BHASSERT(NULL != parent_owner, "Wrong TrackableObject::dealloc called (parent_owner is NULL)");
//	pthread_mutex_lock(&m_mutex);
//	Instance()->dealloc(ptr, parent_owner);
//	pthread_mutex_unlock(&m_mutex);
//}

size_t TrackableObject::rc_msize(void *ptr)
{
//#ifdef LOCKFREE_TRACKABLE_OBJ
	ccs->EnterCountingCriticalSection();
//#endif
	size_t size = Instance()->rc_msize(ptr, this);
//#ifdef LOCKFREE_TRACKABLE_OBJ
	ccs->LeaveCountingCriticalSection();
//#endif
	return size;
}

TrackableObject::TrackableObject()
	: m_Locked(0), m_Blocked(0), m_lock_count(0)
{
	//TODO: added, remove it
//	m_MemHandling=NULL;
	ccs = Instance()->GetCriticalSection();

//	pthread_mutexattr_t mattr;
//	pthread_mutexattr_init(&mattr);
//	pthread_mutexattr_settype(&mattr, PTHREAD_MUTEX_RECURSIVE);
//	pthread_mutex_init(&m_mutex, &mattr);

}
TrackableObject::TrackableObject(size_t comp_size, size_t uncomp_size, size_t cmp_rls, size_t uncmp_rls, string hugedir)
	: m_Locked(0), m_Blocked(0), m_lock_count(0)
{
	ccs = Instance(comp_size, uncomp_size, cmp_rls, uncmp_rls, hugedir)->GetCriticalSection();
//	pthread_mutexattr_t mattr;
//	pthread_mutexattr_init(&mattr);
//	pthread_mutexattr_settype(&mattr, PTHREAD_MUTEX_RECURSIVE);
//	pthread_mutex_init(&m_mutex, &mattr);
}

TrackableObject::~TrackableObject()
{

}

void TrackableObject::Lock()
{
	if(TrackableType() == TO_PACK)
	{
		ccs->EnterCountingCriticalSection();
		m_lock_count++;
		m_Locked = true;
		ccs->LeaveCountingCriticalSection();

		//Locking is done externally
		if(!((AttrPack*)this)->IsEmpty() && IsCompressed())
			Uncompress();
	}
	else
	{
		//EnterCriticalSection(cs);
		ccs->EnterCountingCriticalSection();
		m_lock_count++;
		m_Locked = true;
		ccs->LeaveCountingCriticalSection();
		//LeaveCriticalSection(cs);
	}
	if(m_lock_count == 32766)
	{
		//printf("lock count is very large\n");
		rccontrol << lock << "TrackableObject locked to many times" << unlock;
		BHERROR("TrackableObject locked too many times");
	}
}

void TrackableObject::Unlock()
{
	ccs->EnterCountingCriticalSection();
	m_lock_count--;
	if(m_lock_count < 0)
	{
//		rclog << lock << "Internal error: An object unlocked too many times in memory manager." << unlock;
		m_lock_count = 0;
		ccs->LeaveCountingCriticalSection();
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(!"Internal error: An object unlocked too many times in memory manager.");
	}
	m_lock_count ? m_Locked = true : m_Locked = false;
	ccs->LeaveCountingCriticalSection();
}

void TrackableObject::ConstructionLock()
{
	//All objects enter critical section here
	//EnterCriticalSection(cs);
	ccs->EnterCountingCriticalSection();
	m_lock_count++;
	//if(TrackableType() == TO_PACK)
		//printf("cons locking: %d\n",m_lock_count);
	m_Locked = true;
	ccs->LeaveCountingCriticalSection();
	//LeaveCriticalSection(cs);
	if(m_lock_count == 32766)
	{
		rccontrol << lock << "TrackableObject locked to many times" << unlock;
		BHERROR("TrackableObject locked too many times");
	}
}

void TrackableObject::ConstructionUnlock()
{
	ccs->EnterCountingCriticalSection();
	m_lock_count--;
	//if(TrackableType() == TO_PACK)
		//printf("cons ulocking: %d\n",m_lock_count);
	if(m_lock_count < 0)
	{
		m_lock_count = 0;
		//BHERROR("TrackableObject unlocked too many times");
	}
	m_lock_count ? m_Locked = true : m_Locked = false;
	ccs->LeaveCountingCriticalSection();
}
void TrackableObject::DestructionLock()
{
	ccs->EnterCountingCriticalSection();
	m_Locked = true;
	ccs->LeaveCountingCriticalSection();
}

