/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifdef _SYSTEM_WRSYNC_H_

#include "RCException.h"
#include "RCSystem.h"
#include "WriteSync.h"
#include "core/tools.h"

WriteSync::WriteSync(void):m_nReadingLocksByWriter(0)
{
	pthread_mutexattr_t mattr;
	pthread_mutexattr_init(&mattr);
	pthread_mutexattr_settype(&mattr, PTHREAD_MUTEX_RECURSIVE);
	pthread_mutex_init(&m_WritersCS, &mattr);
	pthread_mutex_init(&m_ReadersCS, &mattr);

#ifdef __GNUC__
	int retCode = sem_init ( &m_NoReadersSem ,
				 0,    //not shared
				 0 );  //not signaled
	if (retCode != 0) throw InternalRCException("Can not create the synchronization sempahore object");
#else
	m_NoReadersEvent = CreateEvent(NULL, true, true, NULL);
	if (m_NoReadersEvent==NULL) throw InternalRCException("Can not create the synchronization event object");
#endif
}

WriteSync::~WriteSync(void)
{
	#ifdef __GNUC__
		int retCode = sem_destroy ( &m_NoReadersSem );
		if (retCode != 0) rclog << lock << "Error while closing the synchronization semaphore object" << unlock;
	#else
		if (!CloseHandle(m_NoReadersEvent)) rclog << lock << "Error while closing the synchronization event object" << unlock;
	#endif
	pthread_mutex_destroy(&m_ReadersCS);
	pthread_mutex_destroy(&m_WritersCS);
}

void WriteSync::LockByReader()
{
	#ifdef __GNUC__
	pthread_t thrId = pthread_self();
	#else
	unsigned long thrId = GetCurrentThreadId();
	#endif
	int i = 0;
	pthread_mutex_lock(&m_ReadersCS);
	if (m_Readers.empty()) {
	#ifdef __GNUC__
		m_Readers.push_back(thrId);
	#else
		if (!ResetEvent(m_NoReadersEvent))
			throw InternalRCException("Can not reset the synchronization event object");
		m_Readers.push_back(thrId);
	#endif
	}
	pthread_mutex_unlock(&m_ReadersCS);
}

void WriteSync::UnlockByReader()
{
	#ifdef __GNUC__
	pthread_t thrId = pthread_self();
	#else
	unsigned long thrId = GetCurrentThreadId();
	#endif

	pthread_mutex_lock(&m_ReadersCS);
	int i = (int)m_Readers.size()-1;
	while (i >= 0 && m_Readers[i] != thrId) i--;
	if (i < 0) throw InternalRCException("Reader's unlock not preceded by lock");
	for (; i < (int)m_Readers.size()-1; i++) m_Readers[i] = m_Readers[i+1];
	m_Readers.pop_back();
	if (m_Readers.size()==m_nReadingLocksByWriter)
	{
		m_nReadingLocksByWriter = 0;
		#ifdef __GNUC__
		   if( !sem_post ( &m_NoReadersSem ) )
			throw InternalRCException("Can not set the synchronization event object");
	   	   if (!m_Readers.empty())
			;
		#else
		if (!SetEvent(m_NoReadersEvent))
			throw InternalRCException("Can not set the synchronization event object");
		if (!m_Readers.empty())
			if (!ResetEvent(m_NoReadersEvent))
				throw InternalRCException("Can not reset the synchronization event object");
		#endif
	}
	pthread_mutex_unlock(&m_ReadersCS);
}

void WriteSync::WaitForNoReaders()
{
	#ifdef __GNUC__
	pthread_t thrId = pthread_self();
	#else
	unsigned long thrId = GetCurrentThreadId();
	#endif

	pthread_mutex_lock(&m_WritersCS);
	pthread_mutex_lock(&m_ReadersCS);
	for (unsigned int i = 0; i < m_Readers.size(); i++) {
		#ifdef __GNUC__
		if( pthread_equal ( m_Readers[i], thrId) ) m_nReadingLocksByWriter++;
		#else
		if (m_Readers[i]==thrId) m_nReadingLocksByWriter++;
		#endif
	}
	if (m_Readers.size()==m_nReadingLocksByWriter)
	{
			m_nReadingLocksByWriter = 0;
			pthread_mutex_unlock(&m_ReadersCS);
			pthread_mutex_unlock(&m_WritersCS);
			return;
	}
	pthread_mutex_unlock(&m_ReadersCS);

	#ifdef __GNUC__
	int retCode = sem_wait( &m_NoReadersSem );
	if( retCode != 0 )
		throw InternalRCException("Waiting for the synchronization event object failed");
	#else
	unsigned long res = WaitForSingleObject(m_NoReadersEvent, INFINITE);
	if (res != WAIT_OBJECT_0)
		throw InternalRCException("Waiting for the synchronization event object failed");
	#endif
	pthread_mutex_unlock(&m_WritersCS);
}

bool WriteSync::NoReaders()
{
	bool result = false;
	#ifdef __GNUC__
	pthread_t thrId = pthread_self();
	#else
	unsigned long thrId = GetCurrentThreadId();
	#endif

	int ownLocks = 0;
	pthread_mutex_lock(&m_ReadersCS);
	for (unsigned int i = 0; i < m_Readers.size(); i++) {
		#ifdef __GNUC__
		if( pthread_equal ( m_Readers[i], thrId) ) ownLocks++;
		#else
		if (m_Readers[i]==thrId) ownLocks++;
		#endif
	}
	if (m_Readers.size()==ownLocks) result = true;
	pthread_mutex_unlock(&m_ReadersCS);
	return result;
}

#endif // _SYSTEM_WRSYNC_H_

