/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "RCDataTypes.h"
#include "common/bhassert.h"
#include "core/tools.h"
#include "vc/VirtualColumn.h"

using namespace std;
RCBString::RCBString()				// null string
{
	null = true;
	len = 0;
	val = 0;
	pos = 0;
	persistent = false;
	zero_term = false;
}

RCBString::RCBString(const char *val, int len, bool persistent, bool zero_term)
	:	persistent(persistent), zero_term(zero_term)
{
	// NOTE: we allow val to be NULL. In this case, no value will be copied (just reserve a place for future use). Only persistent!
	pos = 0;
	this->null = false;
	if(len == -1) {
		this->len = *(ushort*)val;
		if(persistent == false)
			this->val = const_cast<char*>(val) + sizeof(ushort);
		else {
			this->val = new char [this->len + (zero_term ? 1 : 0)];
			if(val)
				memcpy(this->val, val + sizeof(ushort), this->len);
			if(zero_term)
				this->val[len] = 0;
		}
	} else if(len == -2) {
		this->len = *(int*)val;
		if(persistent == false)
			this->val = const_cast<char*>(val) + sizeof(int);
		else {
			this->val = new char [this->len + (zero_term ? 1 : 0)];
			if(val)
				memcpy(this->val, val + sizeof(int), this->len);
			if(zero_term)
				this->val[len] = 0;
		}
	} else {
		if(len == 0) {
			if(val)
				this->len = (int)strlen(val);
			else
				this->len = 0;
		} else
			this->len = (ushort)len;
		if(persistent == false)
			this->val = const_cast<char*>(val);
		else {
			this->val = new char [this->len + (zero_term ? 1 : 0)];
			if(val)
				memcpy(this->val, val, this->len);
			if(zero_term)
				this->val[len] = 0;
		}
	}
}

RCBString::RCBString(const RCBString& rcbs)
	: pos(rcbs.pos), persistent(rcbs.persistent), zero_term(rcbs.zero_term)
{
	this->null = rcbs.null;
	if(!null) {
		len = rcbs.len;
		if(persistent) {
			val = new char [len + pos + (zero_term ? 1 : 0)];
			memcpy(val, rcbs.val, len + pos);
			if(zero_term)
				val[len+pos] = 0;
		} else
			val = rcbs.val;
	} else {
		len = 0;
		val = 0;
		pos = 0;
		persistent = false;
		zero_term = false;
	}
}

RCBString::~RCBString()
{
	if(persistent)
		delete [] val;
}

RCDataType& RCBString::operator=(const RCDataType& rcdt)
{
	if (this == &rcdt)
		return *this;

	if(rcdt.GetValueType() == STRING_TYPE)
		*this = (RCBString&)rcdt;
	else
		BHERROR("bad cast");

	return *this;
}

bool RCBString::Parse(RCBString& in, RCBString& out)
{
	out = in;
	return true;
}

AttributeType RCBString::Type() const
{
	return RC_STRING;
}

void RCBString::PutString(char* &dest, ushort len, bool move_ptr) const
{
	BHASSERT(this->len <= len, "should be 'this->len <= len'");
	if(this->len == 0)
		memset(dest, ' ', len);
	else {
		memcpy(dest, val, this->len);
		memset(dest+this->len, ' ', len - this->len);
	}
	if(move_ptr)
		dest += len;
}

void RCBString::PutVarchar(char* &dest, uchar prefixlen, bool move_ptr) const
{
	if(prefixlen == 0)
		PutString(dest, len);
	if(len == 0) {
		memset(dest, 0, prefixlen);
		if(move_ptr)
			dest += prefixlen;
	} else {
		switch (prefixlen) {
			case 1 :
				*(uchar*)dest = (uchar)len;
				break;
			case 2 :
				*(ushort*)dest = (ushort)len;
				break;
			case 4 :
				*(uint*)dest = (uint)len;
				break;
			default:
				BHERROR("not implemented");
		}
		memcpy(dest + prefixlen, val, len);
		if(move_ptr)
			dest += prefixlen + len;
	}
}

void RCBString::Put(char* &dest, ushort len) const
{
	if(len)
		PutString(dest, len);
	else
		PutString(dest, this->len);
}

RCBString& RCBString::operator=(const RCBString& rcbs)
{
	if(this == &rcbs)
		return *this;

	null = rcbs.null;
	zero_term = rcbs.zero_term;
	if(persistent)
		delete [] val;
	persistent = rcbs.persistent;
	if(null) {
		val = 0;
		len = 0;
		pos = 0;
		zero_term = false;
	} else {
		len = rcbs.len;
		persistent = rcbs.persistent;
		pos = rcbs.pos;
		if(persistent) {
			val = new char [len + pos + (zero_term ? 1 : 0)];
			memcpy(val, rcbs.val, len + pos);
			if(zero_term)
				val[len] = 0;
		} else
			val = rcbs.val;
	}
	return *this;
}

RCBString::operator std::string()
{
	if(len)
		return std::string(val, len);
	return std::string();
}

char& RCBString::operator*()
{
	return val[pos];
}

RCBString& RCBString::operator++()
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(len > 0);
	pos++;
	len--;
	return *this;
}

RCBString& RCBString::operator++(int)
{
	return ++(*this);
}

RCBString& RCBString::operator--()
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(pos > 0);
	pos--;
	len++;
	return *this;
}

RCBString& RCBString::operator--(int)
{
	return --(*this);
}

char& RCBString::operator[](size_t pos) const
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(pos < len);		// Out of RCBString. Note: val is not ended by '\0'.
	return val[this->pos + pos];
}

RCBString& RCBString::operator+=(ushort pos)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT((int)len - pos >= 0);
	this->pos = this->pos + (ushort)pos;
	this->len -= pos;
	return *this;
}

RCBString RCBString::operator+(ushort pos)
{
	return RCBString(this->val + pos, len - pos);
}

bool RCBString::Like(const RCBString& pattern)
{
	//	MakePersistent();
	if(pattern.IsEmpty())
		return this->IsEmpty();
	char *p=pattern.val;		// just for short...
	int was_wild=0;				// are we in "after %" mode?
	uint cur_p = 0,  cur_p_beg = 0;		// pattern positions
	uint cur_s = 0,  cur_s_beg = 0;		// source positions
	do {
		while(cur_p<pattern.len && p[cur_p]=='%') {		// first omit all %
			was_wild=1;
			cur_p++;
		}
		cur_s_beg = cur_s;
		cur_p_beg = cur_p;
		do {								// internal loop: try to match a part between %...%
			while( cur_p<pattern.len && cur_s<len &&				// find the first match...
				  (val[pos+cur_s]==p[cur_p] || p[cur_p]=='_') &&
				  p[cur_p]!='%' ) {
				cur_s++;
				cur_p++;
			}
			if(cur_s<len && ((cur_p<pattern.len && p[cur_p]!='%') || cur_p>=pattern.len)) {		// not matching (loop finished prematurely) - try the next source position
				if(was_wild==0) return false;	// no % up to now => the first non-matching is critical
				cur_p = cur_p_beg;
				cur_s = ++cur_s_beg;			// step forward in the source, rewind the matching pointers
			}
			if(cur_p<pattern.len && cur_s==len && p[cur_p]!='%') return false;	// end of the source - no more chances for matching
			if(cur_p==pattern.len && cur_s==len) return true;
		}
		while(cur_p<pattern.len && p[cur_p]!='%');		// try the next match position
	}
	while(cur_p<pattern.len && cur_s<len);
	return true;
}

RCBString& RCBString::MakePersistent(bool zero_term)
{
	if(!persistent || this->zero_term != zero_term) {
		this->zero_term = zero_term;
		int tmp_len = len + pos +(zero_term ? 1 : 0);
		char* n_val = new char [tmp_len];
		memcpy(n_val, val, len + pos);
		if(persistent) {
			delete [] val;
			val = 0;
		}
		if(zero_term)
			n_val[len+pos] = 0;
		val = n_val;
		persistent = true;
	}
	return *this;
}

bool RCBString::GreaterEqThanMin(_int64 txt_min)
{
	if(null == true)
		return false;
	unsigned char *min = (unsigned char *)(&txt_min);
	uint min_len = 8;
	while(min_len > 0 && min[min_len - 1] == '\0') 
		min_len--;
	for(uint i = 0; i < min_len && i < len; i++)
		if(((unsigned char*)val)[i + pos] < min[i])
			return false;
		else if(((unsigned char*)val)[i + pos] > min[i])
			return true;
	if(len < min_len)
		return false;
	return true;
}

bool RCBString::GreaterEqThanMinUTF(_int64 txt_min, DTCollation col, bool use_full_len)
{
	if(null == true)
		return false;
	if(RequiresUTFConversions(col)) {
		uint useful_len = 0;
		uint min_byte_len = memchr((char*)(&txt_min), 0, 8) ? strlen((char*)&txt_min) : 8;
		if(!use_full_len) {
			uint min_charlen = col.collation->cset->numchars(col.collation,(char*)(&txt_min), (char*)(&txt_min)+min_byte_len);

			int next_char_len, chars_included = 0;
			while(true) {
				if(useful_len >= len || chars_included == min_charlen)
				  break;
				next_char_len = col.collation->cset->mbcharlen(col.collation, (uchar)val[useful_len+pos]);
				assert("wide character unrecognized" && next_char_len > 0);
				useful_len += next_char_len;
				chars_included++;
			}
		} else
			useful_len = len;
		return col.collation->coll->strnncoll(col.collation, (uchar*)val, useful_len, (uchar*) &txt_min, min_byte_len, 0) >= 0;
	} else
		return GreaterEqThanMin(txt_min);
}

bool RCBString::LessEqThanMax(_int64 txt_max)
{
	if(null == true)
		return false;
	unsigned char *max = (unsigned char *)(&txt_max);
	for(uint i=0;i<8 && i<len;i++)
		if(((unsigned char*)val)[i+pos]>max[i])
			return false;
		else if(((unsigned char*)val)[i+pos]<max[i])
			return true;
	return true;
}

bool RCBString::LessEqThanMaxUTF(_int64 txt_max, DTCollation col, bool use_full_len)
{
	if(null == true)
		return false;
	if(RequiresUTFConversions(col)) {
		uint useful_len = 0;
		uint max_byte_len = memchr((char*)(&txt_max), 0, 8) ? strlen((char*)&txt_max) : 8;
		if(!use_full_len) {

			uint max_charlen = col.collation->cset->numchars(col.collation,(char*)(&txt_max), (char*)(&txt_max)+max_byte_len);

			int next_char_len, chars_included = 0;
			while(true) {
				if(useful_len >= len || chars_included == max_charlen)
					break;
				next_char_len = col.collation->cset->mbcharlen(col.collation, (uchar)val[useful_len+pos]);
				assert("wide character unrecognized" && next_char_len > 0);
				useful_len += next_char_len;
				chars_included++;
			}
		} else
			useful_len = len;
		return col.collation->coll->strnncoll(col.collation, (uchar*)val, useful_len, (uchar*) &txt_max, max_byte_len, 0) <= 0;
	} else
		return LessEqThanMax(txt_max);
}

bool RCBString::IsEmpty() const
{
	if(null == true)
		return false;
	return len == 0 ? true : false;
}

bool RCBString::IsNullOrEmpty() const
{
	return ((len == 0 || null) ? true : false);
}

size_t RCBString::size() const
{
	return (size_t)len;
}

bool RCBString::operator==(const RCDataType& rcdt) const
{
	if(null || rcdt.IsNull())
		return false;
	if(rcdt.GetValueType() == STRING_TYPE)
		return strcmp(*this, (RCBString&)rcdt) == 0;
	BHERROR("bad cast");
	return false;
}

bool RCBString::operator==(const RCBString& rcs) const
{
	if(null || rcs.IsNull())
		return false;
	return strcmp(*this, rcs) == 0;
}

bool RCBString::operator<(const RCDataType& rcdt) const
{
	if(null || rcdt.IsNull())
		return false;
	if(rcdt.GetValueType() == STRING_TYPE)
		return strcmp(*this, (RCBString&)rcdt)< 0;
	BHERROR("bad cast");
	return false;
}

bool RCBString::operator>(const RCDataType& rcdt) const
{
	if(null || rcdt.IsNull())
		return false;
	if(rcdt.GetValueType() == STRING_TYPE)
		return strcmp(*this, (RCBString&)rcdt) > 0;
	BHERROR("bad cast");
	return false;
}

//bool RCBString::Greater(const RCDataType& rcdt, DTCollation collation) const
//{
//	if(null || rcdt.IsNull())
//		return false;
//	if(rcdt.GetValueType() == STRING_TYPE)
//		return strcmp(*this, (RCBString&)rcdt, collation.collation) > 0;
//	BHERROR("bad cast");
//	return false;
//}

bool RCBString::operator>=(const RCDataType& rcdt) const
{
	if(null || rcdt.IsNull())
		return false;
	if(rcdt.GetValueType() == STRING_TYPE)
		return strcmp(*this, (RCBString&)rcdt) >= 0;
	BHERROR("bad cast");
	return false;
}

bool RCBString::operator<=(const RCDataType& rcdt) const
{
	if(null || rcdt.IsNull())
		return false;
	if(rcdt.GetValueType() == STRING_TYPE)
		return strcmp(*this, (RCBString&)rcdt) <= 0;
	BHERROR("bad cast");
	return false;
}

bool RCBString::operator!=(const RCDataType& rcdt) const
{
	if(null || rcdt.IsNull())
		return true;
	if(rcdt.GetValueType() == STRING_TYPE)
		return strcmp(*this, (RCBString&)rcdt) != 0;
	BHERROR("bad cast");
	return true;
}

uint RCBString::GetHashCode() const
{
	if(null)
		return 0;
	uint hc = 0;
	int a = 1040021;
	for(uint i = 0; i < len; i++)
		hc = (hc*a + val[i])&1048575;
	return hc;
}

size_t RCBString::GetStorageByteSize() const
{
	return sizeof(null) + (null ? 0 : sizeof(zero_term) + sizeof(len) + sizeof(pos) + (len + pos));
}

void RCBString::ToByteStream(char*& buf) const
{
	store_bytes(buf, null);
	if(!null) {
		store_bytes(buf, zero_term);
		store_bytes(buf, pos);
		this->PutVarchar(buf, 4, true);
	}
}

void RCBString::AssignFromByteStream(char*& buf)
{
	if(persistent)
		delete [] val;
	len = 0;
	pos = 0;
	val = 0;
	zero_term = false;
	persistent = true;
	unstore_bytes(null, buf);
	if(null) {
		zero_term = false;
	} else 	{
		unstore_bytes(zero_term, buf);
		unstore_bytes(pos, buf);
		unstore_bytes(len, buf);
		val = new char [len + (zero_term ? 1 : 0)];
		memcpy(val, buf, len);
		buf += len;
		if(zero_term)
			val[len] = 0;
	}
}

RCBString& RCBString::RightTrim(RCBString& str)
{
	int last = str.len - 1;
	while(last >= 0 && str.val[last--] == ' ')
		--str.len;
	if(str.zero_term)
		str.val[str.len] = 0;
	return str;
}

std::ostream& operator<<(std::ostream &out, const RCBString& rcbs)
{
	out.write(rcbs.val + rcbs.pos, rcbs.len);
	return out;
}

std::size_t strlen(const RCBString& rcbs)
{
	return rcbs.len;
}

void strcpy(char* dest, const RCBString& rcbs)
{
	memcpy(dest, rcbs.val+rcbs.pos, rcbs.len);
	dest[rcbs.len] = 0;
}

void strcpy(uchar* dest, const RCBString& rcbs)
{
	memcpy(dest, (uchar*)rcbs.val+rcbs.pos, rcbs.len);
	dest[rcbs.len] = 0;
}

// TODO: finish this function, peristent, zero_term, etc.
void strcpy(RCBString& dest, const RCBString& rcbs)
{
	dest = rcbs;
}

void strncpy(char* dest, const RCBString& rcbs, unsigned count)
{
	strncpy((uchar*)dest, rcbs, count);
}

void strncpy(uchar* dest, const RCBString& rcbs, unsigned count)
{
	uint len =  (rcbs.len - rcbs.pos) < count ? (rcbs.len - rcbs.pos) : count;
	memcpy(dest, (uchar*)(rcbs.val + rcbs.pos), len);
	if(len <= count)
		memset(dest + len, 0, count - len);
}

int strcmp(const RCBString& rcbs1, const RCBString& rcbs2)
{
	int ret = 0;
	int len = min(rcbs1.len, rcbs2.len);
	if(len == 0) {
		if(rcbs1.len == 0 && rcbs2.len == 0)
			ret = 0;
		else if(rcbs1.len == 0)
			ret = -1;
		else
			ret = 1;
	} else if(rcbs1.len != rcbs2.len) {
		ret = memcmp(rcbs1.val + rcbs1.pos, rcbs2.val + rcbs2.pos, len);
		if(ret == 0)
			if(rcbs1.len < rcbs2.len)
				return -1;
			else
				return 1;
	} else
		ret = memcmp(rcbs1.val + rcbs1.pos, rcbs2.val + rcbs2.pos, len);

	return ret;
}

//int strcmp(const RCBString& rcbs1, const RCBString& rcbs2, CHARSET_INFO* charset)
//{
//	int ret = 0;
//	int len = max(rcbs1.len, rcbs2.len);
//	if(len == 0) {
//		if(rcbs1.len == 0 && rcbs2.len == 0)
//			ret = 0;
//		else if(rcbs1.len == 0)
//			ret = -1;
//		else
//			ret = 1;
//		return ret;
//	} 
//	
//	int buf_len = len * charset->strxfrm_multiply;
//	uchar* buf1 = new uchar[buf_len];
//	uchar* buf2 = new uchar[buf_len];
//	charset->coll->strnxfrm(charset, buf1, buf_len, (uchar*)(rcbs1.val + rcbs1.pos), len);
//	charset->coll->strnxfrm(charset, buf2, buf_len, (uchar*)(rcbs2.val + rcbs2.pos), len);
//	ret = memcmp(buf1, buf2, buf_len);
//	delete [] buf1;
//	delete [] buf2;
//	if(rcbs1.len != rcbs2.len) {
//		if(ret == 0)
//			if(rcbs1.len < rcbs2.len)
//				return -1;
//			else
//				return 1;
//	} 
//
//	return ret;
//}

int stricmp(const RCBString& rcbs1, const RCBString& rcbs2)
{
	int ret = 0;
	int len = min(rcbs1.len, rcbs2.len);
	if(len == 0) {
		if(rcbs1.len == 0 && rcbs2.len == 0)
			ret = 0;
		else if(rcbs1.len == 0)
			ret = -1;
		else
			ret = 1;
	} else if(rcbs1.len != rcbs2.len) {
		ret = strncasecmp(rcbs1.val+rcbs1.pos, rcbs2.val+rcbs2.pos, len);
		if(ret == 0)
			if(rcbs1.len < rcbs2.len)
				return -1;
			else
				return 1;
	} else
		ret = strncasecmp(rcbs1.val+rcbs1.pos, rcbs2.val+rcbs2.pos, len);

	return ret;
}

bool operator!=(const RCBString& rcbs1, const RCBString& rcbs2)
{
	if(rcbs1.IsNull() || rcbs2.IsNull())
		return true;
	return strcmp(rcbs1, rcbs2) != 0;
}

////////////////////////////////////////////////////////////////////////////////////

TextStat::TextStat(ushort no_chars_to_omit)
	: no_chars_to_omit(no_chars_to_omit)
{
	Reset();
}

TextStat::TextStat(TextStat &sec)
{
	int i = 0;
	for(; i < 256; i++) {
		existTable[i] = sec.existTable[i];
		for(int j = 0; j < 64; j++)
			encodeTable[i][j] = sec.encodeTable[i][j];
	}

	for(i = 0; i < 64; i++)
		lenTable[i] = sec.lenTable[i];

	no_chars_to_omit = sec.no_chars_to_omit;
	no_encoded_chars = sec.no_encoded_chars;
	max_string_size = sec.max_string_size;
	nontrivial_encoding = sec.nontrivial_encoding;
}

TextStat::~TextStat(void)
{
}

void TextStat::Reset()
{
	int i = 0;
	for(; i < 256; i++) {
		existTable[i] = 0;
		for(int j = 0; j < 64; j++)
			encodeTable[i][j] = 0;
	}
	for(i = 0; i < 64; i++)
		lenTable[i] = 0;
	no_encoded_chars = 0;
	max_string_size = 0;
	nontrivial_encoding = false;
}

void TextStat::AddString(const RCBString& rcbs)
{
	int to_encode = (int)rcbs.size() - no_chars_to_omit;
	uint stop = (to_encode < 64 ? to_encode : 64) + no_chars_to_omit;

	uint l = 0;
	for(uint i = no_chars_to_omit; i < stop; i++, l++)
		existTable[(uchar)rcbs.val[i]] |= (_uint64)(1ULL << l);
	if(to_encode > max_string_size)
		max_string_size = to_encode;
}

bool TextStat::AddChar(uchar v, int pos)	// return false if out of range
{
	assert(pos - no_chars_to_omit >= 0);
	if(pos - no_chars_to_omit < 64) {
		existTable[v] |= (_uint64)(1ULL << (pos - no_chars_to_omit));
		if(pos + 1 > max_string_size)
			max_string_size = pos + 1;		// size = position + 1
	}
	else
		return false;
	return true;
}

void TextStat::CreateEncoding()
{
	int encode_len = 0;
	no_encoded_chars = 0;
	int i = 0;
	for(; i < 64; i++)
		lenTable[i] = 0;
	for(i = 0; i < 256; i++)
		for(int j = 0; j < 64; j++)
			if(existTable[i] & (_uint64)(1ULL << j))
				lenTable[j]++;

	for(int j = 0; j < 64; j++) {
		lenTable[j] = GetBitLen((uint)lenTable[j]);		// Note that for 256 values (all possible characters) we need 9 bits! Because no character (end of text) must be encoded too.
		encode_len += lenTable[j];
		if(encode_len > 63) {
			no_encoded_chars = j - 1;
			for(; j < 64; j++)
				lenTable[j] = 0;
			break;
		}
	}

	if(encode_len <= 63) {
		no_encoded_chars = 64;		// 64 when encodeLen did not reach 64 bits, but the encoder overpassed 64 characters and was ended (possible e.g. when a lot of identical strings are present)
		if(max_string_size<=no_encoded_chars)
			no_encoded_chars = -1;					// i.e. the whole column is done - in most cases when encodeLen <= 63
	}

	int count = 0;
	for(int j = 0; j < 64; j++) {
		count = 1;
		for(i = 0; i < 256; i++) {
			if(existTable[i] & (_uint64)(1ULL << (int)j)) {
				if(count==256)								// i.e. the last character of full character list
					encodeTable[i][j] = 0;
				else
					encodeTable[i][j] = count++;
			} else
				encodeTable[i][j] = 0;					// may mean no characters (i.e. character not possible) or 255
		}
	}
}

void TextStat::CreateFull(int len)
{
	int encode_len = 0;
	max_string_size = len - no_chars_to_omit;
	for(int j = 0; j < 64; j++) {
		lenTable[j] = (j<max_string_size ? 9 : 0);						// 9, because we must use 9 bits to store 256 values and special 0 (end of text)
		encode_len += lenTable[j];
		if(encode_len > 63) {
			no_encoded_chars = j - 1;
			for(; j < 64; j++)
				lenTable[j] = 0;
			break;
		}
		for(int i = 0; i < 255; i++)
			encodeTable[i][j] = i+1;
		encodeTable[255][j] = 0;
	}
	if(encode_len <= 63) {
		if(max_string_size <= 64)
			no_encoded_chars = -1;	// i.e. the whole column is done - in most cases when encodeLen <= 63
		else
			no_encoded_chars = 64;	// 64 when encode_len did not reach 64 bits, but the encoder overpassed 64 characters and was ended (possible e.g. when a lot of identical strings are present)
	}
}

_uint64 TextStat::Encode(const RCBString& rcbs)
{
	_uint64 res=0;
	_uint64 val;
	ushort stop_encode = no_encoded_chars + no_chars_to_omit;			// note that no_encoded_chars <= 64
	if(no_encoded_chars==-1)														// all encoded
		stop_encode = max_string_size + no_chars_to_omit;

	int p = 0;
	for(ushort l = no_chars_to_omit; l < stop_encode; l++, p++)	{		// l and p have not the same value!
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(p<64);
		if(lenTable[p]>0) {
			res = res << lenTable[p];
			if(l < rcbs.size()) {
				nontrivial_encoding = true;
				val = encodeTable[(uchar)rcbs.val[l]][p];
				if(val==0) {
					if(lenTable[p]==9)							// 0 means 256 in this case
						val = 256;
					else
						return NULL_VALUE_64;					// impossible character found
				}
				res += val;
			}
		}
	}
	return res;
}

_uint64 TextStat::MaxCode()
{
	_uint64 res=0;
	int max;
	ushort stop_encode = (no_encoded_chars == -1 ? max_string_size : no_encoded_chars);
	for(ushort l = 0; l < stop_encode; l++) {
		if(lenTable[l]>0){
			res = res << lenTable[l];			// find a maximal code for this position
			if(lenTable[l]==9)
				res += 256;
			else {
				max=0;
				for(int j=0;j<256;j++)
					if( max<encodeTable[j][l] )
						max = encodeTable[j][l];
				res += _uint64(max);
			}
		}
	}
	return res;
}

void ConvertToBinaryForm(VirtualColumn* vc, const MIIterator& mit, RCBString& buf, DTCollation coll)
{
	if(!vc->IsNull(mit)) {
		RCBString s;
		vc->GetValueString(s, mit);
		coll.collation->coll->strnxfrm(coll.collation, (uchar*)buf.val, buf.len, (uchar*)(s.val + s.pos), s.len);
	} else
		buf.null = true;
}
